/*
    This file is part of corona-13.

    corona-13 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    corona-13 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with corona-13. If not, see <http://www.gnu.org/licenses/>.
*/

#include "guided3v/config.h"
#include "guided3v/guided.h"
#include "guided3v/pixel_est.h"
#include "guided3v/stats.h"

#include "camera.h"
#define FILTER_ATOMIC   // needed for spline filter even in the absence of lt :(
#include "filter.h"
#include "framebuffer.h"
#include "matrix3.h"
#include "pathspace/nee.h"
#include "points.h"
#include "pointsampler.h"
#include "render.h"
#include "screenshot.h"
#include "spectrum.h"
#include "view.h"

#include <assert.h>
#include <float.h>

typedef struct fake_randoms_t
{
  int enabled;
  float rand[40];
}
fake_randoms_t;

typedef struct pointsampler_t
{
  fake_randoms_t *rand;
  guided_t *guided;
  render_mode_t render_mode; // initialized to rm_init by zeroing
  pixel_est_t *px_est;
  int n_learn_frames;
  double start_time;
  double learn_time;
}
pointsampler_t;

static void splat_pixel_intensity(pointsampler_t *s, const path_t *path, float throughput);

const char prev_cache_name[] = "prev_gmm.dump";

void pointsampler_print_info(FILE *f)
{
  fprintf(f, "mutations: guided sampling with vertex triplets\n");
  fprintf(f, "learning took %gs (%d iterations)\n",
      rt.pointsampler->learn_time,
      rt.pointsampler->n_learn_frames);
  fprintf(f, "Compile Flags: VERTEX_PAIRS = %d, ENABLE_LT = %d, LT_LEARN_ONLY = %d, MF_COUNT = %d\n", VERTEX_PAIRS, ENABLE_LT, LT_LEARN_ONLY, MF_COUNT);
}

pointsampler_t *pointsampler_init(uint64_t frame)
{
  pointsampler_t *s = calloc(1, sizeof(*s));
  s->rand = calloc(rt.num_threads, sizeof(*s->rand));

#if REL_CONTRIBUTION
  s->px_est = pixel_est_init(view_width(), view_height(), 1 + 6);
#endif

  return s;
}

float pointsampler(path_t *p, int dim)
{
  // pure random mersenne twister
  const int tid = common_get_threadid();
  if(rt.pointsampler->rand[tid].enabled)
    return rt.pointsampler->rand[tid].rand[dim];
  return points_rand(rt.points, tid);
}

void pointsampler_splat(path_t *p, mf_t value)
{
  assert(0); // only called by the sampler?
}

static void splat_path(pointsampler_t *s, path_t *path, int lt)
{
  guided_t *g = s->guided;

  const int ev = lt ? 0 : path->length - 1;
  const int sv = lt ? path->length - 1 : 0;
  if(!((path->v[ev].mode & s_emit) && (path->v[sv].mode & s_sensor))) return;

  md_t mcf_dwp = path_measurement_contribution_dwp(path, 0, path->length - 1);
  if(mf_all(mf_lte(md_2f(mcf_dwp), mf_set1(0.0f)))) return;

#if ENABLE_LT
  const bool lt_active = LT_LEARN_ONLY
    ? (s->render_mode == rm_learn)
    : true;
#else
  const bool lt_active = false;
#endif

  if(!lt_active) assert(!lt);

  md_t throughput;
  path_t *camera_path = NULL;
  path_t *light_path = NULL;
  if(lt_active)
  {
    path_t reversed;
    path_reverse(&reversed, path);

    camera_path = lt ? &reversed : path;
    light_path  = lt ? path      : &reversed;
  }
  else camera_path = path;
  md_t pdf_camera_path = guided_eval_pdf_pt(g, camera_path);
  md_t pdf_total;
  if(lt_active)
  {
    guided_reuse_lobe_list(g, path->length, lt);
    md_t pdf_light_path = guided_eval_pdf_lt(g, light_path);
    pdf_total = md_add(pdf_light_path, pdf_camera_path);
  }
  else pdf_total = pdf_camera_path;
  throughput = md_div(mcf_dwp, md_set1(md_hsum(pdf_total)));
  path->throughput = md_2f(throughput);

  if(!mf_any(mf_gt(md_2f(throughput), mf_set1(0.0f))) || !mf_all(mf_lt(md_2f(throughput), mf_set1(FLT_MAX)))) return;
  render_splat(path, md_2f(throughput));

  if(path->length < 3) return;

  if(s->render_mode == rm_learn)
  {
    splat_pixel_intensity(s, path, md_hsum(throughput));
    guided_successful_path(g, path);
  }

  guided_record_path(g, camera_path, md_hsum(throughput), s->render_mode);
}

static void trace_camera_path(path_t *path)
{
  pointsampler_t *s = rt.pointsampler;
  guided_t *g = s->guided;

#if ENABLE_LT && GUIDE_LT
  const bool guided_lt_active = LT_LEARN_ONLY
    ? (s->render_mode == rm_learn)
    : true;
#else
  const bool guided_lt_active = false;
#endif

  guided_new_path(g);
  while(1)
  {
    const int err = guided_extend(g, path);
    if(err) return;
    if(path->v[path->length-1].mode & s_emit) break;

#if ENABLE_NEE
    if(nee_sample(path)) return;
    splat_path(s, path, 0);
    int old_length = path->length;
    path_pop(path);
    assert(path->length == old_length - 1);
    if(guided_lt_active)
      guided_invalidate_lobe_list(g, path->length, false);
#endif
  }
  splat_path(s, path, 0);
}

#if ENABLE_LT
static void trace_light_path(path_t *path)
{
  pointsampler_t *s = rt.pointsampler;
  guided_t *g = s->guided;

  guided_new_path(g);
  path->v[0].mode = s_emit;
  while(1)
  {
    const int err = guided_extend(g, path);
    if(err) return;

    if(nee_sample(path)) return;
    splat_path(s, path, 1);
    int old_length = path->length;
    path_pop(path);
    assert(path->length == old_length - 1);
    guided_invalidate_lobe_list(g, path->length, true);
  }
}
#endif

void pointsampler_mutate(path_t *curr, path_t *tent)
{
  path_init(tent, tent->index, tent->sensor.camid);
  trace_camera_path(tent);
#if ENABLE_LT
#if LT_LEARN_ONLY
  pointsampler_t *s = rt.pointsampler;
  if(s->render_mode == rm_learn)
#endif
  {
    path_init(tent, tent->index, tent->sensor.camid);
    trace_light_path(tent);
  }
#endif
}

void pointsampler_mutate_with_pixel(path_t *curr, path_t *tent, float i, float j)
{
  assert(0);
}

int pointsampler_accept(path_t *curr, path_t *tent) { return 0; }

void pointsampler_clear() {
  // TODO clear cache
}

void pointsampler_cleanup(pointsampler_t *s)
{
#if REL_CONTRIBUTION
  pixel_est_cleanup(s->px_est);
#endif
  if(s->guided != NULL) guided_cleanup(s->guided);

  free(s->rand);
  free(s);
}
void pointsampler_set_large_step(pointsampler_t *t, float p_large_step) {}
void pointsampler_reset_thread(pointsampler_t *t) {}
void pointsampler_finalize(pointsampler_t *s) {}

void pointsampler_enable_fake_random(pointsampler_t *s)
{
  assert(0);
  const int tid = common_get_threadid();
  s->rand[tid].enabled = 1;
}

void pointsampler_disable_fake_random(pointsampler_t *s)
{
  assert(0);
  const int tid = common_get_threadid();
  s->rand[tid].enabled = 0;
}

void pointsampler_set_fake_random(pointsampler_t *s, int dim, float rand)
{
  assert(0);
  const int tid = common_get_threadid();
  s->rand[tid].rand[dim] = rand;
}

static int is_power_2(unsigned int x)
{
  return (x & (x - 1)) == 0;
}

void pointsampler_prepare_frame(pointsampler_t *s)
{
  if(s->guided == NULL)
  { // delayed init so scene aabb is available
    s->guided = guided_init();
    if(rt.load_cache)
    { // load cache from previous frame (if file exists)
      guided_load_cache(s->guided, prev_cache_name);
    }
  }

  printf("\n"); // dont overwrite the s/frame line with debug output

  guided_t *c = s->guided;

  if(rt.frames == 0)
  {
    s->render_mode = rm_learn;
    s->start_time = common_time_wallclock();
  }

  if(rt.frames > 0 && s->render_mode == rm_learn)
  {
    int learn_iter = rt.frames;
    guided_update(c, s->px_est, learn_iter);
    if(learn_iter > 0 && is_power_2(learn_iter))
    {
      char filename[1024];
      snprintf(filename, sizeof(filename), "%s_learn%05d", rt.output_filename, learn_iter);
      view_write_images(filename);
    }

    guided_print_stats(c, stderr, 0);

    if(learn_iter > 0)
    {
#if WRITE_DEBUG_INFO
      char filename[1024];
      snprintf(filename, sizeof(filename), "%s%s_cache%05d", rt.basename, rt.output_filename, learn_iter);
      guided_dump_cache_debug(c, filename);
#endif

#if WRITE_STATS
      {
        char filename[1024];
        snprintf(filename, sizeof(filename), "%s%s_stats%05d.json", rt.basename, rt.output_filename, learn_iter);
        FILE *stats_file = fopen(filename, "w");
        guided_print_stats(c, stats_file, 1);
        fclose(stats_file);
      }
#endif
    }
    guided_prepare_frame(c);
  }

  if(s->render_mode == rm_learn)
  {
    double elapsed_time = common_time_wallclock() - s->start_time;
    int stop_learn = 0;
    stop_learn |= (rt.learn_spp != -1) && (rt.frames >= rt.learn_spp);
    stop_learn |= (rt.learn_time != -1) && (elapsed_time >= rt.learn_time);
    if(stop_learn) pointsampler_stop_learning(s);
  }

  if(s->render_mode == rm_render)
  {
    int render_iter = rt.frames - s->n_learn_frames;
    assert(render_iter >= 0);
    if(render_iter > 0 && is_power_2(render_iter))
    {
      char filename[1024];
      snprintf(filename, sizeof(filename), "%s_render%05d", rt.output_filename, render_iter);
      view_write_images(filename);
    }
  }

  guided_reset_stats(c);
#ifdef LEARN_CLEAR
  if(s->render_mode == rm_learn)
  {
    view_clear_frame(rt.view);
    printf("learn iter %ld\n", rt.frames);
  }
#endif
}

void pointsampler_stop_learning(pointsampler_t *s)
{
  s->learn_time = common_time_wallclock() - s->start_time;
  s->render_mode = rm_render;
  guided_dump_cache(s->guided, prev_cache_name);
  view_clear_frame(rt.view);
  s->n_learn_frames = rt.frames;
  {
    guided_t *c = s->guided;
    char filename[1024];
    snprintf(filename, sizeof(filename), "%s%s_stats.json", rt.basename, rt.output_filename);
    FILE *stats_file = fopen(filename, "w");
    guided_print_stats(c, stats_file, 1);
    fclose(stats_file);
  }
}

static void splat_pixel_intensity(pointsampler_t *s, const path_t *path, float throughput)
{
#if REL_CONTRIBUTION
  if(!(throughput > 0.f)) return; // fast path for == 0.0
  if(!(throughput < FLT_MAX)) return; // reject inf
  if(!(throughput == throughput)) return; // filter NaN
  pixel_est_splat(s->px_est, path->sensor.pixel_i, path->sensor.pixel_j, throughput);
#endif
}
