#include "em.h"
#include "fakegaussian2.h"
#include "matrix.h"
#include "lobe.h"
#include "stats.h"
#include "svd_symm.h"
#include "util.h"

#include "corona_common.h"

#include <assert.h>
#include <float.h>
#include <stdlib.h>

#if 0
// from corona_common.h
#define M_PI 3.14159265358979323846
#define MAX(a, b) ((a) > (b) ? (a) : (b))
#define MIN(a, b) ((a) < (b) ? (a) : (b))
#define CLAMP(a, m, M) MIN(MAX(a, m), M)
#endif

static void gmm_fit_init_components(gmm_fit_t *gmm, em_sample_t *samples, uint32_t n_samples, float prior_radius);
static double gmm_fit_em_iter(gmm_fit_t *gmm, em_sample_t *samples, uint32_t n_samples, float prior_radius);
static int gmm_fit_split_component(const gmm_fit_comp_t *parent, gmm_fit_comp_t children[2], em_sample_t *samples, uint32_t n_samples, float log_total_weight, float prior_radius);
static void gmm_fit_split_component_em(em_sample_t *samples, uint32_t length, gmm_fit_comp_t children[2], float prior_radius);
#if 0
static uint32_t gmm_fit_partition_samples(em_sample_t *samples, uint32_t length, const float center[GUIDE_DIM], const float axis[GUIDE_DIM]);
#endif
static uint32_t gmm_fit_partition_samples_em(em_sample_t *samples, uint32_t length, gmm_fit_comp_t children[2], float prior_radius);

static void gmm3_fit_init_components(gmm3_fit_t *gmm, em_sample_t *samples, uint32_t n_samples, float prior_radius);
static double gmm3_fit_em_iter(gmm3_fit_t *gmm, em_sample_t *samples, uint32_t n_samples, float prior_radius);

static int gmm3_fit_split_component(const gmm3_fit_comp_t *parent, gmm3_fit_comp_t children[2], em_sample_t *samples, uint32_t n_samples, float log_total_weight, float prior_radius);
static void gmm3_fit_split_component_em(em_sample_t *samples, uint32_t length, gmm3_fit_comp_t children[2], float prior_radius);
static uint32_t gmm3_fit_partition_samples_em(em_sample_t *samples, uint32_t length, gmm3_fit_comp_t children[2], float prior_radius);

static void accum_sample(gmm_comp_accum_t *accum, const float weight, const float *x);
//static void accum_reweight(gmm_comp_accum_t *accum, float new_weight);
static void accum_merge(const gmm_comp_accum_t *a, const gmm_comp_accum_t *b, gmm_comp_accum_t *result);
static void accum_get_cov(const gmm_comp_accum_t *accum, double *cov);
static void accum_from_comp(gmm_comp_accum_t *accum, const gmm_fit_comp_t *comp);
static void accum_from_comp_with_weight(gmm_comp_accum_t *accum, const gmm_fit_comp_t *comp, float weight);
static double accum_ess(gmm_comp_accum_t *accum);

static void comp_from_accum(gmm_fit_comp_t *comp, gmm_comp_accum_t *accum, float prior_radius);
static void comp_from_accum_no_prior(gmm_fit_comp_t *comp, gmm_comp_accum_t *accum);
static double comp_eval(const gmm_fit_comp_t *comp, const float *x);
static double comp_eval_with_weight(const gmm_fit_comp_t *comp, const float *x);
static double comp_log_likelihood(const gmm_fit_comp_t *comp, float log_total_weight);
static double comp_logdet(const gmm_fit_comp_t *comp);
#if 0
static void comp_copy_to_prior(gmm_fit_comp_t *comp);
#endif

static int svd_d(const double *a, int n, double *w, double *v);
static void regularize_d(int n, double *w);

static void restrict_cov(double *cov, float prior_radius, float ess);

static void accum3_sample(gmm3_comp_accum_t *accum, const float weight, const float *x);
static double comp3_eval_with_weight(const gmm3_fit_comp_t *comp, const float *x);
static double comp3_eval(const gmm3_fit_comp_t *comp, const float *x);
static void comp3_from_accum(gmm3_fit_comp_t *comp, gmm3_comp_accum_t *accum, float prior_radius);
static void accum3_get_cov(const gmm3_comp_accum_t *accum, double *cov);
static void comp3_set_cov(gmm3_fit_comp_t *comp, double *cov);
static void accum3_from_comp(gmm3_comp_accum_t *accum, const gmm3_fit_comp_t *comp);
static double comp3_log_likelihood(const gmm3_fit_comp_t *comp, float log_total_weight);
static double comp3_logdet(const gmm3_fit_comp_t *comp);
static void accum3_merge(const gmm3_comp_accum_t *a, const gmm3_comp_accum_t *b, gmm3_comp_accum_t *result);
static void comp3_from_accum_no_prior(gmm3_fit_comp_t *comp, gmm3_comp_accum_t *accum);

static const uint32_t scratch_item_size = 8;

void gmm_fit_init(gmm_fit_t *gmm)
{
  const uint32_t n_comp_init = 8;
  *gmm = (gmm_fit_t) {
    .n_components = 0,
    .components_allocated = n_comp_init,
    .components = malloc(n_comp_init * sizeof(gmm_fit_comp_t)),
    .accums = malloc(n_comp_init * sizeof(gmm_comp_accum_t)),
    .scratch = malloc(n_comp_init * scratch_item_size),
    .heap = heap_init(n_comp_init),
  };
}

void gmm_fit_cleanup(gmm_fit_t *gmm)
{
  free(gmm->components);
  free(gmm->accums);
  free(gmm->scratch);
  heap_cleanup(gmm->heap);
}

void gmm_fit_resize(gmm_fit_t *gmm, uint32_t n_components)
{
  gmm->n_components = n_components;
  if(gmm->n_components <= gmm->components_allocated) return;
  do gmm->components_allocated *= 2;
  while(gmm->components_allocated < gmm->n_components);
  gmm->components = realloc(gmm->components, sizeof(gmm_fit_comp_t) * gmm->components_allocated);
  gmm->accums = realloc(gmm->accums, sizeof(gmm_comp_accum_t) * gmm->components_allocated);
  gmm->scratch = realloc(gmm->scratch, scratch_item_size * gmm->components_allocated);
  heap_cleanup(gmm->heap);
  gmm->heap = heap_init(gmm->components_allocated);
}

void gmm_fit_samples(gmm_fit_t *gmm, em_sample_t *samples, uint32_t n_samples, float prior_radius, guided_stats_t *stats)
{
  assert(gmm->n_components > 0);
  assert(n_samples > 0);
  double start = common_time_wallclock();
  gmm_fit_init_components(gmm, samples, n_samples, prior_radius);
  double after_init = common_time_wallclock();
  if(gmm->n_components < 2) return;

  double prev_log_likelihood = -FLT_MAX;
  int iter = 0;
  for(iter = 0; iter < EM_ITER; iter++)
  {
    double log_likelihood = gmm_fit_em_iter(gmm, samples, n_samples, prior_radius);
    double change = fabs(log_likelihood - prev_log_likelihood) / fabs(prev_log_likelihood);
    if(change < EM_STOP_TOL) break;
    prev_log_likelihood = log_likelihood;
  }
  double after_em = common_time_wallclock();
  stats_record_em(stats, start, after_init, after_em, iter);
}

void gmm_fit_reduce(gmm_fit_t *gmm, float threshold, double total_weight)
{
  float accumulated_cost = 0;
  while(accumulated_cost < threshold)
  {
    float min_cost = FLT_MAX;
    uint32_t i_min = 0, j_min = 0;
    const uint32_t N = gmm->n_components;
    for(uint32_t i = 0; i < N; i++)
    {
      gmm_comp_accum_t accum_i;
      accum_from_comp(&accum_i, &gmm->components[i]);
      float logdet_i = comp_logdet(&gmm->components[i]);
      for(uint32_t j = i + 1; j < N; j++)
      {
        float logdet_j = comp_logdet(&gmm->components[j]);
        gmm_comp_accum_t accum_j, result;
        accum_from_comp(&accum_j, &gmm->components[j]);
        accum_merge(&accum_i, &accum_j, &result);
        gmm_fit_comp_t merged_comp;
        comp_from_accum_no_prior(&merged_comp, &result);
        float logdet_ij = comp_logdet(&merged_comp);
        float cost = (merged_comp.weight / total_weight) * logdet_ij
          - (gmm->components[i].weight / total_weight) * logdet_i
          - (gmm->components[j].weight / total_weight) * logdet_j;
        if(cost < min_cost)
        {
          min_cost = cost;
          i_min = i;
          j_min = j;
        }
      }
    }
    if(accumulated_cost + min_cost < threshold)
    {
      gmm_comp_accum_t accum_i, accum_j, result;
      accum_from_comp(&accum_i, &gmm->components[i_min]);
      accum_from_comp(&accum_j, &gmm->components[j_min]);
      accum_merge(&accum_i, &accum_j, &result);
      gmm_fit_comp_t merged_comp;
      comp_from_accum_no_prior(&merged_comp, &result);
      gmm->components[i_min] = merged_comp;
      gmm->components[j_min] = gmm->components[gmm->n_components - 1];
      gmm->n_components--;
      accumulated_cost += min_cost;
    }
    else break;
  }
}

uint32_t em_sample_filter_aabb(em_sample_t *samples, uint32_t n_samples, const float *cell_aabb)
{
  uint32_t new_sample_count = 0;
  for(uint32_t i = 0; i < n_samples; i++)
  {
    const em_sample_t *sample = &samples[i];
    if(util_point_in_aabb(cell_aabb, &sample->x[V2OFFSET]))
      samples[new_sample_count++] = *sample;
  }
  return new_sample_count;
}

//--- static functions ---------------------------------------------------------

static void gmm_fit_init_components(gmm_fit_t *gmm, em_sample_t *samples, uint32_t n_samples, float prior_radius)
{ // pca part init
  typedef struct range_t {
    uint32_t offset;
    uint32_t length;
  } range_t;
  assert(sizeof(range_t) <= scratch_item_size);
  range_t *sample_ranges = gmm->scratch;
  heap_clear(gmm->heap);

  float log_total_weight;
  { // we need this to compute log likelihoods later
    double total_weight = 0;
    for(uint32_t sample_i = 0; sample_i < n_samples; sample_i++)
      total_weight += samples[sample_i].weight;
    log_total_weight = log(total_weight);
  }

  uint32_t components_initialized = 1;
  {
    gmm_comp_accum_t accum = {};
    for(uint32_t i = 0; i < n_samples; i++) accum_sample(&accum, samples[i].weight, samples[i].x);
    comp_from_accum(&gmm->components[0], &accum, prior_radius);
    sample_ranges[0] = (range_t) { .offset = 0, .length = n_samples };
    float log_l = comp_log_likelihood(&gmm->components[0], log_total_weight);

    heap_insert(gmm->heap, 0, -log_l);
  }
  while((components_initialized < gmm->n_components) && !heap_empty(gmm->heap))
  {
    float log_l;
    uint64_t comp_i;
    heap_remove(gmm->heap, &comp_i, &log_l);
    assert(comp_i < components_initialized);
    range_t range = sample_ranges[comp_i];

    gmm_fit_comp_t *parent = &gmm->components[comp_i];
    gmm_fit_comp_t children[2];
    int dim = gmm_fit_split_component(parent, children, samples + range.offset, range.length, log_total_weight, prior_radius);
    if(dim == -1) continue;
    gmm_fit_split_component_em(samples + range.offset, range.length, children, prior_radius);
#if 0
    const float *v = &parent->E_T[GUIDE_DIM * dim]; // the eigenvector we use as split axis
    uint32_t split = gmm_fit_partition_samples(samples + range.offset, range.length, parent->mean, v);
#endif

    {
      double ll_sum_parent = 0;
      double ll_sum_children = 0;
      double ll_norm = 0;
      double w2sum = 0;

      for(uint32_t sample_i = 0; sample_i < range.length; sample_i++)
      {
        em_sample_t *sample = &samples[range.offset + sample_i];
        double eval[2];
        for(int k = 0; k < 2; k++) eval[k] = comp_eval_with_weight(&children[k], sample->x);
        double eval_sum = eval[0] + eval[1];
        if(eval_sum > 0)
        {
          ll_sum_children += log(eval_sum / (children[0].weight + children[1].weight)) * sample->weight;
          ll_sum_parent += log(comp_eval(parent, sample->x)) * sample->weight;
          ll_norm += sample->weight;
          w2sum += sample->weight * sample->weight;
        }
      }

      double ess = ll_norm * ll_norm / w2sum;
      double ll_parent = (ll_sum_parent / ll_norm);
      double ll_children = (ll_sum_children / ll_norm);
      double k_parent = (GUIDE_DIM * (GUIDE_DIM+1)) / 2 + GUIDE_DIM;
      double k_children = 2 * k_parent + 1;
      //double bic_parent = log(ess) * k_parent - 2. * ll_parent * ess;
      //double bic_children = log(ess) * k_children - 2. * ll_children * ess;
      //if(bic_parent < bic_children) continue;
      double aic_parent = 2. * k_parent - 2 * ess * ll_parent;
      double aic_children = 2. * k_children - 2 * ess * ll_children;
      if(aic_parent < aic_children) continue;
    }

    uint32_t split = gmm_fit_partition_samples_em(samples + range.offset, range.length, children, prior_radius);

    if(split == 0 || range.length - split == 0) continue;

    uint32_t child_idx[] = { comp_i, components_initialized };
    sample_ranges[child_idx[0]] = (range_t) { range.offset, split };
    sample_ranges[child_idx[1]] = (range_t) { range.offset + split, range.length - split };
    for(int k = 0; k < 2; k++)
    {
      gmm->components[child_idx[k]] = children[k];
      float log_l_child = comp_log_likelihood(&children[k], log_total_weight);
      heap_insert(gmm->heap, child_idx[k], -log_l_child);
    }
    components_initialized++;
  }
  gmm_fit_resize(gmm, components_initialized);
}

static double gmm_fit_em_iter(gmm_fit_t *gmm, em_sample_t *samples, uint32_t n_samples, float prior_radius)
{
  for(uint32_t comp_i = 0; comp_i < gmm->n_components; comp_i++)
    gmm->accums[comp_i] = (gmm_comp_accum_t) {0};

  double *eval_buffer = gmm->scratch;
  double log_likelihood_sum = 0;
  double log_likelihood_norm = 0;

  double weight_sum = 0;
  for(uint32_t comp_i = 0; comp_i < gmm->n_components; comp_i++)
    weight_sum += gmm->components[comp_i].weight;
  for(uint32_t i = 0; i < n_samples; i++)
  {
    em_sample_t *sample = &samples[i];
    for(uint32_t comp_i = 0; comp_i < gmm->n_components; comp_i++)
      eval_buffer[comp_i] = comp_eval_with_weight(&gmm->components[comp_i], sample->x);
    double eval_sum = 0;
    for(uint32_t comp_i = 0; comp_i < gmm->n_components; comp_i++)
      eval_sum += eval_buffer[comp_i];
    if(eval_sum == 0) continue;
    double pdf = 0;
    for(uint32_t comp_i = 0; comp_i < gmm->n_components; comp_i++)
    {
      float weight = (eval_buffer[comp_i] / eval_sum) * sample->weight;
      accum_sample(&gmm->accums[comp_i], weight, sample->x);
      pdf += eval_buffer[comp_i] / weight_sum;
    }
    log_likelihood_sum += log(pdf) * sample->weight;
    log_likelihood_norm += sample->weight;
  }
  for(uint32_t comp_i = 0; comp_i < gmm->n_components; comp_i++)
    comp_from_accum(&gmm->components[comp_i], &gmm->accums[comp_i], prior_radius);
  return log_likelihood_sum / log_likelihood_norm;
}

static int gmm_fit_split_component(const gmm_fit_comp_t *parent, gmm_fit_comp_t children[2], em_sample_t *samples, uint32_t n_samples, float log_total_weight, float prior_radius)
{ // compute covariances for splits along every eigenvector in single pass over samples
  // then evaluate their likelihood and return the best one
  gmm_comp_accum_t accums[GUIDE_DIM * 2] = {};
  for(uint32_t sample_i = 0; sample_i < n_samples; sample_i++)
  {
    float x_centered[GUIDE_DIM];
    for(int k = 0; k < GUIDE_DIM; k++) x_centered[k] = samples[sample_i].x[k] - parent->mean[k];
    float d[GUIDE_DIM];
    mat_mulv(parent->E_T, x_centered, GUIDE_DIM, GUIDE_DIM, d);
    for(int k = 0; k < GUIDE_DIM; k++)
    {
      gmm_comp_accum_t *accum = &accums[2 * k + (d[k] > 0)];
      accum_sample(accum, samples[sample_i].weight, samples[sample_i].x);
    }
  }
  gmm_fit_comp_t comps[GUIDE_DIM * 2];
  for(uint32_t i = 0; i < GUIDE_DIM * 2; i++)
  {
#if 0
#if 1
    accum_from_comp_with_weight(&comps[i].prior, parent, 1.f); // use parent as prior
#else
    comps[i].prior = parent->prior;
    for(int k = 0; k < GUIDE_DIM; k++)
      comps[i].prior.mean[k] = accums[i].mean[k];
#endif
#endif
    comp_from_accum(&comps[i], &accums[i], prior_radius);
  }
#if 1
  double max_likelihood = -FLT_MAX;
#else
  double parent_ll = comp_log_likelihood(parent, log_total_weight);
  double max_likelihood = parent_ll;
#endif
  int axis = -1;
  for(int d = 0; d < GUIDE_DIM; d++)
  {
    if(comps[d * 2].weight == 0 || comps[d * 2 + 1].weight == 0) continue;
    float log_l_combined = 0;
    for(int k = 0; k < 2; k++)
    {
      gmm_fit_comp_t *comp = &comps[d * 2 + k];
      log_l_combined += comp_log_likelihood(comp, log_total_weight);
    }
    if(log_l_combined > max_likelihood)
    {
      max_likelihood = log_l_combined;
      axis = d;
    }
  }
  if(axis != -1) for(int k = 0; k < 2; k++)
    children[k] = comps[axis * 2 + k];
  return axis;
}

static void gmm_fit_split_component_em(em_sample_t *samples, uint32_t length, gmm_fit_comp_t children[2], float prior_radius)
{
  gmm_comp_accum_t accums[2] = {};
  double prev_log_likelihood = -FLT_MAX;
  for(int em_iter = 0; em_iter < EM_ITER; em_iter++)
  {
    double log_likelihood_sum = 0;
    double log_likelihood_norm = 0;
    for(int k = 0; k < 2; k++) accums[k] = (gmm_comp_accum_t) {0};
    for(uint32_t sample_i = 0; sample_i < length; sample_i++)
    {
      const em_sample_t *sample = &samples[sample_i];
      double eval[2];
      for(int k = 0; k < 2; k++) eval[k] = comp_eval_with_weight(&children[k], sample->x);
      double eval_sum = 0;
      for(int k = 0; k < 2; k++) eval_sum += eval[k];
      if(eval_sum == 0) continue;
      log_likelihood_norm += sample->weight;
      log_likelihood_sum += log(eval_sum / (children[0].weight + children[1].weight)) * sample->weight;
      for(int k = 0; k < 2; k++)
      {
        float weight = (eval[k] / eval_sum) * sample->weight;
        accum_sample(&accums[k], weight, sample->x);
      }
    }
    for(int k = 0; k < 2; k++)
      comp_from_accum(&children[k], &accums[k], prior_radius);
    double log_likelihood = log_likelihood_sum / log_likelihood_norm;
    double change = fabs(log_likelihood - prev_log_likelihood) / fabs(prev_log_likelihood);
    if(change < EM_STOP_TOL) break;
    prev_log_likelihood = log_likelihood;
  }
}

static uint32_t gmm_fit_partition_samples_em(em_sample_t *samples, uint32_t length, gmm_fit_comp_t children[2], float prior_radius)
{
  gmm_comp_accum_t accums[2] = {};
  for(int k = 0; k < 2; k++) accums[k] = (gmm_comp_accum_t) {0};
  uint32_t split = 0;
  for(uint32_t i = 0; i < length; i++)
  {
    double eval[2];
    for(int k = 0; k < 2; k++) eval[k] = comp_eval_with_weight(&children[k], samples[i].x);
    int maxcomp = (eval[0] > eval[1]) ? 0 : 1;
    accum_sample(&accums[maxcomp], samples[i].weight, samples[i].x);
    if(maxcomp == 0)
    {
      em_sample_t tmp = samples[i];
      samples[i] = samples[split];
      samples[split] = tmp;
      split++;
    }
  }
  for(int k = 0; k < 2; k++)
    comp_from_accum(&children[k], &accums[k], prior_radius);
  return split;
}

#if 0
static uint32_t gmm_fit_partition_samples(em_sample_t *samples, uint32_t length, const float center[GUIDE_DIM], const float axis[GUIDE_DIM])
{
  uint32_t split = 0;
  for(uint32_t i = 0; i < length; i++)
  {
    float pos = 0;
    float *x = samples[i].x;
    for(int k = 0; k < GUIDE_DIM; k++) pos += (x[k] - center[k]) * axis[k];
    if(pos <= 0)
    {
      em_sample_t tmp = samples[i];
      samples[i] = samples[split];
      samples[split] = tmp;
      split++;
    }
  }
  return split;
}
#endif

static void gmm3_fit_init_components(gmm3_fit_t *gmm, em_sample_t *samples, uint32_t n_samples, float prior_radius)
{ // pca part init
  typedef struct range_t {
    uint32_t offset;
    uint32_t length;
  } range_t;
  assert(sizeof(range_t) <= scratch_item_size);
  range_t *sample_ranges = gmm->scratch;
  heap_clear(gmm->heap);

  float log_total_weight;
  { // we need this to compute log likelihoods later
    double total_weight = 0;
    for(uint32_t sample_i = 0; sample_i < n_samples; sample_i++)
      total_weight += samples[sample_i].weight;
    log_total_weight = log(total_weight);
  }

  uint32_t components_initialized = 1;
  {
    gmm3_comp_accum_t accum = {};
    for(uint32_t i = 0; i < n_samples; i++)
      accum3_sample(&accum, samples[i].weight, samples[i].x + V2OFFSET);
    comp3_from_accum(&gmm->components[0], &accum, prior_radius);
    sample_ranges[0] = (range_t) { .offset = 0, .length = n_samples };
    float log_l = comp3_log_likelihood(&gmm->components[0], log_total_weight);
    heap_insert(gmm->heap, 0, -log_l);
  }
  while((components_initialized < gmm->n_components) && !heap_empty(gmm->heap))
  {
    float log_l;
    uint64_t comp_i;
    heap_remove(gmm->heap, &comp_i, &log_l);
    assert(comp_i < components_initialized);
    range_t range = sample_ranges[comp_i];

    gmm3_fit_comp_t *parent = &gmm->components[comp_i];
    gmm3_fit_comp_t children[2];
    int dim = gmm3_fit_split_component(parent, children, samples + range.offset, range.length, log_total_weight, prior_radius);
    if(dim == -1) continue;
    gmm3_fit_split_component_em(samples + range.offset, range.length, children, prior_radius);

    {
      double ll_sum_parent = 0;
      double ll_sum_children = 0;
      double ll_norm = 0;
      double w2sum = 0;

      for(uint32_t sample_i = 0; sample_i < range.length; sample_i++)
      {
        em_sample_t *sample = &samples[range.offset + sample_i];
        double eval[2];
        for(int k = 0; k < 2; k++) eval[k] = comp3_eval_with_weight(&children[k], sample->x);
        double eval_sum = eval[0] + eval[1];
        if(eval_sum > 0)
        {
          ll_sum_children += log(eval_sum / (children[0].weight + children[1].weight)) * sample->weight;
          ll_sum_parent += log(comp3_eval(parent, sample->x)) * sample->weight;
          ll_norm += sample->weight;
          w2sum += sample->weight * sample->weight;
        }
      }

      double ess = ll_norm * ll_norm / w2sum;
      double ll_parent = (ll_sum_parent / ll_norm);
      double ll_children = (ll_sum_children / ll_norm);
      double k_parent = (3 * (3+1)) / 2 + 3;
      double k_children = 2 * k_parent + 1;
      //double bic_parent = log(ess) * k_parent - 2. * ll_parent * ess;
      //double bic_children = log(ess) * k_children - 2. * ll_children * ess;
      //if(bic_parent < bic_children) continue;
      double aic_parent = 2. * k_parent - 2 * ess * ll_parent;
      double aic_children = 2. * k_children - 2 * ess * ll_children;
      if(aic_parent < aic_children) continue;
    }

    uint32_t split = gmm3_fit_partition_samples_em(samples + range.offset, range.length, children, prior_radius);

    if(split == 0 || range.length - split == 0) continue;

    uint32_t child_idx[] = { comp_i, components_initialized };
    sample_ranges[child_idx[0]] = (range_t) { range.offset, split };
    sample_ranges[child_idx[1]] = (range_t) { range.offset + split, range.length - split };
    for(int k = 0; k < 2; k++)
    {
      gmm->components[child_idx[k]] = children[k];
      float log_l_child = comp3_log_likelihood(&children[k], log_total_weight);
      heap_insert(gmm->heap, child_idx[k], -log_l_child);
    }
    components_initialized++;
  }
  gmm3_fit_resize(gmm, components_initialized);
}

static double gmm3_fit_em_iter(gmm3_fit_t *gmm, em_sample_t *samples, uint32_t n_samples, float prior_radius)
{
  for(uint32_t comp_i = 0; comp_i < gmm->n_components; comp_i++)
    gmm->accums[comp_i] = (gmm3_comp_accum_t) {0};

  double *eval_buffer = gmm->scratch;
  double log_likelihood_sum = 0;
  double log_likelihood_norm = 0;

  double weight_sum = 0;
  for(uint32_t comp_i = 0; comp_i < gmm->n_components; comp_i++)
    weight_sum += gmm->components[comp_i].weight;
  for(uint32_t i = 0; i < n_samples; i++)
  {
    const em_sample_t *sample = &samples[i];
    const float *x = sample->x + V2OFFSET;
    for(uint32_t comp_i = 0; comp_i < gmm->n_components; comp_i++)
      eval_buffer[comp_i] = comp3_eval_with_weight(&gmm->components[comp_i], x);
    double eval_sum = 0;
    for(uint32_t comp_i = 0; comp_i < gmm->n_components; comp_i++)
      eval_sum += eval_buffer[comp_i];
    if(eval_sum == 0) continue;
    double pdf = 0;
    for(uint32_t comp_i = 0; comp_i < gmm->n_components; comp_i++)
    {
      float weight = (eval_buffer[comp_i] / eval_sum) * sample->weight;
      accum3_sample(&gmm->accums[comp_i], weight, x);
      pdf += eval_buffer[comp_i] / weight_sum;
    }
    log_likelihood_sum += log(pdf) * sample->weight;
    log_likelihood_norm += sample->weight;
  }
  for(uint32_t comp_i = 0; comp_i < gmm->n_components; comp_i++)
    comp3_from_accum(&gmm->components[comp_i], &gmm->accums[comp_i], prior_radius);

  return log_likelihood_sum / log_likelihood_norm;
}

static int gmm3_fit_split_component(const gmm3_fit_comp_t *parent, gmm3_fit_comp_t children[2], em_sample_t *samples, uint32_t n_samples, float log_total_weight, float prior_radius)
{ // compute covariances for splits along every eigenvector in single pass over samples
  // then evaluate their likelihood and return the best one
  const int dim = 3;
  gmm3_comp_accum_t accums[3 * 2] = {};
  for(uint32_t sample_i = 0; sample_i < n_samples; sample_i++)
  {
    float x_centered[dim];
    for(int k = 0; k < dim; k++)
      x_centered[k] = samples[sample_i].x[V2OFFSET + k] - parent->mean[k];
    float d[dim];
    mat_mulv(parent->E_T, x_centered, dim, dim, d);
    for(int k = 0; k < dim; k++)
    {
      gmm3_comp_accum_t *accum = &accums[2 * k + (d[k] > 0)];
      accum3_sample(accum, samples[sample_i].weight, samples[sample_i].x + V2OFFSET);
    }
  }
  gmm3_fit_comp_t comps[dim * 2];
  for(uint32_t i = 0; i < dim * 2; i++)
    comp3_from_accum(&comps[i], &accums[i], prior_radius);

  double max_likelihood = -FLT_MAX;
  int axis = -1;
  for(int d = 0; d < dim; d++)
  {
    if(comps[d * 2].weight == 0 || comps[d * 2 + 1].weight == 0) continue;
    float log_l_combined = 0;
    for(int k = 0; k < 2; k++)
    {
      gmm3_fit_comp_t *comp = &comps[d * 2 + k];
      log_l_combined += comp3_log_likelihood(comp, log_total_weight);
    }
    if(log_l_combined > max_likelihood)
    {
      max_likelihood = log_l_combined;
      axis = d;
    }
  }
  if(axis != -1) for(int k = 0; k < 2; k++)
    children[k] = comps[axis * 2 + k];
  return axis;
}

static void gmm3_fit_split_component_em(em_sample_t *samples, uint32_t length, gmm3_fit_comp_t children[2], float prior_radius)
{
  gmm3_comp_accum_t accums[2] = {};
  double prev_log_likelihood = -FLT_MAX;
  for(int em_iter = 0; em_iter < EM_ITER; em_iter++)
  {
    double log_likelihood_sum = 0;
    double log_likelihood_norm = 0;
    for(int k = 0; k < 2; k++) accums[k] = (gmm3_comp_accum_t) {0};
    for(uint32_t sample_i = 0; sample_i < length; sample_i++)
    {
      const em_sample_t *sample = &samples[sample_i];
      double eval[2];
      for(int k = 0; k < 2; k++) eval[k] = comp3_eval_with_weight(&children[k], sample->x + V2OFFSET);
      double eval_sum = 0;
      for(int k = 0; k < 2; k++) eval_sum += eval[k];
      if(eval_sum == 0) continue;
      log_likelihood_norm += sample->weight;
      log_likelihood_sum += log(eval_sum / (children[0].weight + children[1].weight)) * sample->weight;
      for(int k = 0; k < 2; k++)
      {
        float weight = (eval[k] / eval_sum) * sample->weight;
        accum3_sample(&accums[k], weight, sample->x + V2OFFSET);
      }
    }
    for(int k = 0; k < 2; k++)
      comp3_from_accum(&children[k], &accums[k], prior_radius);
    double log_likelihood = log_likelihood_sum / log_likelihood_norm;
    double change = fabs(log_likelihood - prev_log_likelihood) / fabs(prev_log_likelihood);
    if(change < EM_STOP_TOL) break;
    prev_log_likelihood = log_likelihood;
  }
}

static uint32_t gmm3_fit_partition_samples_em(em_sample_t *samples, uint32_t length, gmm3_fit_comp_t children[2], float prior_radius)
{
  gmm3_comp_accum_t accums[2] = {};
  for(int k = 0; k < 2; k++) accums[k] = (gmm3_comp_accum_t) {0};
  uint32_t split = 0;
  for(uint32_t i = 0; i < length; i++)
  {
    double eval[2];
    for(int k = 0; k < 2; k++) eval[k] = comp3_eval_with_weight(&children[k], samples[i].x);
    int maxcomp = (eval[0] > eval[1]) ? 0 : 1;
    accum3_sample(&accums[maxcomp], samples[i].weight, samples[i].x + V2OFFSET);
    if(maxcomp == 0)
    {
      em_sample_t tmp = samples[i];
      samples[i] = samples[split];
      samples[split] = tmp;
      split++;
    }
  }
  for(int k = 0; k < 2; k++)
    comp3_from_accum(&children[k], &accums[k], prior_radius);
  return split;
}

static void accum_sample(gmm_comp_accum_t *accum, const float weight, const float *x)
{
  accum->weight += weight;
  accum->weight2 += weight * weight;
  double update_weight = weight / accum->weight;
  if(!(update_weight < FLT_MAX)) return;
  double delta[GUIDE_DIM];
  for(int k = 0; k < GUIDE_DIM; k++)
  {
    delta[k] = x[k] - accum->mean[k];
    accum->mean[k] += update_weight * delta[k];
  }
  uint32_t offset = 0;
  for(int i = 0; i < GUIDE_DIM; i++)
  {
    double delta_prime = x[i] - accum->mean[i];
    for(int j = i; j < GUIDE_DIM; j++)
    {
      accum->cov[offset++] += weight * delta[j] * delta_prime;
    }
  }
}

#if 0
static void accum_reweight(gmm_comp_accum_t *accum, float new_weight)
{ // does not adjust weight2!
  const int ACC_N = (GUIDE_DIM*(GUIDE_DIM+1))/2;
  double correction = new_weight / accum->weight;
  for(int k = 0; k < ACC_N; k++) accum->cov[k] *= correction;
  accum->weight = new_weight;
}
#endif

static void accum_merge(const gmm_comp_accum_t *a, const gmm_comp_accum_t *b, gmm_comp_accum_t *result)
{
  result->weight = a->weight + b->weight;
  result->weight2 = a->weight2 + b->weight2;

  double delta[GUIDE_DIM];
  double update_weight = b->weight / result->weight;
  for(int k = 0; k < GUIDE_DIM; k++)
  {
    delta[k] = b->mean[k] - a->mean[k];
    result->mean[k] = a->mean[k] + update_weight * delta[k];
  }
  uint32_t offset = 0;
  for(int i = 0; i < GUIDE_DIM; i++)
  {
    double delta_prime = b->mean[i] - result->mean[i];
    for(int j = i; j < GUIDE_DIM; j++)
    {
      result->cov[offset] = a->cov[offset] + b->cov[offset]
        + b->weight * delta[j] * delta_prime;
      offset++;
    }
  }
}

static void accum_get_cov(const gmm_comp_accum_t *accum, double *cov)
{
  uint32_t offset = 0;
  for(int i = 0; i < GUIDE_DIM; i++) for(int j = i; j < GUIDE_DIM; j++)
    cov[i * GUIDE_DIM + j] = accum->cov[offset++] / accum->weight;
  for(int i = 0; i < GUIDE_DIM; i++) for(int j = 0; j < i; j++)
    cov[i * GUIDE_DIM + j] = cov[j * GUIDE_DIM + i];
}

static void accum_from_comp(gmm_comp_accum_t *accum, const gmm_fit_comp_t *comp)
{
  accum_from_comp_with_weight(accum, comp, comp->weight);
}

static void accum_from_comp_with_weight(gmm_comp_accum_t *accum, const gmm_fit_comp_t *comp, float weight)
{ // does not fill weight2!
  accum->weight = weight;
  for(int k = 0; k < GUIDE_DIM; k++) accum->mean[k] = comp->mean[k];

  float cov[GUIDE_DIM * GUIDE_DIM], E[GUIDE_DIM * GUIDE_DIM], w[GUIDE_DIM];
  for(int k = 0; k < GUIDE_DIM; k++) w[k] = 1.f/comp->w_inv[k];
  mat_transpose(comp->E_T, GUIDE_DIM, GUIDE_DIM, E);
  mat_compute_cov(GUIDE_DIM, E, w, cov);

  uint32_t offset = 0;
  for(int i = 0; i < GUIDE_DIM; i++) for(int j = i; j < GUIDE_DIM; j++)
    accum->cov[offset++] = cov[i * GUIDE_DIM + j] * accum->weight;
}

static double accum_ess(gmm_comp_accum_t *accum)
{
  double ess = accum->weight * accum->weight / accum->weight2;
  if(!(ess >= 0 && ess < FLT_MAX)) ess = 0;
  return ess;
}

static void comp_set_cov(gmm_fit_comp_t *comp, double *cov)
{
  double E[GUIDE_DIM * GUIDE_DIM], w[GUIDE_DIM];
  int svd_err = svd_d(cov, GUIDE_DIM, w, E);
  if(svd_err)
  {
    comp->weight = 0;
    return;
  }
  regularize_d(GUIDE_DIM, w);
  for(int k = 0; k < GUIDE_DIM; k++) comp->w_inv[k] = 1./w[k];
  float E_f[GUIDE_DIM * GUIDE_DIM];
  for(int k = 0; k < GUIDE_DIM * GUIDE_DIM; k++) E_f[k] = E[k];
  mat_transpose(E_f, GUIDE_DIM, GUIDE_DIM, comp->E_T);
}

static void comp_from_accum(gmm_fit_comp_t *comp, gmm_comp_accum_t *accum, float prior_radius)
{
  float ess = accum_ess(accum);
  if(!(accum->weight > 0.))
  {
    comp->weight = 0;
    return;
  }

#if 1
  comp->weight = accum->weight;
  for(int k = 0; k < GUIDE_DIM; k++) comp->mean[k] = accum->mean[k];

  double cov[GUIDE_DIM * GUIDE_DIM];
  accum_get_cov(accum, cov);
  restrict_cov(cov, prior_radius, ess);
  comp_set_cov(comp, cov);
#else
  float t_prior = FIXED_PRIOR_N0 / (ess + FIXED_PRIOR_N0);
  t_prior = CLAMP(t_prior, 0.f, 1.f);
  float t_data = CLAMP(1.f - t_prior, 0.f, 1.f);
  gmm_comp_accum_t prior = comp->prior;
  float weight = accum->weight;
  accum_reweight(&prior, t_prior);
  accum_reweight(accum, t_data);
  gmm_comp_accum_t result;
  accum_merge(&prior, accum, &result);
  comp_from_accum_no_prior(comp, &result);
  if(comp->weight > 0) comp->weight = weight;
#endif
}

static void comp_from_accum_no_prior(gmm_fit_comp_t *comp, gmm_comp_accum_t *accum)
{
  comp->weight = accum->weight;
  for(int k = 0; k < GUIDE_DIM; k++) comp->mean[k] = accum->mean[k];

  double cov[GUIDE_DIM * GUIDE_DIM];
  accum_get_cov(accum, cov);

  comp_set_cov(comp, cov);
}

static double comp_eval(const gmm_fit_comp_t *comp, const float *x)
{
  float x_centered[GUIDE_DIM];
  for(int k = 0; k < GUIDE_DIM; k++) x_centered[k] = x[k] - comp->mean[k];
  float d[GUIDE_DIM];
  mat_mulv(comp->E_T, x_centered, GUIDE_DIM, GUIDE_DIM, d);
  double pdf = 1;
  for(int k = 0; k < GUIDE_DIM; k++) pdf *= fakegaussian_pdf(d[k] * comp->w_inv[k]) * comp->w_inv[k];
  return pdf;
}

static double comp_eval_with_weight(const gmm_fit_comp_t *comp, const float *x)
{
  return comp->weight * comp_eval(comp, x);
}

static double comp_log_likelihood(const gmm_fit_comp_t *comp, float log_total_weight)
{
  double log_l = comp->weight * (
      log(comp->weight) - log_total_weight
      - 0.5 * GUIDE_DIM * logf(2. * M_PI)
      - 0.5 * comp_logdet(comp)
      - 0.5 * GUIDE_DIM
      );
  return log_l;
}

static double comp_logdet(const gmm_fit_comp_t *comp)
{
  double logdet = 0;
  for(int i = 0; i < GUIDE_DIM; i++) logdet -= log(comp->w_inv[i]);
  return 2. * logdet;
}

// computes svd of n x n matrix in double precision (internally):
// v * diag(w)^2 * v^T = a
// in case this needs to be faster:
// * try single precision only
// * specialized svd for symmetric matrices (tridiagonal form)
static int svd_d(const double *a, int n, double *w, double *v)
{
  for(int i = 0; i < n * n; i++) v[i] = a[i];
  int svd_err = svd_symm(n, v, w);
  if(svd_err) return svd_err;
  for(int i = 0; i < n; i++) w[i] = sqrt(MAX(0., w[i]));
  return 0;
}

static void regularize_d(int n, double *w)
{
  for(int k = 0; k < n; k++) w[k] = MAX(w[k], COV_MIN_REG);
}

#if 0
static void comp_copy_to_prior(gmm_fit_comp_t *comp)
{
  accum_from_comp_with_weight(&comp->prior, comp, 1.f);
}
#endif

static void restrict_cov(double *cov, float prior_radius, float ess)
{
  float REG_N0 = GUIDE_DIM;
  float t_reg = REG_N0 / (ess - 1.f + REG_N0);
  float t_cov = 1.f - t_reg;
  t_cov = CLAMP(t_cov, 0.f, 1.f);

  float t_prior = (ess - 1.f) / 3.f;
  t_prior = CLAMP(t_prior, 0.f, 1.f);

  const int ISO_DIM = GUIDE_DIM / 3;
  for(int i = 0; i < ISO_DIM; i++) for(int j = i; j < ISO_DIM; j++)
  {
    double iso = 0;
    for(int k = 0; k < 3; k++) iso += cov[(3 * i + k) * GUIDE_DIM + (3 * j + k)];
    iso /= 3.;
    float r2 = (i == j) ? prior_radius * prior_radius : 0.f;
    iso = iso * t_prior + r2 * (1.f - t_prior);
    for(int k = 0; k < 3; k++)
    {
      double *c = &cov[(3 * i + k) * GUIDE_DIM + (3 * j + k)];
      *c = *c * t_cov + iso * t_reg;
    }
    for(int ki = 0; ki < 3; ki++) for(int kj = ki + 1; kj < 3; kj++)
      cov[(3 * i + ki) * GUIDE_DIM + (3 * j + kj)] *= t_cov;
    for(int ki = 0; ki < 3; ki++) for(int kj = 0; kj < ki; kj++)
      cov[(3 * i + ki) * GUIDE_DIM + (3 * j + kj)] *= t_cov;
  }

  // mirror the whole matrix
  for(int i = 0; i < GUIDE_DIM; i++) for(int j = 0; j < i; j++)
    cov[i * GUIDE_DIM + j] = cov[j * GUIDE_DIM + i];
}

//--- camera mixture -----------------------------------------------------------

void gmm3_fit_init(gmm3_fit_t *gmm)
{
  const uint32_t n_comp_init = 8;
  *gmm = (gmm3_fit_t) {
    .n_components = 0,
    .components_allocated = n_comp_init,
    .components = malloc(n_comp_init * sizeof(gmm3_fit_comp_t)),
    .accums = malloc(n_comp_init * sizeof(gmm3_comp_accum_t)),
    .scratch = malloc(n_comp_init * scratch_item_size),
    .heap = heap_init(n_comp_init),
  };
}

void gmm3_fit_cleanup(gmm3_fit_t *gmm)
{
  free(gmm->components);
  free(gmm->accums);
  free(gmm->scratch);
  heap_cleanup(gmm->heap);
}

void gmm3_fit_resize(gmm3_fit_t *gmm, uint32_t n_components)
{
  gmm->n_components = n_components;
  if(gmm->n_components <= gmm->components_allocated) return;
  do gmm->components_allocated *= 2;
  while(gmm->components_allocated < gmm->n_components);
  gmm->components = realloc(gmm->components, sizeof(gmm3_fit_comp_t) * gmm->components_allocated);
  gmm->accums = realloc(gmm->accums, sizeof(gmm3_comp_accum_t) * gmm->components_allocated);
  gmm->scratch = realloc(gmm->scratch, scratch_item_size * gmm->components_allocated);
  heap_cleanup(gmm->heap);
  gmm->heap = heap_init(gmm->components_allocated);
}

void gmm3_copy_from_full_gmm(gmm3_fit_t *gmm, const gmm_fit_t *full_gmm)
{
  gmm3_fit_resize(gmm, full_gmm->n_components);

  for(uint32_t comp_i = 0; comp_i < gmm->n_components; comp_i++)
  {
    gmm3_fit_comp_t *new_comp = &gmm->components[comp_i];
    gmm_fit_comp_t *old_comp = &full_gmm->components[comp_i];

    new_comp->weight = old_comp->weight;
    for(int k = 0; k < 3; k++) new_comp->mean[k] = old_comp->mean[V2OFFSET + k];

    double full_cov[GUIDE_DIM * GUIDE_DIM], E[GUIDE_DIM * GUIDE_DIM], w[GUIDE_DIM];
    for(int k = 0; k < GUIDE_DIM; k++) w[k] = 1./old_comp->w_inv[k];
    float E_f[GUIDE_DIM * GUIDE_DIM];
    mat_transpose(old_comp->E_T, GUIDE_DIM, GUIDE_DIM, E_f);
    for(int k = 0; k < GUIDE_DIM * GUIDE_DIM; k++) E[k] = E_f[k];
    mat_compute_cov_d(GUIDE_DIM, E, w, full_cov);

    double cov_v2[3 * 3];
    mat_extract_d(full_cov, GUIDE_DIM, GUIDE_DIM, V2OFFSET, V2OFFSET + 3, V2OFFSET, V2OFFSET + 3, cov_v2);
    comp3_set_cov(new_comp, cov_v2);
  }
}

void gmm3_fit_samples(gmm3_fit_t *gmm, em_sample_t *samples, uint32_t n_samples, float prior_radius)
{
  if(gmm->n_components == 0) return;
  assert(n_samples > 0);
  gmm3_fit_init_components(gmm, samples, n_samples, prior_radius);
  if(gmm->n_components < 2) return;

  double prev_log_likelihood = -FLT_MAX;
  for(int i = 0; i < EM_ITER; i++)
  {
    double log_likelihood = gmm3_fit_em_iter(gmm, samples, n_samples, prior_radius);
    double change = fabs(log_likelihood - prev_log_likelihood) / fabs(prev_log_likelihood);
    if(change < EM_STOP_TOL) break;
    prev_log_likelihood = log_likelihood;
  }
}

void gmm3_fit_reduce(gmm3_fit_t *gmm, float threshold, double total_weight)
{
  float accumulated_cost = 0;
  while(accumulated_cost < threshold)
  {
    float min_cost = FLT_MAX;
    uint32_t i_min = 0, j_min = 0;
    const uint32_t N = gmm->n_components;
    for(uint32_t i = 0; i < N; i++)
    {
      gmm3_comp_accum_t accum_i;
      accum3_from_comp(&accum_i, &gmm->components[i]);
      float logdet_i = comp3_logdet(&gmm->components[i]);
      for(uint32_t j = i + 1; j < N; j++)
      {
        float logdet_j = comp3_logdet(&gmm->components[j]);
        gmm3_comp_accum_t accum_j, result;
        accum3_from_comp(&accum_j, &gmm->components[j]);
        accum3_merge(&accum_i, &accum_j, &result);
        gmm3_fit_comp_t merged_comp;
        comp3_from_accum_no_prior(&merged_comp, &result);
        float logdet_ij = comp3_logdet(&merged_comp);
        float cost = (merged_comp.weight / total_weight) * logdet_ij
          - (gmm->components[i].weight / total_weight) * logdet_i
          - (gmm->components[j].weight / total_weight) * logdet_j;
        if(cost < min_cost)
        {
          min_cost = cost;
          i_min = i;
          j_min = j;
        }
      }
    }
    if(accumulated_cost + min_cost < threshold)
    {
      gmm3_comp_accum_t accum_i, accum_j, result;
      accum3_from_comp(&accum_i, &gmm->components[i_min]);
      accum3_from_comp(&accum_j, &gmm->components[j_min]);
      accum3_merge(&accum_i, &accum_j, &result);
      gmm3_fit_comp_t merged_comp;
      comp3_from_accum_no_prior(&merged_comp, &result);
      gmm->components[i_min] = merged_comp;
      gmm->components[j_min] = gmm->components[gmm->n_components - 1];
      gmm->n_components--;
      accumulated_cost += min_cost;
    }
    else break;
  }
}

static void accum3_sample(gmm3_comp_accum_t *accum, const float weight, const float *x)
{
  accum->weight += weight;
  accum->weight2 += weight * weight;
  double update_weight = weight / accum->weight;
  if(!(update_weight < FLT_MAX)) return;
  double delta[3];
  for(int k = 0; k < 3; k++)
  {
    delta[k] = x[k] - accum->mean[k];
    accum->mean[k] += update_weight * delta[k];
  }
  uint32_t offset = 0;
  for(int i = 0; i < 3; i++)
  {
    double delta_prime = x[i] - accum->mean[i];
    for(int j = i; j < 3; j++)
    {
      accum->cov[offset++] += weight * delta[j] * delta_prime;
    }
  }
}

static double comp3_eval(const gmm3_fit_comp_t *comp, const float *x)
{
  float x_centered[3];
  for(int k = 0; k < 3; k++) x_centered[k] = x[k] - comp->mean[k];
  float d[3];
  mat_mulv(comp->E_T, x_centered, 3, 3, d);
  double pdf = 1;
  for(int k = 0; k < 3; k++) pdf *= fakegaussian_pdf(d[k] * comp->w_inv[k]) * comp->w_inv[k];
  return pdf;
}

static double comp3_eval_with_weight(const gmm3_fit_comp_t *comp, const float *x)
{
  return comp->weight * comp3_eval(comp, x);
}

static void comp3_set_cov(gmm3_fit_comp_t *comp, double *cov)
{
  double E[3 * 3], w[3];
  int svd_err = svd_d(cov, 3, w, E);
  if(svd_err)
  {
    comp->weight = 0;
    return;
  }
  regularize_d(3, w);
  for(int k = 0; k < 3; k++) comp->w_inv[k] = 1./w[k];
  float E_f[3 * 3];
  for(int k = 0; k < 3 * 3; k++) E_f[k] = E[k];
  mat_transpose(E_f, 3, 3, comp->E_T);
}

static void comp3_from_accum(gmm3_fit_comp_t *comp, gmm3_comp_accum_t *accum, float prior_radius)
{
  float ess = accum->weight * accum->weight / accum->weight2;
  if(!(ess > 0 && ess < FLT_MAX)) ess = 0;
  if(!(accum->weight > 0.))
  {
    comp->weight = 0;
    return;
  }

  comp->weight = accum->weight;
  for(int k = 0; k < 3; k++) comp->mean[k] = accum->mean[k];

  double cov[3 * 3];
  accum3_get_cov(accum, cov);

  const float N0 = 3.f;
  float t_prior = N0 / (ess - 1.f + N0);
  for(int i = 0; i < 3; i++) for(int j = 0; j < 3; j++)
  { // shrink towards isotropic prior
    cov[i * 3 + j] *= (1.f - t_prior);
    if(i == j) cov[i * 3 + j] += t_prior * (prior_radius * prior_radius);
  }

  comp3_set_cov(comp, cov);
}

static void accum3_get_cov(const gmm3_comp_accum_t *accum, double *cov)
{
  uint32_t offset = 0;
  for(int i = 0; i < 3; i++) for(int j = i; j < 3; j++)
    cov[i * 3 + j] = accum->cov[offset++] / accum->weight;
  for(int i = 0; i < 3; i++) for(int j = 0; j < i; j++)
    cov[i * 3 + j] = cov[j * 3 + i];
}

static void accum3_from_comp(gmm3_comp_accum_t *accum, const gmm3_fit_comp_t *comp)
{ // does not fill weight2!
  accum->weight = comp->weight;
  for(int k = 0; k < 3; k++) accum->mean[k] = comp->mean[k];

  float cov[3 * 3], E[3 * 3], w[3];
  for(int k = 0; k < 3; k++) w[k] = 1.f/comp->w_inv[k];
  mat_transpose(comp->E_T, 3, 3, E);
  mat_compute_cov(3, E, w, cov);

  uint32_t offset = 0;
  for(int i = 0; i < 3; i++) for(int j = i; j < 3; j++)
    accum->cov[offset++] = cov[i * 3 + j] * accum->weight;
}

static double comp3_log_likelihood(const gmm3_fit_comp_t *comp, float log_total_weight)
{
  int dim = 3;
  double log_l = comp->weight * (
      log(comp->weight) - log_total_weight
      - 0.5 * dim * logf(2. * M_PI)
      - 0.5 * comp3_logdet(comp)
      - 0.5 * dim
      );
  return log_l;
}

static double comp3_logdet(const gmm3_fit_comp_t *comp)
{
  double logdet = 0;
  for(int i = 0; i < 3; i++) logdet -= log(comp->w_inv[i]);
  return 2. * logdet;
}

static void accum3_merge(const gmm3_comp_accum_t *a, const gmm3_comp_accum_t *b, gmm3_comp_accum_t *result)
{
  result->weight = a->weight + b->weight;
  result->weight2 = a->weight2 + b->weight2;

  double delta[3];
  double update_weight = b->weight / result->weight;
  for(int k = 0; k < 3; k++)
  {
    delta[k] = b->mean[k] - a->mean[k];
    result->mean[k] = a->mean[k] + update_weight * delta[k];
  }
  uint32_t offset = 0;
  for(int i = 0; i < 3; i++)
  {
    double delta_prime = b->mean[i] - result->mean[i];
    for(int j = i; j < 3; j++)
    {
      result->cov[offset] = a->cov[offset] + b->cov[offset]
        + b->weight * delta[j] * delta_prime;
      offset++;
    }
  }
}

static void comp3_from_accum_no_prior(gmm3_fit_comp_t *comp, gmm3_comp_accum_t *accum)
{
  comp->weight = accum->weight;
  for(int k = 0; k < 3; k++) comp->mean[k] = accum->mean[k];

  double cov[3 * 3];
  accum3_get_cov(accum, cov);

  comp3_set_cov(comp, cov);
}
