#include "config.h"
#include "guided.h"
#include "lobe.h"
#include "lobe_list.h"
#include "matrix.h"
#include "model.h"
#include "pixel_est.h"
#include "reservoir.h"
#include "stats.h"
#include "util.h"

#include "accel.h"
#include "corona_common.h"
#include "heap.h"
#include "matrix3.h"
#include "lights.h"
#include "pathspace.h"
#include "pathspace/nee.h"
#include "points.h"
#include "sampler_common.h"
#include "shader.h"
#include "spectrum.h"
#include "threads.h"
#include "view.h"

#include <stdlib.h>

typedef struct guided_tls_t
{
  lobe_list_t lobe_list_pt;
  lobe_list_t lobe_list_lt;
} guided_tls_t;

typedef struct guided_t
{
  model_t model;

  guided_tls_t *tls;

#if REL_CONTRIBUTION
  float *pixel_val;
#endif

  // for synchronization of thread pool
  uint64_t job;
  uint64_t num_jobs;

  guided_stats_t *stats;
} guided_t;

static lobe_range_t guided_lobe_lookup(guided_t *c, const path_t *path, int vi);

static float guided_p_extend_bsdf(const path_t *path, int v);
//static int guided_raydiff_radius(const path_t *path, float step, float *radius);

static lobe_range_t lobe_list_conditional(lobe_list_t *lobe_list, const model_t *m, const float xpp[3], const float xp[3], lobe_range_t lobe_range, bool lt);

guided_t *guided_init()
{
  // zero-initialize whole struct (so we know which members are not initialized and free is safe)
  guided_t *c = calloc(1, sizeof(guided_t));

  c->stats = stats_init();

  model_init(&c->model, c->stats);

  c->tls = calloc(rt.num_threads, sizeof(*c->tls));
  for(int t = 0; t < rt.num_threads; t++)
  {
    lobe_list_init(&c->tls[t].lobe_list_pt, 64);
    lobe_list_init(&c->tls[t].lobe_list_lt, 64);
  }

#if REL_CONTRIBUTION
  c->pixel_val = malloc(sizeof(float) * view_width() * view_height());
  for(int i = 0; i < view_width() * view_height(); i++) c->pixel_val[i] = 1.f;
#endif

  return c;
}

void guided_cleanup(guided_t *c)
{
  model_cleanup(&c->model);

  for (int t = 0; t < rt.num_threads; t++)
  {
    lobe_list_cleanup(&c->tls[t].lobe_list_pt);
    lobe_list_cleanup(&c->tls[t].lobe_list_lt);
  }
  free(c->tls);

#if REL_CONTRIBUTION
  free(c->pixel_val);
#endif

  stats_cleanup(c->stats);

  free(c);
}

void guided_update(guided_t *c, pixel_est_t *px_est, int num_samples)
{
  // update pixel estimate
  if(num_samples > 0) {
#if REL_CONTRIBUTION
    pixel_est_estimate(px_est, num_samples, PIXEL_EST_N_MIN, c->pixel_val);

    int wd = view_width();
    int ht = view_height();
    for(int i = 0; i < wd * ht; i++) c->pixel_val[i] = MAX(c->pixel_val[i], PIXEL_EST_MIN);
#if WRITE_DEBUG_INFO
    int write_screenshot = (((unsigned int) num_samples) & ((unsigned int) num_samples - 1)) == 0;
    if(write_screenshot)
    {
      char filename[1024];
      snprintf(filename, sizeof(filename), "pixel_est_%05d", num_samples);
      screenshot_write(filename, c->pixel_val, 1, 0, 1, wd, ht, 1.f);
    }
#endif
#else
    assert(px_est == NULL);
#endif
  }

  model_update(&c->model);

  stats_collect(c->stats);
}

void guided_prepare_frame(guided_t *c)
{
  model_prepare_frame(&c->model);
}

static lobe_range_t guided_lobe_lookup(guided_t *c, const path_t *path, int vi)
{
  const bool lt = path->v[0].mode & s_emit;
  guided_tls_t *tls = &c->tls[common_get_threadid()];
  lobe_list_t *lobe_list = lt ? &tls->lobe_list_lt : &tls->lobe_list_pt;
  lobe_range_t *lobe_range = &lobe_list->ranges[vi];
  if((lobe_range->end - lobe_range->begin) >= 0) // did lookup already
    return *lobe_range;
  lobe_range->begin = lobe_list->end;

  if(lt && vi == 0) return (lobe_range_t) { .begin = 0, .end = 0 };

  if(vi == 0) assert(path->length >= 2); // need second vertex for lookup
  const float *xp = path->v[(vi == 0) ? 1 : vi].hit.x;
  uint32_t n_lobes;
  while(1)
  {
    const uint32_t n_max = lobe_list->allocated - lobe_range->begin;
    uint32_t *lobes = lobe_list->lobes + lobe_range->begin;
    if(vi == 0) n_lobes = model_fill_lobe_list_camera(&c->model, xp, lobes, n_max);
    else
    {
      const float *xpp = path->v[vi-1].hit.x;
      n_lobes = model_fill_lobe_list(&c->model, xpp, xp, lobes, n_max);
    }

    if(n_lobes <= n_max) break; // valid lookup
    // realloc and repeat lookup
    lobe_list_extend(lobe_list);
  }

  uint32_t *lobes = lobe_list->lobes + lobe_range->begin;
  stats_record_lobes_found(c->stats, n_lobes, vi);

  if(vi == 0)
  {
    float *lobe_cdf = lobe_list->cdf + lobe_range->begin;
    uint32_t n_lobes_new = 0;
    for(uint32_t i = 0; i < n_lobes; i++)
    {
      const uint32_t lobe_id = lobes[i];
      assert(lobe_id < c->model.cam_lobes.n);
      const cam_lobe_t *lobe = &c->model.cam_lobes.lobes[lobe_id];
      if(!util_point_in_aabb(lobe->aabb, xp)) continue;
      if(!lobe_type_is_supported(lobe->type, path, 0)) continue;

      lobes[n_lobes_new] = lobe_id;
      lobe_cdf[n_lobes_new] = c->model.cam_lobe_cdf[lobe_id];
      if(lobe_id > 0) lobe_cdf[n_lobes_new] -= c->model.cam_lobe_cdf[lobe_id - 1];
      n_lobes_new++;
    }
    n_lobes = n_lobes_new;
    lobe_range->end = lobe_range->begin + n_lobes;
  }
  else
  {
    // filter results based on modes
    uint32_t n_lobes_new = 0;
    for(uint32_t i = 0; i < n_lobes; i++)
    {
      const uint32_t lobe_id = lobes[i];
      assert(lobe_id < c->model.lobes.n);
      const lobe_t *lobe = &c->model.lobes.lobes[lobe_id];
      if(!util_point_in_aabb(lobe->aabb_v2, xp)) continue;
      if(!lobe_type_is_supported(lobe->type, path, vi)) continue;

      lobes[n_lobes_new] = lobe_id;
      n_lobes_new++;
    }
    n_lobes = n_lobes_new;

    stats_record_lobes_filtered(c->stats, n_lobes, vi);

    lobe_range->end = lobe_range->begin + n_lobes;
    const float *xpp = path->v[vi-1].hit.x;
    *lobe_range = lobe_list_conditional(lobe_list, &c->model, xpp, xp, *lobe_range, lt);
  }

  n_lobes = lobe_range->end - lobe_range->begin;
  stats_record_lobes_used(c->stats, n_lobes, vi);
  lobe_list->end = lobe_range->end;
  return *lobe_range;
}

static int guided_sample_pos_first(guided_t *c, path_t *path, const uint32_t lobe_index)
{
  assert(!(path->v[0].mode & s_emit));
  int v = path->length;

  // start from scratch at camera:
  // init essentials on new path and v[0]
  path->tangent_frame_scrambling = 0.1f + points_rand(rt.points, common_get_threadid())*(0.9f-0.1f);
  mf_t lambda_pdf;
  if(mf_all(mf_eq(path->lambda, mf_set1(0.0f))))
  { // if it's already set, don't sample a new one (for bdpt)
    // so do we stratify these now or not?
    float lf[MF_COUNT];
    float xi = pointsampler(path, s_dim_lambda);
    for(int l=0;l<MF_COUNT;l++)
      lf[l] = fmodf(xi + l/(float)MF_COUNT, 1.0f);
      //lf[l] = fmodf(pointsampler(path, s_dim_lambda) + l/(float)MF_COUNT, 1.0f);
    path->lambda = spectrum_sample_lambda(mf_loadu(lf), &lambda_pdf);
    path->time = view_sample_time(path, pointsampler(path, s_dim_time)); // unit pdf, /= 1.0f to throughput.
  }
  if(!mf_any(mf_gt(view_cam_sample(path), mf_set1(0.0f)))) return ERR_CAM_SAMPLE;

  path->v[0].mode = s_sensor;

  // sample start point always in global exterior medium:
  shader_exterior_medium(path);

  // we're actually creating two vertices at this point. the first is on the camera:
  path->length++;
  v++;
  memset(path->v+path->length, 0, sizeof(vertex_t));
  memset(path->e+path->length, 0, sizeof(edge_t));

  path->v[0].throughput = mf_set1(1.0f); // dummy
  path->v[1].throughput = mf_set1(1.0f); // dummy

  // TODO check if pdf zero only after projection!
  const float *xp = path->v[0].hit.x;

  const cam_lobe_t* sampled_lobe = &c->model.cam_lobes.lobes[lobe_index];
  float random[3];
  for(int k = 0; k < 2; k++) random[k] = points_rand(rt.points, common_get_threadid());
  if(sampled_lobe->type & t_vol_v3)
  {
    random[2] = points_rand(rt.points, common_get_threadid());
    cam_lobe_sample_pos_3d(sampled_lobe, random, path->v[v].hit.x);
    double pdf = cam_lobe_pdf_pos_3d_dw(sampled_lobe, xp, path->v[v].hit.x);
    if(!(pdf > 0. && pdf < FLT_MAX)) return ERR_PDF_ZERO_3D;
    // tentative mode for new vertex
    path->v[v].mode = s_volume;
  }
  else
  {
    cam_lobe_sample_pos_2d(sampled_lobe, random, xp, path->v[v].hit.x);
    double pdf = cam_lobe_pdf_pos_2d_dw(sampled_lobe, xp, path->v[v].hit.x);
    if(!(pdf > 0. && pdf < FLT_MAX)) return ERR_PDF_ZERO_2D;
  }
  return 0;
}

static int guided_sample_pos_next(guided_t *c, path_t *path, const uint32_t lobe_index)
{
  bool lt = path->v[0].mode & s_emit;
  const int v = path->length;

  // minimal init on next data points:
  memset(path->v+path->length, 0, sizeof(vertex_t));
  memset(path->e+path->length, 0, sizeof(edge_t));

  // sample from this model (has adjusted lobe weights, too)
  const lobe_t* sampled_lobe = &c->model.lobes.lobes[lobe_index];

  const float *xpp = path->v[v-2].hit.x;
  const float *xp = path->v[v-1].hit.x;
  // based on specularity of vertex, sample (constrained) bsdf or phase function instead
  const float p_bsdf = guided_p_extend_bsdf(path, path->length);
  const bool in_volume = (lt ? t_vol_v1 : t_vol_v3) & sampled_lobe->type;
  if(points_rand(rt.points, common_get_threadid()) < p_bsdf)
  {
    path->v[v].throughput = path->v[v-1].throughput;
    // sample the bsdf at vertex v-1. also inits e[v] with volume information.
    path->v[v-1].culled_modes = lobe_type_get_cull_mask(sampled_lobe->type);
    const mf_t throughput = shader_sample(path);
    path->v[v-1].culled_modes = 0;

    if(!mf_any(mf_gt(throughput, mf_set1(0.0f))))
      return ERR_BSDF_SAMPLE; // return without incrementing path length

    if(in_volume)
    {
      // sample distance from guide cache
      const float random = points_rand(rt.points, common_get_threadid());
      if(lobe_sample_dist(sampled_lobe, random, xpp, xp, path->e[v].omega, &path->e[v].dist, lt))
      {
        // sampled a negative distance
        return ERR_DIST_SAMPLE;
      }
      double pdf = lobe_pdf_dist(sampled_lobe, xpp, xp, path->e[v].omega, path->e[v].dist, lt);
      if(!(pdf > 0. && pdf < FLT_MAX)) return ERR_DIST_SAMPLE;
    }
    else path->e[v].dist = 1.0f; // dummy to generate some vertex position.
    // we'll overwrite this distance soon anyways.

    for(int k = 0; k < 3; k++)
      path->v[v].hit.x[k] = path->v[v-1].hit.x[k] + path->e[v].dist * path->e[v].omega[k];
  }
  else
  {
    float random[3];
    for(int k = 0; k < 2; k++) random[k] = points_rand(rt.points, common_get_threadid());
    if(in_volume)
    { // next vertex in volume
      random[2] = points_rand(rt.points, common_get_threadid());
      lobe_sample_pos_3d(sampled_lobe, random, xpp, xp, path->v[v].hit.x, lt);
      double pdf = lobe_pdf_pos_3d_dw(sampled_lobe, xpp, xp, path->v[v].hit.x, lt);
      if(!(pdf > 0. && pdf < FLT_MAX)) return ERR_PDF_ZERO_3D;
    }
    else
    { // next vertex on surface
      lobe_sample_pos_2d(sampled_lobe, random, xpp, xp, path->v[v].hit.x, lt);
      double pdf = lobe_pdf_pos_2d_dw(sampled_lobe, xpp, xp, path->v[v].hit.x, lt);
      if(!(pdf > 0. && pdf < FLT_MAX)) return ERR_PDF_ZERO_2D;
    }
  }

  // ignore path throughput for now, since we compute it after sampling the whole path anyway.
  path->v[v].throughput = mf_set1(1.f); // dummy

  // set tentative mode and flags at new vertex for path_project()
  lobe_type_set_tentative(sampled_lobe->type, path, v);
  return 0;
}

// for inner vertices, else use global p_uniform_cam
static float guided_p_uniform(guided_t *c, uint32_t lobe_index)
{
#ifdef P_UNIFORM_FIXED
  return P_UNIFORM;
#else
  return c->gmm.lobes[lobe_index].p_uniform;
#endif
}

void guided_new_path(guided_t *c)
{
  stats_record_sampled_path(c->stats);
  guided_tls_t *tls = &c->tls[common_get_threadid()];
  lobe_list_clear(&tls->lobe_list_pt);
  lobe_list_clear(&tls->lobe_list_lt);
}

void guided_invalidate_lobe_list(guided_t *g, int new_length, bool lt)
{ // assume old path was one vertex longer
  guided_tls_t *tls = &g->tls[common_get_threadid()];
  // invalidate other list
  lobe_list_t *lobe_list = lt ? &tls->lobe_list_pt : &tls->lobe_list_lt;
  lobe_list_clear_entry(lobe_list, 0);
  lobe_list_clear_entry(lobe_list, 1);
  lobe_list_update_end(lobe_list);
  lobe_list->last_path_length = new_length + 1;
}

void guided_reuse_lobe_list(guided_t *g, int new_length, bool lt)
{
  guided_tls_t *tls = &g->tls[common_get_threadid()];
  // invalidate other list
  lobe_list_t *lobe_list = lt ? &tls->lobe_list_pt : &tls->lobe_list_lt;
  lobe_list_shift_new_path(lobe_list, new_length);
}

int guided_extend(guided_t *c, path_t *path)
{
  guided_tls_t *tls = &c->tls[common_get_threadid()];
#define ERR(e) {stats_record_err_extend(c->stats, e); return e;}
#define UNINITIALIZED 0xabcdef

  const bool lt = path->v[0].mode & s_emit;
  int v = path->length; // current vertex we're initing right now.
  if(v >= PATHSPACE_MAX_VERTS) ERR(ERR_MAX_LENGTH);

  // trying to continue path from vertex on envmap
  if(v > 0 && path->v[v-1].flags & s_environment) ERR(ERR_ENV_ABSORB);
  if(v > 0 && path->v[v-1].material_modes == 0) ERR(ERR_NO_MATERIAL);

  // might have been overwritten by nee, don't accidentally terminate path
  path->v[v-1].throughput = mf_set1(1.f);

  uint32_t lobe_index = UNINITIALIZED;
  if(v == 0)
  {
    const model_t *m = &c->model;
    bool do_sample_uniform =
      lt ? true
         : points_rand(rt.points, common_get_threadid()) < m->p_uniform_cam;
    if(do_sample_uniform)
    { // fall back to path_extend() for uniform samples
      const int err = path_extend(path) ? ERR_UNIFORM_EXTEND_FIRST : NO_ERROR;
      ERR(err);
    }

    lobe_index = sample_cdf(m->cam_lobe_cdf, m->cam_lobes.n, points_rand(rt.points, common_get_threadid()));
    const int err = guided_sample_pos_first(c, path, lobe_index);
    if(err) ERR(err);
    // update v, since guided_sample_pos_first() did init the vertex on the camera
    v = path->length;
  }
  else
  {
#if GUIDE_LT == 0
    if(lt)
    {
      const int err = path_extend(path) ? ERR_UNIFORM_EXTEND : NO_ERROR;
      ERR(err)
    }
#endif

    const lobe_range_t lobe_range = guided_lobe_lookup(c, path, path->length - 1);
    const uint32_t n_lobes = lobe_range.end - lobe_range.begin;
    const lobe_list_t *lobe_list = lt ? &tls->lobe_list_lt : &tls->lobe_list_pt;
    const uint32_t *lobes = lobe_list->lobes + lobe_range.begin;
    const float *lobe_cdf = lobe_list->cdf + lobe_range.begin;

    int sample_uniform = n_lobes == 0;
    if(!sample_uniform)
    {
      const uint32_t lobe_list_index = sample_cdf(lobe_cdf, n_lobes, points_rand(rt.points, common_get_threadid()));
      lobe_index = lobes[lobe_list_index];
      const float p_uniform = guided_p_uniform(c, lobe_index);
      sample_uniform = points_rand(rt.points, common_get_threadid()) < p_uniform;
    }

    if(sample_uniform)
    {
      // fall back to path_extend() for uniform samples
      // note that we mostly still need the bvh lookup for mis computation below.
      const int err = path_extend(path) ? ERR_UNIFORM_EXTEND : NO_ERROR;
      path->v[v].throughput = mf_set1(1.);
      ERR(err)
    }
    else
    {
      const int err = guided_sample_pos_next(c, path, lobe_index);
      if(err) ERR(err);
    }
  }

  assert(lobe_index != UNINITIALIZED);
  assert(v >= 1);

  path->e[v].dist = FLT_MAX;
  path->v[v].hit.prim = INVALID_PRIMID;

  // TODO
  //// mutation mode requesting volume distance without medium?
  //// this may happen if mutated to the other side of a surface.
  //if(path->e[v].vol.shader == -1 && (path->v[v].mode & s_volume)) ERR(9);
  assert(!(path->e[v].vol.shader == -1 && (path->v[v].mode & s_volume)));

  const int err = path_project(path, v, s_propagate_mutate);
  if(err)
  {
    stats_record_err_project(c->stats, err);
    return 100 + err;
  }

  lobe_type_t lobe_type;
  if(v > 1)
  {
    const mf_t bsdf = shader_brdf(path, v-1); // sets the mode
    if(!mf_any(mf_gt(bsdf, mf_set1(0.0f)))) ERR(ERR_BSDF_ZERO);

    assert(path->v[v-1].mode != s_absorb);
    lobe_type = c->model.lobes.lobes[lobe_index].type;
  }
  else
  {
    const cam_lobe_t *lobe = &c->model.cam_lobes.lobes[lobe_index];
    // fail if we sampled the first vertex and it is outside the lobe's aabb.
    // this avoids a raycast against the bvh for mis.
    if(!util_point_in_aabb(lobe->aabb, path->v[1].hit.x))
      ERR(ERR_FIRST_OUTSIDE_AABB);
    lobe_type = lobe->type;
  }
  lobe_type_t sampled_type = lobe_type_from_path(path, v-1);
  if(sampled_type != lobe_type) ERR(ERR_WRONG_FLAGS);

  assert(mf_any(mf_gt(path->v[v].throughput, mf_set1(0.f))));

  path->length++; // now a valid vertex

  ERR(NO_ERROR);
#undef UNINITIALIZED
#undef ERR
}

// pdf for sampling vertex v on path
md_t guided_pdf_extend_dwp(guided_t *c, path_t *path, int v)
{
  assert(v > 0);
  const bool lt = path->v[0].mode & s_emit;

  const float *xp = path->v[v - 1].hit.x;
  const float *xn = path->v[v].hit.x;

  float p_uniform_cam = c->model.p_uniform_cam;

  md_t uniform_pdf;
  {
    mf_t uniform_pdf_mf;
    if(v == 1)
    {
      if(lt) uniform_pdf_mf = light_sampler_pdf_extend(path);
      else   uniform_pdf_mf = mf_mul(view_cam_pdf(path, 0), mf_set1(p_uniform_cam));
    }
    else uniform_pdf_mf = shader_pdf(path, v-1);
    uniform_pdf = mf_2d(mf_mul(uniform_pdf_mf, shader_vol_pdf(path, v)));
  }

#if GUIDE_LT == 0
  if(lt) return uniform_pdf;
#endif

  guided_tls_t *tls = &c->tls[common_get_threadid()];
  const lobe_range_t lobe_range = guided_lobe_lookup(c, path, v - 1);
  const uint32_t n_lobes = lobe_range.end - lobe_range.begin;
  if(n_lobes == 0) return uniform_pdf;

  const lobe_list_t *lobe_list = lt ? &tls->lobe_list_lt : &tls->lobe_list_pt;
  const uint32_t *lobes = lobe_list->lobes + lobe_range.begin;
  const float *lobe_cdf = lobe_list->cdf + lobe_range.begin;

  if(v == 1)
  {
    if(lt) return uniform_pdf;

    double pdf_total = 0;
    for(uint32_t i = 0; i < n_lobes; i++)
    {
      const double p_lobe = lobe_cdf[i];
      const cam_lobe_t *lobe = &c->model.cam_lobes.lobes[lobes[i]];
      double lobe_i_pdf = (path->v[v].mode & s_volume)
                        ? cam_lobe_pdf_pos_3d_dw(lobe, xp, xn)
                        : cam_lobe_pdf_pos_2d_dw(lobe, xp, xn);

      if(!(lobe_i_pdf > 0) || !(lobe_i_pdf < FLT_MAX)) lobe_i_pdf = 0;
      pdf_total += lobe_i_pdf * p_lobe;
    }

    // converts from solid angle to projected solid angle
    const double jacobian = 1.0;// /path_lambert(path, 0, path->e[1].omega);
    md_t pdf_cam_connect = md_mul(mf_2d(view_cam_pdf_connect(path, 0)), md_set1(jacobian)); // aperture pdf
    md_t guided_pdf = md_mul(md_set1(pdf_total), pdf_cam_connect);
    return md_add(md_mul(guided_pdf, md_set1(1.0 - p_uniform_cam)), uniform_pdf);
  }
  else
  {
    const float *xpp = path->v[v - 2].hit.x;

    double uniform_total = 0;
    double dist_total = 0;
    double gauss_total = 0;
    const bool sampled_vol = path->v[v].mode & s_volume;
    const lobe_type_t path_lobe_type = lobe_type_from_path(path, v-1);
    for(uint32_t i = 0; i < n_lobes; i++)
    {
      const lobe_t *lobe = &c->model.lobes.lobes[lobes[i]];
      float p_uniform = guided_p_uniform(c, lobes[i]);
      double p_lobe = i > 0 ? lobe_cdf[i] - lobe_cdf[i - 1] : lobe_cdf[i];
      p_lobe = CLAMP(p_lobe, 0, 1);
      uniform_total += p_uniform * p_lobe;

      if(lobe->type != path_lobe_type) continue;
      double pdf_gauss = lobe_pdf_dwp(lobe, path, v, lt);
      if(!(pdf_gauss < FLT_MAX)) pdf_gauss = 0;
      gauss_total += pdf_gauss * (1.0 - p_uniform) * p_lobe;

      double pdf_dist = 1.;
      if(sampled_vol)
      {
        pdf_dist = lobe_pdf_dist(lobe, xpp, xp, path->e[v].omega, path->e[v].dist, lt);
        if(!(pdf_dist > 0) || !(pdf_dist < FLT_MAX)) pdf_dist = 0;
      }
      dist_total += pdf_dist * (1.0 - p_uniform) * p_lobe;
    }
    md_t uniform_pdf_total = md_mul(md_set1(uniform_total), uniform_pdf);

    vertex_scattermode_t cull_mask = lobe_type_get_cull_mask(lobe_type_from_path(path, v-1));
    path->v[v-1].culled_modes = cull_mask;
    const mf_t pdf_bsdf = shader_pdf(path, v-1);
    path->v[v-1].culled_modes = 0;
    const float p_bsdf = guided_p_extend_bsdf(path, v);
    md_t pdf_bsdf_dist = md_mul(mf_2d(pdf_bsdf), md_set1(dist_total * p_bsdf));
    md_t guided_pdf_total = md_add(md_set1((1.-p_bsdf) * gauss_total), pdf_bsdf_dist);
    md_t pdf_total = md_add(uniform_pdf_total, guided_pdf_total);
    return pdf_total;
  }
}

void guided_record_path(guided_t *c, path_t *path, double throughput, render_mode_t render_mode)
{
  if(!(throughput > 0 && throughput < FLT_MAX)) return;

  if(render_mode != rm_learn) return;

  double target = throughput;
#if REL_CONTRIBUTION
  int ii = path->sensor.pixel_i;
  int jj = path->sensor.pixel_j;
  float pixel_intensity = c->pixel_val[jj * view_width() + ii];
  target /= pixel_intensity;
#endif
  for(int v = 1; v < path->length - 1; v++)
  {
    reservoir_sample_t sample;
    sample.type = lobe_type_from_path(path, v);
#if VERTEX_PAIRS
    if(v == 1) sample.type |= t_camera;
#endif
    for(int k = 0; k < 3; k++)
    {
#if VERTEX_PAIRS == 0
      sample.x[k] = path->v[v-1].hit.x[k];
#endif
      sample.x[V2OFFSET + k] = path->v[v  ].hit.x[k];
      sample.x[V2OFFSET + 3 + k] = path->v[v+1].hit.x[k];
    }

    float octree_pos[3];
    for(int k = 0; k < 3; k++) octree_pos[k] = path->v[v].hit.x[k];
#if ENABLE_NEIGHBOR_SPLAT
    { // randomly offset sample (for reservoir lookup only)
      float cell_aabb[6];
      octree_lookup_aabb(&c->model.octree, octree_pos, cell_aabb);
      float side_length = cell_aabb[3] - cell_aabb[0]; // should be the same for all dimensions
      side_length *= NEIGHBOR_SPLAT_RADIUS;
      for(int k = 0; k < 3; k++) octree_pos[k] += (points_rand(rt.points, common_get_threadid()) - 0.5f) * side_length;
    }
#endif
    uint32_t cell_index = octree_lookup_prim(&c->model.octree, octree_pos);
    assert(cell_index < c->model.n_cells);
    reservoir_t *reservoir = &c->model.reservoirs_tls[common_get_threadid()][cell_index];
    float xi[2];
    for(int k = 0; k < 2; k++) xi[k] = points_rand(rt.points, common_get_threadid());
    reservoir_push(reservoir, &sample, target, xi);
  }
}

#if 0
  float scene_scale, scene_bound;
  {
    const float *global_aabb = accel_aabb(rt.accel);
    float shortest_side = global_aabb[3] - global_aabb[0];
    for(int k = 1; k < 3; k++) shortest_side = MIN(shortest_side, global_aabb[3+k] - global_aabb[k]);
    float longest_side = global_aabb[3] - global_aabb[0];
    for(int k = 1; k < 3; k++) longest_side = MAX(longest_side, global_aabb[3+k] - global_aabb[k]);
    scene_scale = shortest_side;
    scene_bound = longest_side;
  }
  float radius = COV_INIT_RADIUS * scene_scale;
#if COV_INIT_RAYDIFF
  {
    float r;
    if(!guided_raydiff_radius(path, COV_INIT_RAYDIFF_SCALE, &r)) radius = r;
  }
#endif
#endif

static float guided_p_extend_bsdf(const path_t *path, int v)
{
  assert(v > 1);
  if(path->v[v-1].material_modes & s_specular) return 1.f;
  // roughness will be set to 0 for index matched materials in the dielectric's
  // prepare() shader, so these should return 1.0 here, too.
  const float r = (path->v[v-1].material_modes & s_volume) ?
    (1.0f - fabsf(path->v[v-1].interior.mean_cos)) :
    path->v[v-1].shading.roughness;
  float p_bsdf = 1.f - sqrtf(MAX(0.f, r))/BSDF_MAX_ROUGHNESS;
  p_bsdf = CLAMP(p_bsdf, 0.f, 1.f);
  return p_bsdf;
}

// evaluate whole path pdf for camera -> light
md_t guided_eval_pdf_pt(guided_t *g, path_t *path)
{
  assert(path->v[0].mode & s_sensor);
  const int end = path->length - 1;
  md_t pdf = md_set1(1.0);
#if ENABLE_NEE
  md_t pdf_nee = md_set1(0);
#endif

  // do lookups for mis in reverse to prevent gaps in lobe list
  for(int e = end; e >= 1; e--) guided_lobe_lookup(g, path, e - 1);

  for(int e = 1; e <= end; e++)
  {
    const int v = e;
    md_t pdf_edge = guided_pdf_extend_dwp(g, path, v);
#if ENABLE_NEE
    if(e == end)
    {
      pdf_nee = md_div(mf_2d(nee_pdf(path, end)), md_set1(path_G(path, end)));
      pdf_nee = md_mul(pdf, pdf_nee);
    }
#endif
    pdf = md_mul(pdf, pdf_edge);
  }

#if ENABLE_NEE
  return md_add(pdf_nee, pdf);
#else
  return pdf;
#endif
}

// evaluate whole path pdf for light -> camera
md_t guided_eval_pdf_lt(guided_t *g, path_t *path)
{
  assert(path->v[0].mode & s_emit);
  const int end = path->length - 1;
  md_t pdf = md_set1(1.0);

  // do lookups for mis in reverse to prevent gaps in lobe list
  for(int e = end - 1; e >= 1; e--) guided_lobe_lookup(g, path, e - 1);

  for(int e = 1; e < end; e++)
  {
    const int v = e;
    md_t pdf_edge = guided_pdf_extend_dwp(g, path, v);
    pdf = md_mul(pdf, pdf_edge);
  }
  md_t pdf_nee = md_div(mf_2d(nee_pdf(path, end)), md_set1(path_G(path, end)));
  pdf = md_mul(pdf, pdf_nee);
  return pdf;
}

void guided_dump_cache_debug(guided_t *c, const char *filename)
{
  {
    char vis_filename[1024];
    snprintf(vis_filename, 1024, "%s_vis", filename);
    FILE *f = fopen(vis_filename, "wb");
    model_serialize_debug(&c->model, f);
    fclose(f);
  }
  {
    char samples_filename[1024];
    snprintf(samples_filename, 1024, "%s_samples", filename);
    FILE *f = fopen(samples_filename, "wb");
    model_serialize_samples_debug(&c->model, f);
    fclose(f);
  }
}

void guided_dump_cache(const guided_t *c, const char *filename)
{
#if 0
  FILE *f = fopen(filename, "wb");
  uint32_t lobe_size = sizeof(lobe_t);
  fwrite(&lobe_size, sizeof(uint32_t), 1, f);
  gmm_lobes_to_file(&c->gmm, f);
  fclose(f);

  view_write_images("dump");
#endif
}

int guided_load_cache(guided_t *c, const char *filename)
{
#if 0
  FILE *f = fopen(filename, "rb");
  if(f == NULL) return 1;
  uint32_t lobe_size;
  fread(&lobe_size, sizeof(uint32_t), 1, f);
  assert(lobe_size == sizeof(lobe_t)); // safety check
  gmm_lobes_from_file(&c->gmm, f);

  gmm_update_bvh(&c->gmm);
  c->p_uniform_cam = P_UNIFORM_FIRST;
  c->n_v1_lobes = 0;
#endif

  return 0;
}

void guided_successful_path(guided_t *c, path_t *path)
{
  stats_record_successful_path(c->stats);
}

void guided_print_stats(guided_t *c, FILE *f, int json)
{
  if(!json)
  {
    model_memsize_t mem;
    model_get_memsize(&c->model, &mem);
    double guiding_mb = mem.guiding / (double) (1024 * 1024);
    double samples_mb = mem.samples / (double) (1024 * 1024);
    fprintf(f, "required memory (MB): %g (guiding), %g (samples)\n", guiding_mb, samples_mb);
  }
  guided_cache_stats_t cache_stats = {
    .num_lobes_v1 = c->model.cam_lobes.n,
    .num_lobes_total = c->model.lobes.n,
  };
  if(json) stats_print_json(c->stats, cache_stats, f);
  else stats_print(c->stats, cache_stats, f);
}

void guided_reset_stats(guided_t *c)
{
  stats_reset(c->stats);
}

#if 0
// taken from raydifferentials_v1, but without surface normal and just returns an isotropic radius
// returns != 0 in fail cases (path vertex is outside camera's viewing frustum)
static int guided_raydiff_radius(const path_t *path, float step, float *radius)
{
  float rd_i[3], rd_j[3];
  assert(path->v[0].mode & s_sensor);
  if(path->sensor.pixel_i < 0.0 || path->sensor.pixel_i >= view_width() ||
     path->sensor.pixel_j < 0.0 || path->sensor.pixel_j >= view_height())
    return 1;
  path_t p;
  float o[3], dotpo, dotpd, t;
  p.time = path->time;
  p.lambda = path->lambda;
  p.sensor = path->sensor;
  p.sensor.pixel_set = 1;
  p.sensor.aperture_set = 1;
  p.sensor.camid_set = 1;
  // sample new pixel offset by requested step size
  p.sensor.pixel_i += step;
  view_cam_sample(&p);
  float n[3];
  for(int k = 0; k < 3; k++) n[k] = path->v[1].hit.x[k] - path->v[0].hit.x[k];
  // intersect with plane at v[1]:
  for(int k=0;k<3;k++) o[k] = path->v[1].hit.x[k] - p.v[0].hit.x[k];
  dotpo = dotproduct(n, o);
  dotpd = dotproduct(n, p.e[1].omega);
  t = dotpo/dotpd;
  for(int k=0;k<3;k++) rd_i[k] = p.v[0].hit.x[k] + t*p.e[1].omega[k] - path->v[1].hit.x[k];
  if(!(dotproduct(rd_i,rd_i) >= 0.0)) return 1;

  // sample new pixel offset by requested step in j direction
  p.sensor.pixel_i = path->sensor.pixel_i;
  p.sensor.pixel_j += step;
  view_cam_sample(&p);
  // intersect with plane at v[1]:
  for(int k=0;k<3;k++) o[k] = path->v[1].hit.x[k] - p.v[0].hit.x[k];
  dotpo = dotproduct(n, o);
  dotpd = dotproduct(n, p.e[1].omega);
  t = dotpo/dotpd;
  for(int k=0;k<3;k++) rd_j[k] = p.v[0].hit.x[k] + t*p.e[1].omega[k] - path->v[1].hit.x[k];
  if(!(dotproduct(rd_j,rd_j) >= 0.0)) return 1;

  float r = sqrtf(MAX(dotproduct(rd_i, rd_i), dotproduct(rd_j, rd_j)));
  if(!(r > 0.f)) return 1;
  *radius = r;
  return 0;
}
#endif

static lobe_range_t lobe_list_conditional(lobe_list_t *lobe_list, const model_t *m, const float xpp[3], const float xp[3], lobe_range_t lobe_range, bool lt)
{ // evaluate marginal pdf of xpp and xp for every lobe, removing the ones that evaluate to zero
  uint32_t n_lobes = lobe_range.end - lobe_range.begin;
  float *cdf = lobe_list->cdf + lobe_range.begin;
  uint32_t *lobes = lobe_list->lobes + lobe_range.begin;

  float x[GUIDE_DIM-3];
  for(int k = 0; k < 3; k++)
  {
#if VERTEX_PAIRS == 0
    x[k] = xpp[k];
#endif
    x[V2OFFSET + k] = xp[k];
  }
  uint32_t lobe_index = 0;
  double weight_sum = 0.;
  for(uint32_t list_i = 0; list_i < n_lobes; list_i++)
  {
    const lobe_t* lobe = &m->lobes.lobes[lobes[list_i]];
    cdf[lobe_index] = lobe_eval_marginal(lobe, x, lt);
    lobes[lobe_index] = lobes[list_i];
    // remove lobes with zero weight
    if(!(cdf[lobe_index] > 0.f && cdf[lobe_index] < FLT_MAX)) continue;
    weight_sum += cdf[lobe_index];
    lobe_index++;
  }
  n_lobes = lobe_index;

  if(n_lobes > 0)
  {
    // normalize weights and build cdf
    cdf[0] /= weight_sum;
    assert(cdf[0] == cdf[0]);
    for(uint32_t i = 1; i < n_lobes; i++)
    {
      cdf[i] = cdf[i - 1] + cdf[i] / weight_sum;
      assert(cdf[i] == cdf[i]);
    }
  }

  lobe_range.end = lobe_range.begin + n_lobes;
  return lobe_range;
}
