#include "lobe.h"

#include "em.h"
#include "matrix.h"
#include "svd_symm.h"
#include "svd2.h"

#include "corona_common.h"
#include "fakegaussian2.h"
#include "matrix2.h"
#include "matrix3.h"
#include "pathspace.h"
#include "points.h"

#include <assert.h>
#include <float.h>
#include <stdbool.h>

static float gauss1d_pdf(float x);
static float gauss1d_sample(float xi);
static float gauss1d_cdf(float x);

static void compute_aabb(const float mean[3], const float E[9], const float w[3], float aabb[6]);
static int svd_d(const double *a, int n, double *w, double *v);
static int mat_generalized_inverse(const double *a, int n, double *a_inv);
static void regularize(int n, float *w);
static void regularize_d(int n, double *w);

static void sample_gauss(const int n, const float *E, const float *w, const float *mean, const float *random, float *sample);
static double eval_gauss(const int n, const float *E, const float *w, const float *mean, const float *x);
static double eval_gauss_inv(const int n, const float *E_T, const float *w_inv, const float *mean, const float *x);

// compute all conditionals and marginals for this lobe based on its 9D covariance
static int lobe_compute_derived(lobe_t *lobe, const double cov[GUIDE_DIM * GUIDE_DIM]);

// compute conditional mean for the third vertex given first two
// covariance is fixed and precomputed per lobe (S_11 - S_12 S_22^-1 S_21)
// mu = mu_1 + S_12 S_22^-1 (a - mu_2)
static void lobe_cond_mean(
    const lobe_t *lobe,
    const float xpp[3],
    const float xp[3],
    float cond_mean[3],
    bool lt);

// project 3d cond. covariance on the plane orthogonal to E[xn|xpp,xp] - xp:
// build basis around direction, rotate and decompose the marginal
// return the basis onb, lower triangular matrix l
static void cond_mean2d(
    const float E_cond[9],
    const float w_cond[3],
    const float xp[3],
    const float cond_mean[3],
    float onb[9],
    float E[4],
    float w[2]);
 
// rotate covariance of last vertex, so that it aligns with the sampled direction.
// then condiditionalize on the other two dimensions, so we get a 1d distribution along the sampled direction.
static void lobe_cond_dist(
    const lobe_t *lobe,
    const float xpp[3],
    const float xp[3],
    const float dir[3], // has to be normalized
    float *mean,
    float *sigma,
    bool lt);

// convert vertex volume measure to solid angle measure for outgoing direction at xp
static double vol_to_solid_angle(const float xp[3], const float xn[3]);

static void sample_pos_2d(const float E[9], const float w[3], const float mean[3], const float random[2], const float xp[3], float xn[3]);

static double pdf_pos2d(const float E[9], const float w[3], const float mean[3], const float xp[3], const float xn[3]);

double lobe_eval_marginal(const lobe_t *lobe, const float xn[GUIDE_DIM-3], bool lt)
{
  if(lt)
  {
#if ENABLE_LT
    assert(GUIDE_DIM == 9);
    float mean[GUIDE_DIM-3];
    for(int k = 0; k < 3; k++) mean[k] = lobe->mean[6+k];
    for(int k = 0; k < 3; k++) mean[3+k] = lobe->mean[3+k];
    return lobe->weight * eval_gauss_inv(GUIDE_DIM-3, lobe->E_v3v2_T, lobe->w_v3v2_inv, mean, xn);
#else
    assert(0);
#endif
  }
  else return lobe->weight * eval_gauss_inv(GUIDE_DIM-3, lobe->E_v1v2_T, lobe->w_v1v2_inv, lobe->mean, xn);
}

int lobe_from_comp(lobe_t *lobe, const gmm_fit_comp_t *comp, lobe_type_t type)
{
  lobe->weight = comp->weight;
  memcpy(lobe->mean, comp->mean, sizeof(float) * GUIDE_DIM);
  lobe->type = type;

  double cov[GUIDE_DIM * GUIDE_DIM], E[GUIDE_DIM * GUIDE_DIM], w[GUIDE_DIM];
  for(int k = 0; k < GUIDE_DIM; k++) w[k] = 1./comp->w_inv[k];
  float E_f[GUIDE_DIM * GUIDE_DIM];
  mat_transpose(comp->E_T, GUIDE_DIM, GUIDE_DIM, E_f);
  for(int k = 0; k < GUIDE_DIM * GUIDE_DIM; k++) E[k] = E_f[k];
  mat_compute_cov_d(GUIDE_DIM, E, w, cov);
  return lobe_compute_derived(lobe, cov);
}

int cam_lobe_from_comp(cam_lobe_t *lobe, const gmm3_fit_comp_t *comp, lobe_type_t type)
{
  lobe->weight = comp->weight;
  for(int k = 0; k < 3; k++) lobe->mean[k] = comp->mean[k];
  lobe->type = (type & t_vol_v2) ? t_vol_v3 : t_surf_v3;

  mat_transpose(comp->E_T, 3, 3, lobe->E);
  for(int k = 0; k < 3; k++) lobe->w[k] = 1./comp->w_inv[k];

  compute_aabb(lobe->mean, lobe->E, lobe->w, lobe->aabb);
  return 0;
}

static int compute_cond_blocks(const double cov[GUIDE_DIM], float *E_marg_T, float *w_marg_inv, float *cond_mean, float *E_cond, float *w_cond)
{
  const int MARG_DIM = GUIDE_DIM - 3;

  const int COND_MEAN_DIM = 3*(GUIDE_DIM-3);
  double cond_mean_d[COND_MEAN_DIM];
  {
    // first and second vertex (for conditional mixture weights): 6x6 E_v1v2, w_v1v2
    double S_v1v2[MARG_DIM * MARG_DIM]; // cov[0:6, 0:6]
    mat_extract_d(cov, GUIDE_DIM, GUIDE_DIM, 0, MARG_DIM, 0, MARG_DIM, S_v1v2);
    double E_v1v2[MARG_DIM * MARG_DIM];
    double w_v1v2[MARG_DIM];
    if(svd_d(S_v1v2, MARG_DIM, w_v1v2, E_v1v2)) return 1;
    regularize_d(MARG_DIM, w_v1v2);
    float E_v1v2_f[MARG_DIM * MARG_DIM];
    for(int k = 0; k < MARG_DIM * MARG_DIM; k++) E_v1v2_f[k] = E_v1v2[k];
    mat_transpose(E_v1v2_f, MARG_DIM, MARG_DIM, E_marg_T);
    for(int k = 0; k < MARG_DIM; k++) w_marg_inv[k] = 1./w_v1v2[k];

    // compute blocks for conditionalization and decompose: 3x3 E_cond, w_cond and 3x6 cond_mean
    double S_v1v2_inv[MARG_DIM * MARG_DIM];
    if(mat_generalized_inverse(S_v1v2, MARG_DIM, S_v1v2_inv)) return 1;

    // cond_mean = cov[6:9, 0:6] S_v1v2_inv
    double tmp[MARG_DIM * 3];
    mat_extract_d(cov, GUIDE_DIM, GUIDE_DIM, MARG_DIM, GUIDE_DIM, 0, MARG_DIM, tmp);
    mat_mul_d(tmp, S_v1v2_inv, 3, MARG_DIM, MARG_DIM, cond_mean_d);
    for(int k = 0; k < COND_MEAN_DIM; k++) cond_mean[k] = cond_mean_d[k];
  }
  {
    double S_v3_cond[9];
    // S_v3_cond = cov[6:9, 6:9] - cond_mean cov[0:6, 6:9]
    mat_extract_d(cov, GUIDE_DIM, GUIDE_DIM, V2OFFSET + 3, GUIDE_DIM, V2OFFSET + 3, GUIDE_DIM, S_v3_cond);
#if DISABLE_COND
    // uses 3x3 marginal for sampling - do nothing
#else
    { // conditionalize 3x3
      double sub[9];
      double tmp[MARG_DIM * 3];
      mat_extract_d(cov, GUIDE_DIM, GUIDE_DIM, 0, MARG_DIM, MARG_DIM, GUIDE_DIM, tmp);
      mat_mul_d(cond_mean_d, tmp, 3, MARG_DIM, 3, sub);
      for(int i = 0; i < 9; i++) S_v3_cond[i] -= sub[i];
    }
#endif
    double E_cond_d[9];
    double w_cond_d[3];
    if(svd_d(S_v3_cond, 3, w_cond_d, E_cond_d)) return 1;
    regularize_d(3, w_cond_d);
    for(int k = 0; k < 3; k++) w_cond[k] = w_cond_d[k];
    for(int k = 0; k < 9; k++) E_cond[k] = E_cond_d[k];
  }
  return 0;
}

static int lobe_compute_derived(lobe_t *lobe, const double cov[GUIDE_DIM * GUIDE_DIM])
{
  // decompose marginal blocks:

  // middle vertex (for aabb): 3x3 E_v2, w_v2
  // last vertex if camera lobe
  {
    double S_v2[9]; // cov[3:6, 3:6]
    mat_extract_d(cov, GUIDE_DIM, GUIDE_DIM, V2OFFSET, V2OFFSET + 3, V2OFFSET, V2OFFSET + 3, S_v2);
    double E_v2[9], w_v2[3];
    if(svd_d(S_v2, 3, w_v2, E_v2)) return 1;
    regularize_d(3, w_v2);

    float E_v2_f[9], w_v2_f[3];
    for(int k = 0; k < 9; k++) E_v2_f[k] = E_v2[k];
    for(int k = 0; k < 3; k++) w_v2_f[k] = w_v2[k];

    const float *aabb_mean = lobe->mean + V2OFFSET;
    compute_aabb(aabb_mean, E_v2_f, w_v2_f, lobe->aabb_v2);
  }

  if(compute_cond_blocks(cov,
    lobe->E_v1v2_T,
    lobe->w_v1v2_inv,
    lobe->cond_mean,
    lobe->E_cond,
    lobe->w_cond
  )) return 1;

#if ENABLE_LT
  assert(GUIDE_DIM == 9);

  // light tracer does only connections to camera, no guided sampling
  if(lobe->type & t_camera) return 0;

  // flip v1 and v3
  double rev_cov[GUIDE_DIM * GUIDE_DIM];
  for(int bi = 0; bi < 3; bi++) for(int bj = 0; bj < 3; bj++)
    for(int i = 0; i < 3; i++) for(int j = 0; j < 3; j++)
      rev_cov[(bi * 3 + i) * GUIDE_DIM + (bj * 3 + j)] =
        cov[((2 - bi) * 3 + i) * GUIDE_DIM + (2 - bj) * 3 + j];

  if(compute_cond_blocks(rev_cov,
    lobe->E_v3v2_T,
    lobe->w_v3v2_inv,
    lobe->rev_cond_mean,
    lobe->rev_E_cond,
    lobe->rev_w_cond
  )) return 1;
#endif

  return 0;
}

double lobe_pdf_dwp(const lobe_t *lobe, const path_t* path, int v, bool lt)
{
  assert(v > 1);

  const float *xpp = path->v[v - 2].hit.x;
  const float *xp = path->v[v - 1].hit.x;
  const float *xn = path->v[v].hit.x;

  // converts from solid angle to projected solid angle
  const double jacobian = 1.0/path_lambert(path, v-1, path->e[v].omega);

  if(path->v[v].mode & s_volume) return jacobian * lobe_pdf_pos_3d_dw(lobe, xpp, xp, xn, lt);
  else return jacobian * lobe_pdf_pos_2d_dw(lobe, xpp, xp, xn, lt);
}

void lobe_sample_pos_2d(const lobe_t *lobe, const float random[2], const float xpp[3], const float xp[3], float xn[3], bool lt)
{
  float cond_mean[3];
  lobe_cond_mean(lobe, xpp, xp, cond_mean, lt);
#if ENABLE_LT
  const float *E_cond = lt ? lobe->rev_E_cond : lobe->E_cond;
  const float *w_cond = lt ? lobe->rev_w_cond : lobe->w_cond;
#else
  const float *E_cond = lobe->E_cond;
  const float *w_cond = lobe->w_cond;
#endif
  sample_pos_2d(E_cond, w_cond, cond_mean, random, xp, xn);
}


void lobe_sample_pos_3d(const lobe_t *lobe, const float random[3], const float xpp[3], const float xp[3], float xn[3], bool lt)
{
  float cond_mean[3];
  lobe_cond_mean(lobe, xpp, xp, cond_mean, lt);
#if ENABLE_LT
  const float *E_cond = lt ? lobe->rev_E_cond : lobe->E_cond;
  const float *w_cond = lt ? lobe->rev_w_cond : lobe->w_cond;
#else
  const float *E_cond = lobe->E_cond;
  const float *w_cond = lobe->w_cond;
#endif
  sample_gauss(3, E_cond, w_cond, cond_mean, random, xn);
}

int lobe_sample_dist(const lobe_t *lobe, float random, const float xpp[3], const float xp[3], const float dir[3], float *dist, bool lt)
{
  float mean, sigma;
  lobe_cond_dist(lobe, xpp, xp, dir, &mean, &sigma, lt);
  float rand0 = gauss1d_cdf(-mean/sigma);
  if(rand0 >= 1.f) return 1; // can't sample any positive distance
  random = random * (1.f - rand0) + rand0;
  *dist = mean + gauss1d_sample(random) * sigma;
  return !(*dist > 0);
}

void cam_lobe_sample_pos_2d(const cam_lobe_t *lobe, const float random[2], const float xp[3], float xn[3])
{
  sample_pos_2d(lobe->E, lobe->w, lobe->mean, random, xp, xn);
}

void cam_lobe_sample_pos_3d(const cam_lobe_t *lobe, const float random[3], float xn[3])
{
  sample_gauss(3, lobe->E, lobe->w, lobe->mean, random, xn);
}

double lobe_pdf_pos_2d_dw(const lobe_t *lobe, const float xpp[3], const float xp[3], const float xn[3], bool lt)
{
  float cond_mean[3];
  lobe_cond_mean(lobe, xpp, xp, cond_mean, lt);
#if ENABLE_LT
  const float *E_cond = lt ? lobe->rev_E_cond : lobe->E_cond;
  const float *w_cond = lt ? lobe->rev_w_cond : lobe->w_cond;
#else
  const float *E_cond = lobe->E_cond;
  const float *w_cond = lobe->w_cond;
#endif
  return pdf_pos2d(E_cond, w_cond, cond_mean, xp, xn);
}

double lobe_pdf_pos_3d_dw(const lobe_t *lobe, const float xpp[3], const float xp[3], const float xn[3], bool lt)
{
  float cond_mean[3];
  lobe_cond_mean(lobe, xpp, xp, cond_mean, lt);
#if ENABLE_LT
  const float *E_cond = lt ? lobe->rev_E_cond : lobe->E_cond;
  const float *w_cond = lt ? lobe->rev_w_cond : lobe->w_cond;
#else
  const float *E_cond = lobe->E_cond;
  const float *w_cond = lobe->w_cond;
#endif
  double pdf = eval_gauss(3, E_cond, w_cond, cond_mean, xn);
  pdf *= vol_to_solid_angle(xp, xn);
  return pdf;
}

double lobe_pdf_dist(const lobe_t *lobe, const float xpp[3], const float xp[3], const float dir[3], float dist, bool lt)
{
  float mean, sigma;
  lobe_cond_dist(lobe, xpp, xp, dir, &mean, &sigma, lt);

  float rand0 = gauss1d_cdf(-mean/sigma);
  if(rand0 >= 1.f) return 0.f; // can't sample any positive distance

  return (gauss1d_pdf((dist - mean) / sigma) / sigma) / (1.f - rand0);
}

double cam_lobe_pdf_pos_2d_dw(const cam_lobe_t *lobe, const float xp[3], const float xn[3])
{
  return pdf_pos2d(lobe->E, lobe->w, lobe->mean, xp, xn);
}

double cam_lobe_pdf_pos_3d_dw(const cam_lobe_t *lobe, const float xp[3], const float xn[3])
{
  double pdf = eval_gauss(3, lobe->E, lobe->w, lobe->mean, xn);
  pdf *= vol_to_solid_angle(xp, xn);
  return pdf;
}

static float gauss1d_pdf(float x)
{
  x *= SIGMA_SCALE;
  return fakegaussian_pdf(x) * SIGMA_SCALE;
}

static float gauss1d_sample(float xi)
{
  return fakegaussian_sample(xi) / SIGMA_SCALE;
}

static float gauss1d_cdf(float x)
{
  x *= SIGMA_SCALE;
  x = CLAMP(x, -4.f, 4.f);
  return fakegaussian_cdf(x);
}

static void compute_aabb(const float mean[3], const float E[9], const float w[3], float aabb[6])
{
  // grow bounding boxes by gaussian support:
  // pretend to sample really far gaussian distances and bound that.
  float Bi[9];
  memcpy(Bi, E, sizeof(Bi));

  for(int k = 0; k < 3; k++)
  {
    const float scale = w[k];
    Bi[0 + k] *= scale;
    Bi[3 + k] *= scale;
    Bi[6 + k] *= scale;
  }

  for(int k = 0; k < 3; k++)
    aabb[k] = aabb[3 + k] = mean[k];
  for(int i = 0; i < 8; i++)
  {
    const float cutoff = 4.0f / SIGMA_SCALE;
    const float g[] =
    {
      (i&1) ? cutoff : -cutoff,
      (i&2) ? cutoff : -cutoff,
      (i&4) ? cutoff : -cutoff
    };
    float gt[3];
    mat3_mulv(Bi, g, gt);
    for(int k = 0; k < 3; k++)
    {
      const float corner = mean[k] + gt[k];
      aabb[    k] = MIN(aabb[    k], corner);
      aabb[3 + k] = MAX(aabb[3 + k], corner);
    }
  }
}

static void regularize(int n, float *w)
{
  for(int k = 0; k < n; k++) w[k] = MAX(w[k], COV_MIN_REG);
}

static void regularize_d(int n, double *w)
{
  for(int k = 0; k < n; k++) w[k] = MAX(w[k], COV_MIN_REG);
}

// computes svd of n x n matrix in double precision (internally):
// v * diag(w)^2 * v^T = a
// in case this needs to be faster:
// * try single precision only
// * specialized svd for symmetric matrices (tridiagonal form)
static int svd_d(const double *a, int n, double *w, double *v)
{
  for(int i = 0; i < n * n; i++) v[i] = a[i];
  int svd_err = svd_symm(n, v, w);
  if(svd_err) return svd_err;
  for(int i = 0; i < n; i++) w[i] = sqrt(MAX(0., w[i]));
  return 0;
}

static void sample_gauss(const int n, const float *E, const float *w, const float *mean, const float *random, float *sample)
{
  float d[n];
  for(int k = 0; k < n; k++) d[k] = gauss1d_sample(random[k]) * w[k];
  mat_mulv(E, d, n, n, sample);
  for(int k = 0; k < n; k++) sample[k] += mean[k];
}

static double eval_gauss(const int n, const float *E, const float *w, const float *mean, const float *x)
{
  float E_T[n * n];
  mat_transpose(E, n, n, E_T);
  float w_inv[n];
  for(int k = 0; k < n; k++) w_inv[k] = 1.f/w[k];
  return eval_gauss_inv(n, E_T, w_inv, mean, x);
}

static double eval_gauss_inv(const int n, const float *E_T, const float *w_inv, const float *mean, const float *x)
{
  float x_centered[n];
  for(int k = 0; k < n; k++) x_centered[k] = x[k] - mean[k];
  float d[n];
  mat_mulv(E_T, x_centered, n, n, d);
  double pdf = 1;
  for(int k = 0; k < n; k++) pdf *= gauss1d_pdf(d[k] * w_inv[k]) * w_inv[k];
  return pdf;
}

static void lobe_cond_mean(
    const lobe_t *lobe,
    const float xpp[3],
    const float xp[3],
    float cond_mean[3],
    bool lt)
{
  float lobe_mean[GUIDE_DIM];
  const float *lobe_cond_mean;
  if(lt)
  {
#if ENABLE_LT
    assert(GUIDE_DIM == 9);
    lobe_cond_mean = lobe->rev_cond_mean;
    for(int k = 0; k < 3; k++) lobe_mean[k]   = lobe->mean[6+k];
    for(int k = 0; k < 3; k++) lobe_mean[3+k] = lobe->mean[3+k];
    for(int k = 0; k < 3; k++) lobe_mean[6+k] = lobe->mean[k];
#else
    assert(0);
#endif
  }
  else
  {
    lobe_cond_mean = lobe->cond_mean;
    for(int k = 0; k < GUIDE_DIM; k++) lobe_mean[k] = lobe->mean[k];
  }

#if DISABLE_COND
  for(int i = 0; i < 3; i++) cond_mean[i] = lobe_mean[V2OFFSET + 3 + i];
#else

  float a[GUIDE_DIM - 3];
  for(int k = 0; k < 3; k++)
  {
#if VERTEX_PAIRS == 0
    a[k] = xpp[k] - lobe_mean[k];
#endif
    a[V2OFFSET + k] = xp[k] - lobe_mean[V2OFFSET + k];
  }
  mat_mulv(lobe_cond_mean, a, 3, GUIDE_DIM-3, cond_mean);
  for(int i = 0; i < 3; i++) cond_mean[i] += lobe_mean[(V2OFFSET+3) + i];
#endif
}

static void cond_mean2d(
    const float E_cond[9],
    const float w_cond[3],
    const float xp[3],
    const float cond_mean[3],
    float onb[9],
    float E[4],
    float w[2])
{
  // build basis
  for (int k = 0; k < 3; k++) onb[k] = cond_mean[k] - xp[k];
  normalise(onb);
  get_onb(onb, onb+3, onb+6);

  float E_onb[9], cov[9];
  mat3_mul(onb, E_cond, E_onb);
  mat_compute_cov(3, E_onb, w_cond, cov);

  // marginalize to 2nd and 3rd dimension and decompose
  double cov_d[4];
  cov_d[0] = cov[4]; cov_d[1] = cov[5];
  cov_d[2] = cov[7]; cov_d[3] = cov[8];
  double w_d[2], E_T_d[4];
  getEVDSymmetric2x2(w_d, E_T_d, cov_d);
  for(int k = 0; k < 2; k++) w[k] = sqrtf(w_d[k]);
  E[0] = E_T_d[0]; E[1] = E_T_d[2];
  E[2] = E_T_d[1]; E[3] = E_T_d[3];
  regularize(2, w);
}

static void lobe_cond_dist(
    const lobe_t *lobe,
    const float xpp[3],
    const float xp[3],
    const float dir[3], // has to be normalized
    float *mean,
    float *sigma,
    bool lt)
{
  float cond_mean[3], onb[9];
  lobe_cond_mean(lobe, xpp, xp, cond_mean, lt);
  for(int k = 0; k < 3; k++) onb[k] = dir[k];
  get_onb(onb, onb+3, onb+6);

  float E_onb[9];
  double cov[9];
  if(lt)
  {
#if ENABLE_LT
    mat3_mul(onb, lobe->rev_E_cond, E_onb);
    double E_onb_d[9], w_cond_d[3];
    for(int k = 0; k < 9; k++) E_onb_d[k] = E_onb[k];
    for(int k = 0; k < 3; k++) w_cond_d[k] = lobe->rev_w_cond[k];
    mat_compute_cov_d(3, E_onb_d, w_cond_d, cov);
#else
    assert(0);
#endif
  }
  else
  {
    mat3_mul(onb, lobe->E_cond, E_onb);
    double E_onb_d[9], w_cond_d[3];
    for(int k = 0; k < 9; k++) E_onb_d[k] = E_onb[k];
    for(int k = 0; k < 3; k++) w_cond_d[k] = lobe->w_cond[k];
    mat_compute_cov_d(3, E_onb_d, w_cond_d, cov);
  }

  double S12S22i[2];
  {
    double S22[4], S22i[4];
    mat_extract_d(cov, 3, 3, 1, 3, 1, 3, S22);
    int err = mat_generalized_inverse(S22, 2, S22i);
    if(err)
    {
      *mean = -1;
      *sigma = 0;
      return;
    }
    double S12[2] = { cov[1], cov[2] };
    mat_mul_d(S12, S22i, 1, 2, 2, S12S22i);
  }
  { // S = S11 - S12 * S22^-1 * S21
    double sigma2 = cov[0] - S12S22i[0] * cov[3] - S12S22i[1] * cov[6];
    *sigma = sqrt(sigma2);
    *sigma = MAX(*sigma, COV_MIN_REG);
  }
  { // mu = mu1 + S12 * S22^-1 (a - mu2)
    float mean_onb[3], xp_onb[3];
    mat3_mulv(onb, cond_mean, mean_onb);
    mat3_mulv(onb, xp, xp_onb);
    float d[2];
    for(int k = 0; k < 2; k++) d[k] = xp_onb[1+k] - mean_onb[1+k];
    *mean = mean_onb[0] + S12S22i[0] * d[0] + S12S22i[1] * d[1] - xp_onb[0];
  }
}

// convert vertex volume measure to solid angle measure for outgoing direction at xp
static double vol_to_solid_angle(const float xp[3], const float xn[3])
{
  double e[3];
  for(int k = 0; k < 3; k++) e[k] = xn[k] - xp[k];
  return dotproduct(e, e);
}

static void sample_pos_2d(const float E[9], const float w[3], const float mean[3], const float random[2], const float xp[3], float xn[3])
{
  float onb[9], E_sample[4], w_sample[2];
  cond_mean2d(E, w, xp, mean, onb, E_sample, w_sample);

  float sample[3] = {};
  float zero[2] = {};
  sample_gauss(2, E_sample, w_sample, zero, random, sample+1);
  // xn = onb^T sample
  float onb_T[9];
  mat3_transpose(onb, onb_T);
  mat3_mulv(onb_T, sample, xn);
  for(int k = 0; k < 3; k++) xn[k] += mean[k];
}

static double pdf_pos2d(const float E[9], const float w[3], const float mean[3], const float xp[3], const float xn[3])
{
  float onb[9], E_sample[4], w_sample[2];
  cond_mean2d(E, w, xp, mean, onb, E_sample, w_sample);

  // project to plane at dist
  float wo[3];
  for (int k = 0; k < 3; k++) wo[k] = xn[k] - xp[k];
  normalise(wo);
  const float t = (dotproduct(onb, mean) - dotproduct(onb, xp)) / dotproduct(onb, wo);
  if (!(t > 0)) return 0.0;
  float offset[3];
  for (int k = 0; k < 3; k++) offset[k] = xp[k] + t * wo[k] - mean[k];

  // transform into onb
  float d[2];
  mat_mulv(onb+3, offset, 2, 3, d);
  float zero[2] = {};
  double pdf = eval_gauss(2, E_sample, w_sample, zero, d);

  // convert to solid angle measure
  pdf *= t * t / fabsf(dotproduct(onb, wo));
  return pdf;
}

static int mat_generalized_inverse(const double *a, int n, double *a_inv)
{
  double a_d[n * n], w_d[n];
  for(int i = 0; i < n * n; i++) a_d[i] = a[i];
  int svd_err = svd_symm(n, a_d, w_d);
  if(svd_err) return svd_err;

  double max_w = w_d[0];
  for(int k = 1; k < n; k++) max_w = MAX(w_d[k], max_w);
  //double thresh = max_w * 0.5 * sqrt(n + n + 1) * FLT_EPSILON;
  double thresh = COV_MIN_REG * COV_MIN_REG;
  for(int i = 0; i < n; i++)
  {
    if(w_d[i] < thresh) w_d[i] = 0;
    else w_d[i] = 1./w_d[i];
  }

  double a_d_T[n * n];
  mat_transpose_d(a_d, n, n, a_d_T);
  double tmp[n * n];
  mat_mul_diag_d(w_d, a_d_T, n, n, tmp);
  mat_mul_d(a_d, tmp, n, n, n, a_inv);

  return 0;
}
