#pragma once

#include "config.h"
#include "lobe_type.h"

#include <stdbool.h>
#include <stdint.h>

typedef struct gmm_fit_comp_t gmm_fit_comp_t;
typedef struct gmm3_fit_comp_t gmm3_fit_comp_t;

// one lobe represents a triplet of vertices on a path: v1->v2->v3
// per lobe:
// weight 1 float
// mean   9 float
// cov   81 float v
//        9 float singular values
// + precomputed 3x6 (S_ab S_bb^{-1}) and 3x3 S_a|b = S_aa - (S_ab S_bb^-1 S_ba) blocks or so
// TODO space efficient representation of rotations?
// storing covariances as triangular matrices instead (cholesky) would need only half the floats
typedef struct lobe_t
{
  float weight;
  float mean[GUIDE_DIM];

  float cond_mean[3*(GUIDE_DIM-3)]; // 3x6 transforming mean: mean_a|b = mean_a + [S_ab S_bb^-1] (x_b - mean_b)
  float E_cond[9];     // 3x3 covariance:        S_a|b    = S_aa - S_ab S_bb^-1 S_ba
  float w_cond[3];

  float aabb_v2[6];
  // marginal of first and second vertex, decomposed
  float E_v1v2_T[(GUIDE_DIM-3)*(GUIDE_DIM-3)];
  float w_v1v2_inv[GUIDE_DIM-3];

#if ENABLE_LT
  float rev_cond_mean[3*(GUIDE_DIM-3)];
  float rev_E_cond[9];
  float rev_w_cond[3];

  float E_v3v2_T[(GUIDE_DIM-3)*(GUIDE_DIM-3)];
  float w_v3v2_inv[GUIDE_DIM-3];
#endif

  float p_uniform;
  lobe_type_t type;
} lobe_t;

// to sample first vertex in scene
typedef struct cam_lobe_t
{
  float weight;
  float E[9];
  float w[3];
  float mean[3];
  lobe_type_t type;
  float aabb[6];
} cam_lobe_t;

typedef struct path_t path_t;

// evaluate weighted marginal pdf of lobe for first two vertices of a triplet
double lobe_eval_marginal(const lobe_t *lobe, const float xn[GUIDE_DIM-3], bool lt);

// make a sampling lobe out of a mixture component
// mainly precomputes submatrices and their decomposition
// return error (!= 0) when some decomposition fails
int lobe_from_comp(lobe_t *lobe, const gmm_fit_comp_t *comp, lobe_type_t type);

// make a (after camera) sampling lobe out of a mixture component
// mainly precomputes submatrices and their decomposition
// return error (!= 0) when some decomposition fails
int cam_lobe_from_comp(cam_lobe_t *lobe, const gmm3_fit_comp_t *comp, lobe_type_t type);

// pdf of vertex v on path v using this lobe in projected solid angle
double lobe_pdf_dwp(const lobe_t *lobe, const path_t* path, int v, bool lt);

// sample third vertex given the first two (either 2d marginal for surfaces or 3d for volumes)
void lobe_sample_pos_2d(const lobe_t *lobe, const float random[2], const float xpp[3], const float xp[3], float xn[3], bool lt);
void lobe_sample_pos_3d(const lobe_t *lobe, const float random[3], const float xpp[3], const float xp[3], float xn[3], bool lt);

// sample 1d distance given first two vertices and fixed direction
int lobe_sample_dist(const lobe_t *lobe, float random, const float xpp[3], const float xp[3], const float dir[3], float *dist, bool lt);

void cam_lobe_sample_pos_2d(const cam_lobe_t *lobe, const float random[2], const float xp[3], float xn[3]);
void cam_lobe_sample_pos_3d(const cam_lobe_t *lobe, const float random[3], float xn[3]);

// evaluate pdf for sampled positions in solid angle measure
double lobe_pdf_pos_2d_dw(const lobe_t *lobe, const float xpp[3], const float xp[3], const float xn[3], bool lt);
double lobe_pdf_pos_3d_dw(const lobe_t *lobe, const float xpp[3], const float xp[3], const float xn[3], bool lt);

double lobe_pdf_dist(const lobe_t *lobe, const float xpp[3], const float xp[3], const float dir[3], float dist, bool lt);

double cam_lobe_pdf_pos_2d_dw(const cam_lobe_t *lobe, const float xp[3], const float xn[3]);
double cam_lobe_pdf_pos_3d_dw(const cam_lobe_t *lobe, const float xp[3], const float xn[3]);
