#pragma once
#include <string.h>
#include <math.h>

// some matrix helper functions
// all matrices are row major

// matrix matrix multiplication
// a: n x m matrix
// b: m x p matrix
// res = a b (n x p matrix)
static inline void mat_mul(
    const float *const restrict a,
    const float *const restrict b,
    const int n,
    const int m,
    const int p,
    float *const restrict res)
{
  memset(res, 0, sizeof(float) * n * p);
  for(int i = 0; i < n; i++)
  {
    for(int j = 0; j < p; j++)
    {
      for(int k = 0; k < m; k++)
      {
        res[i * p + j] += a[i * m + k] * b[k * p + j];
      }
    }
  }
}

// matrix matrix multiplication
// a: n x m matrix
// b: m x p matrix
// res = a b (n x p matrix)
static inline void mat_mul_d(
    const double *const restrict a,
    const double *const restrict b,
    const int n,
    const int m,
    const int p,
    double *const restrict res)
{
  memset(res, 0, sizeof(double) * n * p);
  for(int i = 0; i < n; i++)
  {
    for(int j = 0; j < p; j++)
    {
      for(int k = 0; k < m; k++)
      {
        res[i * p + j] += a[i * m + k] * b[k * p + j];
      }
    }
  }
}

// matrix vector multiplication
// a: n x m matrix
// b: vector of length m
// res = a b (vector of length n)
static inline void mat_mulv(
    const float *const restrict a,
    const float *const restrict b,
    const int n,
    const int m,
    float *const restrict res)
{
  memset(res, 0, sizeof(float) * n);
  for(int i = 0; i < n; i++)
  {
    for(int j = 0; j < m; j++)
    {
      res[i] += a[i * m + j] * b[j];
    }
  }
}

// multiplication with diagonal matrix from the left
// a: diagonal (vector of length n)
// b: n x m matrix
// res = diag(a) b (n x m matrix)
static inline void mat_mul_diag(
    const float *const restrict a,
    const float *const restrict b,
    const int n,
    const int m,
    float *const restrict res)
{
  for(int i = 0; i < n; i++)
  {
    for(int j = 0; j < m; j++)
    {
      res[i*n+j] = a[i] * b[i*n+j];
    }
  }
}

// multiplication with diagonal matrix from the left
// a: diagonal (vector of length n)
// b: n x m matrix
// res = diag(a) b (n x m matrix)
static inline void mat_mul_diag_d(
    const double *const restrict a,
    const double *const restrict b,
    const int n,
    const int m,
    double *const restrict res)
{
  for(int i = 0; i < n; i++)
  {
    for(int j = 0; j < m; j++)
    {
      res[i*n+j] = a[i] * b[i*n+j];
    }
  }
}

// matrix transpose
static inline void mat_transpose(
    const float *const restrict a,
    const int n,
    const int m,
    float *const restrict res)
{
  for(int i = 0; i < n; i++)
  {
    for(int j = 0; j < m; j++)
    {
      res[j * n + i] = a[i * m + j];
    }
  }
}

// matrix transpose
static inline void mat_transpose_d(
    const double *const restrict a,
    const int n,
    const int m,
    double *const restrict res)
{
  for(int i = 0; i < n; i++)
  {
    for(int j = 0; j < m; j++)
    {
      res[j * n + i] = a[i * m + j];
    }
  }
}

// extract block matrix a[i0:i1, j0:j1]
// ranges are exclusive
static inline void mat_extract(
    const float *const restrict a,
    const int n,
    const int m,
    const int i0,
    const int i1,
    const int j0,
    const int j1,
    float *const restrict b
    )
{
  const int n2 = i1 - i0;
  const int m2 = j1 - j0;
  for(int i = 0; i < n2; i++)
  {
    for(int j = 0; j < m2; j++)
    {
      b[i * m2 + j] = a[(i0 + i) * m + (j0 + j)];
    }
  }
}

// extract block matrix a[i0:i1, j0:j1]
// ranges are exclusive
static inline void mat_extract_d(
    const double *const restrict a,
    const int n,
    const int m,
    const int i0,
    const int i1,
    const int j0,
    const int j1,
    double *const restrict b
    )
{
  const int n2 = i1 - i0;
  const int m2 = j1 - j0;
  for(int i = 0; i < n2; i++)
  {
    for(int j = 0; j < m2; j++)
    {
      b[i * m2 + j] = a[(i0 + i) * m + (j0 + j)];
    }
  }
}

// compute covariance from eigendecomposition cov = E diag(w^2) E^T
static inline void mat_compute_cov(
    int n,
    const float *const restrict E,
    const float *const restrict w,
    float *const restrict cov)
{
  float tmp[n * n];
  for(int i = 0; i < n; i++) for(int j = 0; j < n; j++)
    tmp[i * n + j] = E[i * n + j] * w[j];
  float tmp_T[n * n];
  mat_transpose(tmp, n, n, tmp_T);
  mat_mul(tmp, tmp_T, n, n, n, cov);
  // make symmetric (TODO only compute upper triangular before)
  for(int i = 0; i < n; i++) for(int j = 0; j < i; j++)
    cov[i * n + j] = cov[j * n + i];
}

// compute covariance from eigendecomposition cov = E diag(w^2) E^T
static inline void mat_compute_cov_d(
    int n,
    const double *const restrict E,
    const double *const restrict w,
    double *const restrict cov)
{
  double tmp[n * n];
  for(int i = 0; i < n; i++) for(int j = 0; j < n; j++)
    tmp[i * n + j] = E[i * n + j] * w[j];
  double tmp_T[n * n];
  mat_transpose_d(tmp, n, n, tmp_T);
  mat_mul_d(tmp, tmp_T, n, n, n, cov);
  // make symmetric (TODO only compute upper triangular before)
  for(int i = 0; i < n; i++) for(int j = 0; j < i; j++)
    cov[i * n + j] = cov[j * n + i];
}
