#include "bvh.h"
#include "model.h"
#include "octree.h"
#include "reservoir.h"
#include "stats.h"

#include "accel.h"
#include "points.h"
#include "threads.h"

//--- static declarations ------------------------------------------------------

typedef struct lobe_type_count_t
{
  lobe_type_t type;
  uint32_t count;
} lobe_type_count_t;

static uint32_t model_update_octree_before(model_t *m);
static void model_update_octree_after(model_t *m, uint32_t threshold);

static void model_update_bvh(model_t *m);

static void model_update_cam_cdf(model_t *m);

static void *par_update_mixture(void *arg);

static void *par_reset_reservoirs(void *arg);

static void model_update_mixture(model_t *m, uint32_t cell_index);

static uint32_t model_extract_samples(const model_t *m, uint32_t cell_index, lobe_type_t type, em_sample_t *sample_buffer, bool camera);

// type counts must have space for at least LOBE_TYPE_COUNT elements
static uint32_t model_count_sample_types(const model_t *m, uint32_t cell_index, lobe_type_count_t *type_counts, bool camera);

static void lobe_buffer_init(lobe_buffer_t *lobe_buffer);
static void lobe_buffer_cleanup(lobe_buffer_t *lobe_buffer);
static void lobe_buffer_resize(lobe_buffer_t *lobe_buffer, uint32_t n_lobes);

static void cam_lobe_buffer_init(cam_lobe_buffer_t *lobe_buffer);
static void cam_lobe_buffer_cleanup(cam_lobe_buffer_t *lobe_buffer);
static void cam_lobe_buffer_resize(cam_lobe_buffer_t *lobe_buffer, uint32_t n_lobes);

static void model_octree_fill_leaves_sample_counts(model_t *model);
static void model_octree_propagate_sample_counts(model_t *model);
static void model_octree_adapt_collapse(model_t *m, uint32_t threshold);
static void model_octree_adapt_subdivide(model_t *m, uint32_t threshold);
static void model_octree_adapt_finalize(model_t *m);
static void model_octree_subdivide_leaf(model_t *m, uint32_t node_index, const float node_aabb[6]);
static void model_octree_collapse_node(model_t *m, uint32_t node_index, const float node_aabb[6]);

static void model_resize_cells(model_t *m, uint32_t n_cells);

static void model_init_cell(model_t *m, uint32_t cell_index, const float aabb[6]);

static float aabb_min_side(const float aabb[6]);

//--- public interface ---------------------------------------------------------

void model_init(model_t *m, guided_stats_t *stats)
{
  m->reservoirs_tls = malloc(sizeof(reservoir_t*) * rt.num_threads);
  m->update_tls = malloc(sizeof(model_update_tls_t) * rt.num_threads);

  m->n_cells = 0;
  m->cells_allocated = 16;
  m->cells = malloc(sizeof(cell_mixture_t) * m->cells_allocated);

  for(int t = 0; t < rt.num_threads; t++)
  {
    m->reservoirs_tls[t] = malloc(sizeof(reservoir_t) * m->cells_allocated);

    model_update_tls_t *tls = &m->update_tls[t];
    gmm_fit_init(&tls->gmm_fit);
    gmm3_fit_init(&tls->gmm3_fit);
    lobe_buffer_init(&tls->lobe_buffer);
    cam_lobe_buffer_init(&tls->cam_lobe_buffer);
  }

  m->cam_lobe_cdf = NULL;
  m->p_uniform_cam = 1.f;

  const float *global_aabb = accel_aabb(rt.accel);
  octree_init(&m->octree, global_aabb, sizeof(uint32_t));

  m->bvh = NULL;
  m->bvh_prims = NULL;

  model_resize_cells(m, 8 * 8 * 8);
  for(uint32_t offset1 = 0; offset1 < 8; offset1++)
  {
    float child1_aabb[6];
    memcpy(child1_aabb, global_aabb, sizeof(float) * 6);
    octree_compute_child_aabb(child1_aabb, offset1);
    for(uint32_t offset2 = 0; offset2 < 8; offset2++)
    {
      float child2_aabb[6];
      memcpy(child2_aabb, child1_aabb, sizeof(float) * 6);
      octree_compute_child_aabb(child2_aabb, offset2);
      for(uint32_t offset3 = 0; offset3 < 8; offset3++)
      {
        float child3_aabb[6];
        memcpy(child3_aabb, child2_aabb, sizeof(float) * 6);
        octree_compute_child_aabb(child3_aabb, offset3);
        uint32_t cell_index = (offset1 * 8 + offset2) * 8 + offset3;
        model_init_cell(m, cell_index, child3_aabb);
      }
    }
  }
  for(int i = 0; i < 8 + 8 * 8; i++) octree_adapt_subdivide_cell(&m->octree, i);
  octree_adapt_compact(&m->octree);
  for(uint32_t i = 0; i < m->n_cells; i++)
    octree_set_prim(&m->octree, m->cells[i].octree_pos, i);

  lobe_buffer_init(&m->lobes);
  lobe_buffer_init(&m->lobes_back);
  threads_mutex_init(&m->lobes_back_mutex, NULL);

  cam_lobe_buffer_init(&m->cam_lobes);
  cam_lobe_buffer_init(&m->cam_lobes_back);
  threads_mutex_init(&m->cam_lobes_back_mutex, NULL);

  m->stats = stats;
}

void model_cleanup(model_t *m)
{
  lobe_buffer_cleanup(&m->lobes);
  lobe_buffer_cleanup(&m->lobes_back);
  threads_mutex_destroy(&m->lobes_back_mutex);

  cam_lobe_buffer_cleanup(&m->cam_lobes);
  cam_lobe_buffer_cleanup(&m->cam_lobes_back);
  threads_mutex_destroy(&m->cam_lobes_back_mutex);

  free(m->cam_lobe_cdf);

  octree_cleanup(&m->octree);

  if(m->bvh != NULL) guided_bvh_cleanup(m->bvh);
  free(m->bvh_prims);

  for(int t = 0; t < rt.num_threads; t++)
  {
    for(uint32_t i = 0; i < m->n_cells; i++)
      reservoir_cleanup(&m->reservoirs_tls[t][i]);
    free(m->reservoirs_tls[t]);

    model_update_tls_t *tls = &m->update_tls[t];
    gmm_fit_cleanup(&tls->gmm_fit);
    gmm3_fit_cleanup(&tls->gmm3_fit);
    lobe_buffer_cleanup(&tls->lobe_buffer);
    cam_lobe_buffer_cleanup(&tls->cam_lobe_buffer);
  }

  free(m->cells);

  free(m->reservoirs_tls);
  free(m->update_tls);
}

void model_update(model_t *m)
{
#if OCTREE_FIXED == 0
  uint32_t threshold = model_update_octree_before(m);
#endif

  lobe_buffer_resize(&m->lobes_back, 0);
  cam_lobe_buffer_resize(&m->cam_lobes_back, 0);

  // update all mixtures in parallel
  m->job = 0;
  m->num_jobs = m->n_cells;
  threads_t *t = rt.threads;
  for(int k=0; k < rt.num_threads; k++)
    pthread_pool_task_init(t->task + k, &t->pool, par_update_mixture, m);
  pthread_pool_wait(&t->pool);

  {
    lobe_buffer_t tmp = m->lobes;
    m->lobes = m->lobes_back;
    m->lobes_back = tmp;
  }

  {
    cam_lobe_buffer_t tmp = m->cam_lobes;
    m->cam_lobes = m->cam_lobes_back;
    m->cam_lobes_back = tmp;
  }

  model_update_bvh(m);
  model_update_cam_cdf(m);

#if OCTREE_FIXED == 0
  model_update_octree_after(m, threshold);
#endif
}

void model_prepare_frame(model_t *m)
{
  // reset all reservoirs
  m->job = 0;
  m->num_jobs = m->n_cells;
  threads_t *t = rt.threads;
  for(int k=0; k < rt.num_threads; k++)
    pthread_pool_task_init(t->task + k, &t->pool, par_reset_reservoirs, m);
  pthread_pool_wait(&t->pool);
}

uint32_t model_fill_lobe_list(model_t *m, const float xpp[3], const float xp[3], uint32_t *lobe_list, uint32_t n_max)
{
  uint32_t cell_index = octree_lookup_prim(&m->octree, xp);
  assert(cell_index < m->n_cells);

  cell_mixture_t *mixture =  &m->cells[cell_index];

  if(mixture->n_lobes <= n_max)
  {
    for(uint32_t i = 0; i < mixture->n_lobes; i++)
      lobe_list[i] = mixture->lobes_offset + i;
  }
#ifndef NDEBUG
  else for(uint32_t i = 0; i < n_max; i++) lobe_list[i] = -1;
#endif
  return mixture->n_lobes;
}

uint32_t model_fill_lobe_list_camera(model_t *m, const float xp[3], uint32_t *lobe_list, uint32_t n_max)
{
  if(m->bvh == NULL || m->p_uniform_cam == 1.f) return 0;
  uint32_t n_lobes = guided_bvh_lookup(m->bvh, xp, lobe_list, n_max);
  if(n_lobes < n_max) // if n_lobes == n_max cannot be sure to have found everything
  {
    // replace indices into prim array with lobe index
    for(uint32_t i = 0; i < n_lobes; i++)
      lobe_list[i] = m->bvh_prims[lobe_list[i]];
    return n_lobes;
  }
  else return n_max + 1; // request more memory
}

void model_serialize_debug(const model_t *m, FILE *f)
{
  int dim = GUIDE_DIM;
  fwrite(&dim, sizeof(dim), 1, f);
  int N = m->lobes.n;
  fwrite(&N, sizeof(N), 1, f);
  for(int i = 0; i < N; i++)
  {
    lobe_t *lobe = &m->lobes.lobes[i];
    fwrite(&lobe->weight, sizeof(lobe->weight), 1, f);
    uint32_t type = lobe->type;
    fwrite(&type, sizeof(type), 1, f);
    fwrite(lobe->E_cond, sizeof(lobe->E_cond), 1, f);
    fwrite(lobe->w_cond, sizeof(lobe->w_cond), 1, f);
    fwrite(lobe->mean, sizeof(lobe->mean), 1, f);
    fwrite(lobe->cond_mean, sizeof(lobe->cond_mean), 1, f);
    fwrite(lobe->E_v1v2_T, sizeof(lobe->E_v1v2_T), 1, f);
    fwrite(lobe->w_v1v2_inv, sizeof(lobe->w_v1v2_inv), 1, f);
  }
  // now the octree
  const octree_t *octree = &m->octree;
  fwrite(&octree->nodes.n, sizeof(uint32_t), 1, f);
  {
    // assume the lowest bit is used for the leaf bit
    // but actually it is up to the compiler
    octree_idx_t leaf = {.leaf = 1, .index = 0};
    uint32_t bits;
    memcpy(&bits, &leaf, sizeof(bits));
    assert(bits == 1);
  }
  fwrite(octree->aabb, sizeof(float), 6, f);
  fwrite(octree->nodes.nodes, sizeof(octree_idx_t), octree->nodes.n, f);
  // and the mixtures
  fwrite(&m->n_cells, sizeof(uint32_t), 1, f);
  for(uint32_t i = 0; i < m->n_cells; i++)
  {
    cell_mixture_t *mixture = &m->cells[i];
    fwrite(&mixture->n_lobes, sizeof(mixture->n_lobes), 1, f);
    fwrite(&mixture->lobes_offset, sizeof(mixture->lobes_offset), 1, f);
    fwrite(mixture->octree_pos, sizeof(mixture->octree_pos), 1, f);
  }
  // camera mixture
  fwrite(&m->cam_lobes.n, sizeof(uint32_t), 1, f);
  for(uint32_t i = 0; i < m->cam_lobes.n; i++)
  {
    cam_lobe_t *cam_lobe = &m->cam_lobes.lobes[i];
    fwrite(&cam_lobe->weight, sizeof(cam_lobe->weight), 1, f);
    uint32_t type = cam_lobe->type;
    fwrite(&type, sizeof(type), 1, f);
    fwrite(cam_lobe->E, sizeof(cam_lobe->E), 1, f);
    fwrite(cam_lobe->w, sizeof(cam_lobe->w), 1, f);
    fwrite(cam_lobe->mean, sizeof(cam_lobe->mean), 1, f);
  }
}

void model_serialize_samples_debug(const model_t *m, FILE *f)
{
  int dim = GUIDE_DIM;
  fwrite(&dim, sizeof(dim), 1, f);
  uint32_t n_samples_total = 0;
  for(uint32_t cell_i = 0; cell_i < m->n_cells; cell_i++)
    for(uint32_t t = 0; t < rt.num_threads; t++)
      n_samples_total += reservoir_stored_samples(&m->reservoirs_tls[t][cell_i]);
  fwrite(&n_samples_total, sizeof(n_samples_total), 1, f);
  for(uint32_t cell_i = 0; cell_i < m->n_cells; cell_i++)
  {
    for(uint32_t t = 0; t < rt.num_threads; t++)
    {
      reservoir_t *r = &m->reservoirs_tls[t][cell_i];
      for(uint32_t heap_i = 0; heap_i < r->heap->end; heap_i++)
      {
        float weight = -r->heap->vals[heap_i];
        reservoir_sample_t *s = &r->samples[r->heap->keys[heap_i]];
        fwrite(&weight, sizeof(weight), 1, f);
        uint32_t type = s->type;
        fwrite(&type, sizeof(type), 1, f);
        fwrite(&s->x, sizeof(weight), GUIDE_DIM, f);
      }
      for(uint32_t sample_i = r->samples_begin; sample_i < r->capacity + 1; sample_i++)
      {
        float weight = r->equal_weight;
        reservoir_sample_t *s = &r->samples[sample_i];
        fwrite(&weight, sizeof(weight), 1, f);
        uint32_t type = s->type;
        fwrite(&type, sizeof(type), 1, f);
        fwrite(&s->x, sizeof(weight), GUIDE_DIM, f);
      }
    }
  }
}


void model_get_memsize(const model_t *m, model_memsize_t *mem_out)
{
  mem_out->guiding = 0;

  mem_out->guiding += sizeof(cell_mixture_t) * m->n_cells;
  mem_out->guiding += sizeof(cam_lobe_t) * m->cam_lobes.n;
  mem_out->guiding += sizeof(cam_lobe_t) * m->cam_lobes_back.n;
  mem_out->guiding += sizeof(float) * m->cam_lobes.n;
  mem_out->guiding += sizeof(lobe_t) * m->lobes.n;
  mem_out->guiding += sizeof(lobe_t) * m->lobes_back.n;
  mem_out->guiding += octree_get_size(&m->octree);
  mem_out->guiding += guided_bvh_get_size(m->bvh);
  mem_out->guiding += sizeof(uint32_t) * m->cam_lobes.n;
  mem_out->guiding += sizeof(cell_mixture_t) * m->n_cells;


  mem_out->samples = 0;
  for(uint32_t i = 0; i < m->n_cells; i++)
  {
    for(int t = 0; t < rt.num_threads; t++)
    {
      mem_out->samples += sizeof(reservoir_t);
      uint32_t capacity = (m->reservoirs_tls[t][i].capacity + 1);
      mem_out->samples += capacity * sizeof(reservoir_sample_t);
      mem_out->samples += sizeof(heap_t);

      mem_out->samples += sizeof(uint64_t)*capacity;
      mem_out->samples += sizeof(float)   *capacity;
    }
  }
}

//--- static functions ---------------------------------------------------------

static uint32_t model_update_octree_before(model_t *m)
{ // collapse inner nodes if they do not have enough samples
  model_octree_fill_leaves_sample_counts(m);
  model_octree_propagate_sample_counts(m);
#if OCTREE_LEAF_TARGET > 0
  uint32_t *node_data = (uint32_t*) m->octree.node_data;
  uint32_t sample_count_total = 0;
  for(int k = 0; k < 8; k++)
    sample_count_total += node_data[k];
  uint32_t threshold = sample_count_total / OCTREE_LEAF_TARGET;
#else
  uint32_t threshold = OCTREE_SUBDIV_THRESHOLD;
#endif
  model_octree_adapt_collapse(m, threshold);
  return threshold;
}

static void model_update_octree_after(model_t *m, uint32_t threshold)
{ // subdivide leaf nodes (for next iteration) if they have enough samples
  model_octree_fill_leaves_sample_counts(m);
  model_octree_adapt_subdivide(m, threshold);
}

static void model_update_bvh(model_t *m)
{
  // throw away old bvh and rebuild
  if(m->bvh != NULL) guided_bvh_cleanup(m->bvh);

  m->bvh_prims = realloc(m->bvh_prims, sizeof(uint32_t) * m->cam_lobes.n);
  m->bvh = guided_bvh_init(m->bvh_prims, m->cam_lobes.n);

  float *bvh_aabb = guided_bvh_get_prim_aabb(m->bvh);
  for(uint32_t i = 0, N = m->cam_lobes.n; i < N; i++)
  {
    m->bvh_prims[i] = i;
    const cam_lobe_t *lobe = &m->cam_lobes.lobes[i];
    for(int k = 0; k < 6; k++) bvh_aabb[6 * i + k] = lobe->aabb[k];
  }

  guided_bvh_build(m->bvh, NULL);
}

static void model_update_cam_cdf(model_t *m)
{
  if(m->cam_lobes.n == 0)
  {
    m->p_uniform_cam = 1.f;
    return;
  }
  m->p_uniform_cam = P_UNIFORM_FIRST;
  m->cam_lobe_cdf = realloc(m->cam_lobe_cdf, sizeof(float) * m->cam_lobes.n);
  double total = 0.;
  for(uint32_t i = 0, N = m->cam_lobes.n; i < N; i++)
  {
    float weight = m->cam_lobes.lobes[i].weight;
    m->cam_lobe_cdf[i] = weight;
    total += weight;
  }

  double prev_sum = 0.;
  for(uint32_t i = 0, N = m->cam_lobes.n; i < N; i++)
  {
    double p_this = m->cam_lobe_cdf[i] / total;
    m->cam_lobe_cdf[i] = prev_sum + p_this;
    prev_sum += p_this;
  }
}

static void model_octree_fill_leaves_sample_counts(model_t *m)
{
  octree_t *octree = &m->octree;
  uint32_t *node_data = (uint32_t*) octree->node_data;
  for(uint32_t node_i = 0; node_i < octree->nodes.n; node_i++)
  {
    octree_idx_t node = octree->nodes.nodes[node_i];
    if(!node.leaf) continue;
    uint32_t cell_index = node.index;
    assert(cell_index < m->n_cells);
    node_data[node_i] = 0;
    for(int t = 0; t < rt.num_threads; t++)
      node_data[node_i] += m->reservoirs_tls[t][cell_index].samples_seen;
  }
}

static void model_octree_propagate_sample_counts(model_t *m)
{ // postorder traverse to get sample counts for inner nodes
  uint32_t *node_data = (uint32_t*) m->octree.node_data;
  octree_idx_t *nodes = m->octree.nodes.nodes;
  typedef struct stack_item_t {
    uint32_t ready : 1;
    uint32_t index : 31;
  } stack_item_t;
  const int max_stacklen = 8 * (OCTREE_MAX_DEPTH+1);
  stack_item_t stack[max_stacklen];
  uint32_t stack_len = 0;
  for(uint32_t index = 0; index < 8; index++)
    stack[stack_len++] = (stack_item_t) { .ready = false, .index = index};
  while(stack_len > 0)
  {
    stack_item_t node_index = stack[--stack_len];
    octree_idx_t node = nodes[node_index.index];
    if(node.leaf) continue;
    if(!node_index.ready)
    {
      stack[stack_len++].ready = true;
      for(uint32_t offset = 0; offset < 8; offset++)
        stack[stack_len++] = (stack_item_t) { .ready = false, .index = node.index + offset};
      assert(stack_len < max_stacklen);
    }
    else
    {
      assert(node_data[node_index.index] == 0);
      for(uint32_t offset = 0; offset < 8; offset++)
        node_data[node_index.index] += node_data[node.index + offset];
    }
  }
}

static void model_octree_adapt_collapse(model_t *m, uint32_t threshold)
{
  uint32_t *node_data = (uint32_t*) m->octree.node_data;
  octree_nodes_t *nodes = &m->octree.nodes;
  uint32_t N_before = nodes->n;

  typedef struct stack_item_t {
    float aabb[6];
    uint32_t index;
  } stack_item_t;
  stack_item_t stack[8 * OCTREE_MAX_DEPTH];
  uint32_t stack_len = 1;
  stack[0].index = 0;
  memcpy(stack[0].aabb, m->octree.aabb, sizeof(float) * 6);
  while(stack_len > 0)
  {
    stack_item_t parent = stack[--stack_len];
    for(uint32_t offset = 0; offset < 8; offset++)
    {
      uint32_t node_index = parent.index + offset;
      assert(node_index < N_before);
      octree_idx_t node = nodes->nodes[node_index];
      if(node.leaf) continue;
      float node_aabb[6];
      memcpy(node_aabb, parent.aabb, sizeof(float) * 6);
      octree_compute_child_aabb(node_aabb, offset);
      uint32_t sample_count = node_data[node_index];
      if(sample_count > threshold)
      { // push to traverse
        stack[stack_len].index = node.index;
        memcpy(stack[stack_len].aabb, node_aabb, sizeof(float) * 6);
        stack_len++;
        assert(stack_len < 8 * OCTREE_MAX_DEPTH);
      }
      else model_octree_collapse_node(m, node_index, node_aabb);
    }
  }

  model_octree_adapt_finalize(m);
}

static void model_octree_adapt_subdivide(model_t *m, uint32_t threshold)
{
  uint32_t *node_data = (uint32_t*) m->octree.node_data;
  for(uint32_t node_i = 0, N_before = m->octree.nodes.n; node_i < N_before; node_i++)
  {
    octree_idx_t node = m->octree.nodes.nodes[node_i];
    if(!node.leaf) continue;
    uint32_t sample_count = node_data[node_i];
    if(sample_count > threshold)
    {
      float node_aabb[6];
      const float *pos = m->cells[node.index].octree_pos;
      uint32_t idx = octree_lookup_aabb(&m->octree, pos, node_aabb);
      uint32_t depth = octree_lookup_depth(&m->octree, pos);
      assert(idx == node_i);
      if(depth < OCTREE_MAX_DEPTH) model_octree_subdivide_leaf(m, node_i, node_aabb);
    }
  }

  model_octree_adapt_finalize(m);
}

static void model_octree_adapt_finalize(model_t *m)
{
  octree_adapt_compact(&m->octree);

  for(uint32_t cell_i = 0; cell_i < m->n_cells; cell_i++)
  {
    if(!m->cells[cell_i].to_delete) continue;
    m->n_cells--;
    uint32_t last = m->n_cells;
    m->cells[cell_i] = m->cells[last];
    for(int t = 0; t < rt.num_threads; t++)
    {
      reservoir_cleanup(&m->reservoirs_tls[t][cell_i]);
      m->reservoirs_tls[t][cell_i] = m->reservoirs_tls[t][last];
    }
    cell_i--;
  }

  for(uint32_t i = 0; i < m->n_cells; i++)
    octree_set_prim(&m->octree, m->cells[i].octree_pos, i);

  for(uint32_t node_i = 0, N = m->octree.nodes.n; node_i < N; node_i++)
  {
    octree_idx_t node = m->octree.nodes.nodes[node_i];
    if(node.leaf) assert(node.index < m->n_cells);
  }
}

static void model_octree_subdivide_leaf(model_t *m, uint32_t node_index, const float node_aabb[6])
{
  octree_idx_t node = m->octree.nodes.nodes[node_index];
  assert(node.leaf == 1);
  uint32_t cell_index = node.index;

  octree_adapt_subdivide_cell(&m->octree, node_index);

  uint32_t cell_offset = m->n_cells;
  model_resize_cells(m, m->n_cells + 8);

  cell_mixture_t *old_cell = &m->cells[cell_index];
  old_cell->to_delete = true;
  for(uint32_t child_index = 0; child_index < 8; child_index++)
  {
    float child_aabb[6];
    memcpy(child_aabb, node_aabb, sizeof(child_aabb));
    octree_compute_child_aabb(child_aabb, child_index);
    model_init_cell(m, cell_offset + child_index, child_aabb);
    cell_mixture_t *new_cell = &m->cells[cell_offset + child_index];
    new_cell->n_lobes = old_cell->n_lobes;
    new_cell->lobes_offset = old_cell->lobes_offset;
  }

#if 0 // duplicate reservoirs
  for(int t = 0; t < rt.num_threads; t++)
  {
    const reservoir_t *reservoir = &m->reservoirs_tls[t][cell_index];
    // TODO reservoir_copy()?
    float equal_weight = reservoir->equal_weight;
    for(int child_offset = 0; child_offset < 8; child_offset++)
    {
      for(uint32_t sample_i = reservoir->samples_begin; sample_i < reservoir->capacity + 1; sample_i++)
      {
        reservoir_sample_t *sample = &reservoir->samples[sample_i];
        float xi[2];
        for(int k = 0; k < 2; k++) xi[k] = points_rand(rt.points, common_get_threadid());
        reservoir_push(&m->reservoirs_tls[t][cell_offset + child_offset], sample, equal_weight, xi);
      }
      const heap_t *heap = reservoir->heap;
      for(uint32_t heap_i = 0; heap_i < heap->end; heap_i++)
      {
        reservoir_sample_t *sample = &reservoir->samples[heap->keys[heap_i]];
        float weight = -heap->vals[heap_i];
        float xi[2];
        for(int k = 0; k < 2; k++) xi[k] = points_rand(rt.points, common_get_threadid());
        reservoir_push(&m->reservoirs_tls[t][cell_offset + child_offset], sample, weight, xi);
      }
    }
  }
#endif
  // split samples between reservoirs
  for(int t = 0; t < rt.num_threads; t++)
  {
    const reservoir_t *reservoir = &m->reservoirs_tls[t][cell_index];
    float equal_weight = reservoir->equal_weight;
    for(uint32_t sample_i = reservoir->samples_begin; sample_i < reservoir->capacity + 1; sample_i++)
    {
      reservoir_sample_t *sample = &reservoir->samples[sample_i];
      uint32_t child_offset = octree_get_child_offset(node_aabb, &sample->x[V2OFFSET]);
      float xi[2];
      for(int k = 0; k < 2; k++) xi[k] = points_rand(rt.points, common_get_threadid());
      reservoir_push(&m->reservoirs_tls[t][cell_offset + child_offset], sample, equal_weight, xi);
    }
    const heap_t *heap = reservoir->heap;
    for(uint32_t heap_i = 0; heap_i < heap->end; heap_i++)
    {
      reservoir_sample_t *sample = &reservoir->samples[heap->keys[heap_i]];
      float weight = -heap->vals[heap_i];
      uint32_t child_offset = octree_get_child_offset(node_aabb, &sample->x[V2OFFSET]);
      float xi[2];
      for(int k = 0; k < 2; k++) xi[k] = points_rand(rt.points, common_get_threadid());
      reservoir_push(&m->reservoirs_tls[t][cell_offset + child_offset], sample, weight, xi);
    }
  }
}

static void model_octree_collapse_node(model_t *m, uint32_t node_index, const float node_aabb[6])
{
  uint32_t new_cell_index = m->n_cells;
  model_resize_cells(m, m->n_cells + 1);
  model_init_cell(m, new_cell_index, node_aabb);

  const octree_idx_t *nodes = m->octree.nodes.nodes;
  uint32_t stack[8 * OCTREE_MAX_DEPTH];
  uint32_t stack_len = 1;
  stack[0] = node_index;
  while(stack_len > 0)
  {
    octree_idx_t node = nodes[stack[--stack_len]];
    if(node.leaf)
    {
      uint32_t cell_index = node.index;
      m->cells[cell_index].to_delete = true;

      for(int t = 0; t < rt.num_threads; t++)
      {
        reservoir_t *new_reservoir = &m->reservoirs_tls[t][new_cell_index];
        const reservoir_t *reservoir = &m->reservoirs_tls[t][cell_index];
        float equal_weight = reservoir->equal_weight;
        for(uint32_t sample_i = reservoir->samples_begin; sample_i < reservoir->capacity + 1; sample_i++)
        {
          reservoir_sample_t *sample = &reservoir->samples[sample_i];
          float xi[2];
          for(int k = 0; k < 2; k++) xi[k] = points_rand(rt.points, common_get_threadid());
          reservoir_push(new_reservoir, sample, equal_weight, xi);
        }
        const heap_t *heap = reservoir->heap;
        for(uint32_t heap_i = 0; heap_i < heap->end; heap_i++)
        {
          reservoir_sample_t *sample = &reservoir->samples[heap->keys[heap_i]];
          float weight = -heap->vals[heap_i];
          float xi[2];
          for(int k = 0; k < 2; k++) xi[k] = points_rand(rt.points, common_get_threadid());
          reservoir_push(new_reservoir, sample, weight, xi);
        }
      }
    }
    else for(uint32_t i = 0; i < 8; i++) stack[stack_len++] = node.index + i;
  }

  octree_adapt_collapse_cell(&m->octree, node_index);
}

static void *par_update_mixture(void *arg)
{
  model_t *m = arg;
  while(1)
  {
    uint32_t cell_index = __sync_fetch_and_add(&m->job, 1);
    if(cell_index >= m->num_jobs) return 0;
    model_update_mixture(m, cell_index);
  }
}

static void *par_reset_reservoirs(void *arg)
{
  model_t *m = arg;
  while(1)
  {
    uint32_t cell_index = __sync_fetch_and_add(&m->job, 1);
    if(cell_index >= m->num_jobs) return 0;

    for(int t = 0; t < rt.num_threads; t++)
    {
      reservoir_t *reservoir = &m->reservoirs_tls[t][cell_index];
#if RESERVOIR_CLEAR
      reservoir_clear(reservoir);
#else
      reservoir_decay(reservoir, RESERVOIR_DECAY);
#endif
    }
  }
}

static void model_update_mixture(model_t *m, uint32_t cell_index)
{
  model_update_tls_t *tls = &m->update_tls[common_get_threadid()];


  const float cell_size_factor = INIT_RADIUS;
  const cell_mixture_t *cell = &m->cells[cell_index];
  float prior_radius = cell->side_length * cell_size_factor;

  lobe_buffer_resize(&tls->lobe_buffer, 0);
  { // scene mixtures
    lobe_type_count_t sample_counts[N_LOBE_TYPES];
    uint32_t n_counts = model_count_sample_types(m, cell_index, sample_counts, false);

    for(uint32_t type_i = 0; type_i < n_counts; type_i++)
    {
      gmm_fit_t *gmm = &tls->gmm_fit;
      lobe_type_count_t type_count = sample_counts[type_i];
      assert(type_count.count > 0);
      uint32_t component_count = CLAMP(type_count.count, 1, GMM_MAX_COMPONENTS);
      gmm_fit_resize(gmm, component_count);
      uint32_t sample_count = model_extract_samples(m, cell_index, type_count.type, tls->sample_buffer, false);
      gmm_fit_samples(gmm, tls->sample_buffer, sample_count, prior_radius, m->stats);
#if ENABLE_GMM_REDUCTION
      {
        double weight_sum = 0;
        for(uint32_t comp_i = 0; comp_i < gmm->n_components; comp_i++)
          weight_sum += gmm->components[comp_i].weight;
        uint32_t n_before = gmm->n_components;
        gmm_fit_reduce(gmm, 0.6, weight_sum);
      }
#endif

      uint32_t lobes_offset = tls->lobe_buffer.n;
      lobe_buffer_resize(&tls->lobe_buffer, tls->lobe_buffer.n + gmm->n_components);
      for(uint32_t lobe_i = 0; lobe_i < gmm->n_components; lobe_i++)
      {
        lobe_t *lobe = &tls->lobe_buffer.lobes[lobes_offset];
        int err = lobe_from_comp(lobe, &gmm->components[lobe_i], type_count.type);
        if(err) lobe->weight = 0;
        if(lobe->weight > 0) lobes_offset++;
      }
      lobe_buffer_resize(&tls->lobe_buffer, lobes_offset);
    }

    threads_mutex_lock(&m->lobes_back_mutex);
    uint32_t offset = m->lobes_back.n;
    lobe_buffer_resize(&m->lobes_back, offset + tls->lobe_buffer.n);
    memcpy(&m->lobes_back.lobes[offset], tls->lobe_buffer.lobes, sizeof(lobe_t) * tls->lobe_buffer.n);
    threads_mutex_unlock(&m->lobes_back_mutex);
    m->cells[cell_index].n_lobes = tls->lobe_buffer.n;
    m->cells[cell_index].lobes_offset = offset;
  }

  { // camera mixtures
    lobe_type_count_t sample_counts[N_LOBE_TYPES];
    uint32_t n_counts = model_count_sample_types(m, cell_index, sample_counts, true);

    cam_lobe_buffer_resize(&tls->cam_lobe_buffer, 0);

    for(uint32_t type_i = 0; type_i < n_counts; type_i++)
    {
      gmm3_fit_t *gmm3 = &tls->gmm3_fit;
      lobe_type_count_t type_count = sample_counts[type_i];
      assert(type_count.count > 0);
      uint32_t sample_count = model_extract_samples(m, cell_index, type_count.type, tls->sample_buffer, true);

#if 0
      float cell_aabb[6];
      uint32_t node_index = octree_lookup_aabb(&m->octree, m->cells[cell_index].octree_pos, cell_aabb);
      assert(m->octree.nodes.nodes[node_index].index == cell_index);
      sample_count = em_sample_filter_aabb(tls->sample_buffer, sample_count, cell_aabb);
#endif
      uint32_t component_count = CLAMP(sample_count, 0, GMM_MAX_COMPONENTS);
      gmm3_fit_resize(gmm3, component_count);
      gmm3_fit_samples(gmm3, tls->sample_buffer, sample_count, prior_radius);
#if ENABLE_GMM_REDUCTION
      {
        double weight_sum = 0;
        for(uint32_t comp_i = 0; comp_i < gmm3->n_components; comp_i++)
          weight_sum += gmm3->components[comp_i].weight;
        gmm3_fit_reduce(gmm3, 0.3, weight_sum);
      }
#endif

      uint32_t cam_lobes_offset = tls->cam_lobe_buffer.n;
      cam_lobe_buffer_resize(&tls->cam_lobe_buffer, tls->cam_lobe_buffer.n + gmm3->n_components);
      for(uint32_t lobe_i = 0; lobe_i < gmm3->n_components; lobe_i++)
      {
        cam_lobe_t *cam_lobe = &tls->cam_lobe_buffer.lobes[cam_lobes_offset];
        int err = cam_lobe_from_comp(cam_lobe, &gmm3->components[lobe_i], type_count.type);
        if(err) cam_lobe->weight = 0;
        if(cam_lobe->weight > 0) cam_lobes_offset++;
      }
      cam_lobe_buffer_resize(&tls->cam_lobe_buffer, cam_lobes_offset);
    }

    threads_mutex_lock(&m->cam_lobes_back_mutex);
    uint32_t offset = m->cam_lobes_back.n;
    cam_lobe_buffer_resize(&m->cam_lobes_back, offset + tls->cam_lobe_buffer.n);
    memcpy(&m->cam_lobes_back.lobes[offset], tls->cam_lobe_buffer.lobes, sizeof(cam_lobe_t) * tls->cam_lobe_buffer.n);
    threads_mutex_unlock(&m->cam_lobes_back_mutex);
  }
}

static uint32_t model_extract_samples(const model_t *m, uint32_t cell_index, lobe_type_t type, em_sample_t *sample_buffer, bool camera)
{
  uint32_t n_samples = 0;
  for(int t = 0; t < rt.num_threads; t++)
  {
    const reservoir_t *reservoir = &m->reservoirs_tls[t][cell_index];
    float equal_weight = reservoir->equal_weight;
    for(uint32_t sample_i = reservoir->samples_begin; sample_i < reservoir->capacity + 1; sample_i++)
    {
      reservoir_sample_t *reservoir_sample = &reservoir->samples[sample_i];
      if(camera && (((type & t_vol_v2) | t_camera) != (reservoir_sample->type & (t_vol_v2 | t_camera)))) continue;
#if VERTEX_PAIRS
      if(!camera && (type != (reservoir_sample->type & ~t_camera))) continue;
#else
      if(!camera && (type != reservoir_sample->type)) continue;
#endif
      sample_buffer[n_samples].weight = equal_weight;
      memcpy(sample_buffer[n_samples].x, reservoir->samples[sample_i].x, sizeof(float) * GUIDE_DIM);
      n_samples++;
    }
    const heap_t *heap = reservoir->heap;
    for(uint32_t heap_i = 0; heap_i < heap->end; heap_i++)
    {
      reservoir_sample_t *reservoir_sample = &reservoir->samples[heap->keys[heap_i]];
      if(camera && (((type & t_vol_v2) | t_camera) != (reservoir_sample->type & (t_vol_v2 | t_camera)))) continue;
#if VERTEX_PAIRS
      if(!camera && (type != (reservoir_sample->type & ~t_camera))) continue;
#else
      if(!camera && (type != reservoir_sample->type)) continue;
#endif
      sample_buffer[n_samples].weight = -heap->vals[heap_i];
      memcpy(sample_buffer[n_samples].x, reservoir_sample->x, sizeof(float) * GUIDE_DIM);
      n_samples++;
    }
  }
  return n_samples;
}

static uint32_t model_count_sample_types(const model_t *m, uint32_t cell_index, lobe_type_count_t *type_counts, bool camera)
{
  uint32_t n_counts = 0;
  for(int t = 0; t < rt.num_threads; t++)
  {
    const reservoir_t *reservoir = &m->reservoirs_tls[t][cell_index];
    uint32_t sample_i = 0;
    if(sample_i == reservoir->heap->end) sample_i = reservoir->samples_begin;
    while(sample_i < reservoir->capacity + 1)
    {
      lobe_type_t type = reservoir->samples[sample_i].type;
      if(camera && !(type & t_camera)); // skip
      else
      {
#if VERTEX_PAIRS
        type &= ~t_camera; // do not care about this later
#endif
        if(camera) type &= t_vol_v2; // camera lobes care only about volume/not volume

        uint32_t type_index = 0;
        for(; type_index < n_counts; type_index++)
          if(type_counts[type_index].type == type) break;
        if(type_index < n_counts) type_counts[type_index].count++;
        else
        {
          assert(n_counts < N_LOBE_TYPES);
          type_counts[n_counts].type = type;
          type_counts[n_counts].count = 1;
          n_counts++;
        }
      }

      sample_i++;
      // skip empty middle part
      if(sample_i == reservoir->heap->end) sample_i = reservoir->samples_begin;
    }
  }
  return n_counts;
}

static void lobe_buffer_init(lobe_buffer_t *lobe_buffer)
{
  uint32_t initial_size = 32;
  *lobe_buffer = (lobe_buffer_t) {
    .lobes = malloc(sizeof(lobe_t) * initial_size),
    .n = 0,
    .allocated = initial_size,
  };
}

static void lobe_buffer_cleanup(lobe_buffer_t *lobe_buffer)
{
  free(lobe_buffer->lobes);
}

static void lobe_buffer_resize(lobe_buffer_t *lobe_buffer, uint32_t n_lobes)
{
  lobe_buffer->n = n_lobes;
  if(lobe_buffer->n <= lobe_buffer->allocated) return;
  while(lobe_buffer->allocated < lobe_buffer->n)
    lobe_buffer->allocated *= 2;
  lobe_buffer->lobes = realloc(lobe_buffer->lobes, sizeof(lobe_t) * lobe_buffer->allocated);
}

static void cam_lobe_buffer_init(cam_lobe_buffer_t *lobe_buffer)
{
  uint32_t initial_size = 32;
  *lobe_buffer = (cam_lobe_buffer_t) {
    .lobes = malloc(sizeof(cam_lobe_t) * initial_size),
    .n = 0,
    .allocated = initial_size,
  };
}

static void cam_lobe_buffer_cleanup(cam_lobe_buffer_t *lobe_buffer)
{
  free(lobe_buffer->lobes);
}

static void cam_lobe_buffer_resize(cam_lobe_buffer_t *lobe_buffer, uint32_t n_lobes)
{
  lobe_buffer->n = n_lobes;
  if(lobe_buffer->n <= lobe_buffer->allocated) return;
  while(lobe_buffer->allocated < lobe_buffer->n)
    lobe_buffer->allocated *= 2;
  lobe_buffer->lobes = realloc(lobe_buffer->lobes, sizeof(cam_lobe_t) * lobe_buffer->allocated);
}

static void model_resize_cells(model_t *m, uint32_t n_cells)
{
  m->n_cells = n_cells;
  if(m->n_cells <= m->cells_allocated) return;
  while(m->cells_allocated < m->n_cells)
    m->cells_allocated *= 2;
  m->cells = realloc(m->cells, sizeof(cell_mixture_t) * m->cells_allocated);
  for(int t = 0; t < rt.num_threads; t++)
    m->reservoirs_tls[t] = realloc(m->reservoirs_tls[t], sizeof(reservoir_t) * m->cells_allocated);
}

static float aabb_min_side(const float aabb[6])
{
  float min_side = aabb[3] - aabb[0];
  for(int k = 1; k < 3; k++) min_side = MIN(min_side, aabb[3+k] - aabb[k]);
  return min_side;
}

static void model_init_cell(model_t *m, uint32_t cell_index, const float aabb[6])
{
  cell_mixture_t *cell = &m->cells[cell_index];
  for(int k = 0; k < 3; k++)
    cell->octree_pos[k] = aabb[k] + 0.5f * (aabb[3+k] - aabb[k]);
  cell->side_length = aabb_min_side(aabb);
  cell->n_lobes = 0;
  cell->lobes_offset = 0;
  cell->to_delete = false;

  for(int t = 0; t < rt.num_threads; t++)
    reservoir_init(&m->reservoirs_tls[t][cell_index], RESERVOIR_SAMPLE_SIZE / rt.num_threads);
}
