#pragma once

#include "em.h"
#include "lobe.h"
#include "octree.h"
#include "threads.h"

typedef struct guided_bvh_t guided_bvh_t;
typedef struct reservoir_t reservoir_t;
typedef struct guided_stats_t guided_stats_t;

typedef struct cell_mixture_t
{
  uint32_t n_lobes;
  uint32_t lobes_offset;
  float side_length;
  // 3d position (center of cell?), so we know where to insert into octree
  float octree_pos[3];
  bool to_delete;
} cell_mixture_t;

typedef struct lobe_buffer_t
{
  lobe_t *lobes;
  uint32_t n;
  uint32_t allocated;
} lobe_buffer_t;

typedef struct cam_lobe_buffer_t
{
  cam_lobe_t *lobes;
  uint32_t n;
  uint32_t allocated;
} cam_lobe_buffer_t;

typedef struct model_update_tls_t
{
  gmm_fit_t gmm_fit;
  gmm3_fit_t gmm3_fit;

  em_sample_t sample_buffer[RESERVOIR_SAMPLE_SIZE];

  // buffer lobes to append them in one contiguous block to global array
  lobe_buffer_t lobe_buffer;
  cam_lobe_buffer_t cam_lobe_buffer;
} model_update_tls_t;

typedef struct model_t
{
  lobe_buffer_t lobes; // global list of all lobes for sampling
  lobe_buffer_t lobes_back; // back buffer for update
  threads_mutex_t lobes_back_mutex; // used during update only because of possible reallocation

  // same as above for sampling after camera
  cam_lobe_buffer_t cam_lobes;
  cam_lobe_buffer_t cam_lobes_back;
  threads_mutex_t cam_lobes_back_mutex;
  float *cam_lobe_cdf;
  float p_uniform_cam;

  octree_t octree; // organize local mixture models (leaf indices point into arrays below)

  guided_bvh_t *bvh; // for camera lobes
  uint32_t *bvh_prims;

  // leaf cell data for the octree:
  // - one mixture with information about lobes for sampling (indices point into lobes array above)
  // - one reservoir (duplicated across threads) for storing samples
  uint32_t n_cells;
  uint32_t cells_allocated;
  cell_mixture_t *cells;
  reservoir_t **reservoirs_tls; // points to arrays of reservoirs, one for each thread

  model_update_tls_t *update_tls; // per-thread storage used during updates

  guided_stats_t *stats;

  // for synchronization of thread pool
  uint32_t job;
  uint32_t num_jobs;
} model_t;

void model_init(model_t *m, guided_stats_t *stats);
void model_cleanup(model_t *m);

void model_update(model_t *m);

void model_prepare_frame(model_t *m);

uint32_t model_fill_lobe_list(model_t *m, const float xpp[3], const float xp[3], uint32_t *lobe_list, uint32_t n_max);
uint32_t model_fill_lobe_list_camera(model_t *m, const float xp[3], uint32_t *lobe_list, uint32_t n_max);

void model_serialize_debug(const model_t *m, FILE *f);

void model_serialize_samples_debug(const model_t *m, FILE *f);
void model_serialize_debug(const model_t *m, FILE *f);


typedef struct model_memsize_t
{
    uint64_t guiding;
    uint64_t samples;
} model_memsize_t;

void model_get_memsize(const model_t *m, model_memsize_t *mem_out);
