#include "octree.h"

#include <assert.h>
#include <stdlib.h>
#include <string.h>

static void octree_resize_nodes(octree_nodes_t *nodes, uint32_t new_n_nodes);

int octree_init(octree_t *octree, const float aabb[6], uint32_t node_data_size)
{
  memcpy(octree->aabb, aabb, sizeof(octree->aabb));
  octree->node_data_size = node_data_size;

  // initialize octree with 8 leaves and no primitives
  octree->nodes.n = 8;
  octree->nodes.alloc = 8;
  octree->nodes.nodes = malloc(sizeof(octree_idx_t) * octree->nodes.alloc);
  for(int i = 0; i < 8; i++) octree->nodes.nodes[i] = (octree_idx_t) { .leaf = 1, .index = i };

  if(node_data_size != 0)
    octree->node_data = calloc(octree->nodes.n, node_data_size);
  else octree->node_data = NULL;

  // allocate something here so we do not need to check for alloc == 0 later
  octree->nodes_back.n = 0;
  octree->nodes_back.alloc = 8;
  octree->nodes_back.nodes = malloc(sizeof(octree_idx_t) * octree->nodes_back.alloc);

  return 0;
}

void octree_cleanup(octree_t *octree)
{
  free(octree->nodes.nodes);
  free(octree->node_data);
  free(octree->nodes_back.nodes);
}

// TODO handle accesses outside of aabb. right now clamps implicitly?
uint32_t octree_lookup(const octree_t *octree, const float pos[3])
{
  float cell_aabb[6];
  return octree_lookup_aabb(octree, pos, cell_aabb);
}

uint32_t octree_lookup_aabb(const octree_t *octree, const float pos[3], float cell_aabb[6])
{
  memcpy(cell_aabb, octree->aabb, sizeof(float) * 6);
  octree_idx_t index = {0};
  assert(octree->nodes.n >= 8); // we don't have empty octrees
  while(1)
  {
    uint32_t child_offset = octree_get_child_offset(cell_aabb, pos);
    octree_compute_child_aabb(cell_aabb, child_offset);
    uint32_t child_index = index.index + child_offset;
    assert(child_index < octree->nodes.n);
    index = octree->nodes.nodes[child_index];
    if(index.leaf) return child_index;
  }
}

uint32_t octree_lookup_depth(const octree_t *octree, const float pos[3])
{
  float cell_aabb[6];
  memcpy(cell_aabb, octree->aabb, sizeof(float) * 6);
  octree_idx_t index = {0};
  assert(octree->nodes.n >= 8); // we don't have empty octrees
  uint32_t depth = 1;
  while(1)
  {
    uint32_t child_offset = octree_get_child_offset(cell_aabb, pos);
    octree_compute_child_aabb(cell_aabb, child_offset);
    uint32_t child_index = index.index + child_offset;
    assert(child_index < octree->nodes.n);
    index = octree->nodes.nodes[child_index];
    if(index.leaf) return depth;
    depth++;
  }
}

uint32_t octree_lookup_prim(const octree_t *octree, const float pos[3])
{
  uint32_t index = octree_lookup(octree, pos);
  octree_idx_t leaf_node = octree->nodes.nodes[index];
  assert(leaf_node.leaf == 1);
  return leaf_node.index;
}

void octree_set_prim(const octree_t *octree, const float pos[3], uint32_t prim)
{
  uint32_t index = octree_lookup(octree, pos);
  octree_idx_t *leaf_node = &octree->nodes.nodes[index];
  assert(leaf_node->leaf == 1);
  leaf_node->index = prim;
}

uint32_t octree_get_child_offset(const float cell_aabb[6], const float pos[3])
{
  uint32_t offset = 0;
  for(uint32_t k = 0; k < 3; k++)
  {
    float split = cell_aabb[k] + 0.5f * (cell_aabb[3+k] - cell_aabb[k]);
    offset |= (split <= pos[k]) << k;
  }
  return offset;
}

void octree_compute_child_aabb(float cell_aabb[6], uint32_t offset)
{
  for(uint32_t k = 0; k < 3; k++)
  {
    float split = cell_aabb[k] + 0.5 * (cell_aabb[3+k] - cell_aabb[k]);
    if(offset & (1 << k)) cell_aabb[k] = split;
    else cell_aabb[3+k] = split;
  }
}

//--- adaptation ---------------------------------------------------------------

void octree_adapt_subdivide_cell(octree_t *octree, uint32_t node_index)
{
  assert(node_index < octree->nodes.n);
  octree_idx_t *node = &octree->nodes.nodes[node_index];
  assert(node->leaf == 1);
  uint32_t new_index = octree->nodes.n;
  *node = (octree_idx_t) {
    .leaf = 0,
    .index = new_index,
  };
  octree_resize_nodes(&octree->nodes, octree->nodes.n + 8);
  for(int i = 0; i < 8; i++)
    octree->nodes.nodes[new_index + i] = (octree_idx_t) {
      .leaf = 1,
      .index = 0x7fffffff,
    };
}

void octree_adapt_collapse_cell(octree_t *octree, uint32_t node_index)
{
  assert(node_index < octree->nodes.n);
  octree_idx_t *node = &octree->nodes.nodes[node_index];
  assert(node->leaf == 0);
  *node = (octree_idx_t) {
    .leaf = 1,
    .index = 0x7fffffff,
  };
  // child nodes will become unreachable and get cleaned up by octree_adapt_compact call
}

void octree_adapt_compact(octree_t *octree)
{
  // dfs traversal copying the octree to back buffer, compacting on the way
  // (unreachable nodes do not get copied)
 
  struct stack_item_t {
    uint32_t old_offset;
    uint32_t new_offset;
  } stack[OCTREE_MAX_DEPTH * 8];
  stack[0] = (struct stack_item_t) { .old_offset = 0, .new_offset = 0 };
  uint32_t stack_len = 1;
  octree_resize_nodes(&octree->nodes_back, 8);

  while(stack_len > 0)
  {
    stack_len--;
    uint32_t old_offset = stack[stack_len].old_offset;
    uint32_t new_offset = stack[stack_len].new_offset;
    assert(old_offset + 8 <= octree->nodes.n);
    assert(new_offset + 8 <= octree->nodes_back.n);
    for(uint32_t child = 0; child < 8; child++)
    {
      octree_idx_t *old_node = &octree->nodes.nodes[old_offset + child];
      octree_idx_t *new_node = &octree->nodes_back.nodes[new_offset + child];
      if(old_node->leaf) *new_node = *old_node;
      else
      { // reserve node space and push to stack
        uint32_t new_index = octree->nodes_back.n;
        *new_node = (octree_idx_t) {
          .leaf = 0,
          .index = new_index,
        };
        octree_resize_nodes(&octree->nodes_back, new_index + 8);
        assert(stack_len < OCTREE_MAX_DEPTH * 8);
        stack[stack_len++] = (struct stack_item_t) {
          .old_offset = old_node->index,
          .new_offset = new_index,
        };
      }
    }
  }

  { // ping-pong buffers
    octree_nodes_t nodes_tmp = octree->nodes_back;
    octree->nodes_back = octree->nodes;
    octree->nodes = nodes_tmp;
  }

  if(octree->node_data_size != 0)
  {
    free(octree->node_data);
    octree->node_data = calloc(octree->nodes.n, octree->node_data_size);
  }
}


uint64_t octree_get_size(const octree_t *octree)
{
  uint64_t total = 0;
  total += sizeof(octree_t);
  total += sizeof(octree_idx_t) * octree->nodes.n;
  total += sizeof(octree_idx_t) * octree->nodes_back.n;
  total += octree->node_data_size * octree->nodes.n;
  return total;
}

//--- static functions ---------------------------------------------------------

static void octree_resize_nodes(octree_nodes_t *nodes, uint32_t new_n_nodes)
{
  while(nodes->alloc < new_n_nodes) nodes->alloc *= 2;
  nodes->nodes = realloc(nodes->nodes, sizeof(octree_idx_t) * nodes->alloc);
  nodes->n = new_n_nodes;
}
