#pragma once

#include <stdbool.h>
#include <stdint.h>

typedef struct octree_idx_t
{
  uint32_t leaf : 1;
  uint32_t index : 31; // either index into nodes array (leaf == 0) or "payload" set from outside (leaf == 1)
} octree_idx_t;

typedef struct octree_nodes_t
{
  uint32_t n; // length of nodes array
  uint32_t alloc;
  octree_idx_t *nodes; // groups of 8 consecutive child nodes (order?)
} octree_nodes_t;

typedef struct octree_t
{
  float aabb[6]; // [x_min, y_min, z_min, x_max, y_max, z_max]
  // current octree:
  octree_nodes_t nodes;
  uint32_t node_data_size;
  void *node_data;
  // back buffer used for adaptation only:
  octree_nodes_t nodes_back;
} octree_t;

#define OCTREE_MAX_DEPTH 20

int octree_init(octree_t *octree, const float aabb[6], uint32_t node_data_size);
void octree_cleanup(octree_t *octree);

// return index of leaf node in nodes array
uint32_t octree_lookup(const octree_t *octree, const float pos[3]);

uint32_t octree_lookup_aabb(const octree_t *octree, const float pos[3], float aabb[6]);

uint32_t octree_lookup_depth(const octree_t *octree, const float pos[3]);

// return stored prim index of leaf
uint32_t octree_lookup_prim(const octree_t *octree, const float pos[3]);

// store prim index in leaf
void octree_set_prim(const octree_t *octree, const float pos[3], uint32_t prim);

uint32_t octree_get_child_offset(const float cell_aabb[6], const float pos[3]);

void octree_compute_child_aabb(float cell_aabb[6], uint32_t offset);

//--- adaptation ---------------------------------------------------------------

// split an octree leaf into 8 new leaves
// node: index of the leaf node in nodes array (as returned by octree_lookup)
// new nodes have no node data until call to compact!
void octree_adapt_subdivide_cell(octree_t *octree, uint32_t node_index);

// collapse an octree cell and make it a leaf
// node: index of the pointer to the beginning of that node
// new node has no node data until call to compact!
void octree_adapt_collapse_cell(octree_t *octree, uint32_t node_index);

// do a compacting copy of the octree (collapsing will create holes)
// this will also recreate the node_data array
void octree_adapt_compact(octree_t *octree);

uint64_t octree_get_size(const octree_t *octree);
