#include "config.h"
#include "pixel_est.h"

#include "corona_common.h"
#include "screenshot.h"
#include "view.h"
#include <assert.h>
#include <string.h>

#define NUM_DBOR_LEVEL 20

typedef struct pixel_est_t
{
  int num_levels;
  int wd, ht;
  float *dbor; // dbor cascade
#ifdef DETERMINISTIC
  float **dbor_tls;
#endif
  float *mip_val; // mipmap pyramid for pixel values
  float *mip_n; // mipmap pyramid for sample count
  int n_mip;
  int mip_off[]; // indices into mip for each mip level
} pixel_est_t;

pixel_est_t *pixel_est_init(int width, int height, int num_mip_level)
{
  pixel_est_t *px_est = malloc(sizeof(pixel_est_t) + sizeof(int) * num_mip_level);
  px_est->wd = width;
  px_est->ht = height;
  px_est->dbor = calloc(NUM_DBOR_LEVEL * width * height, sizeof(float));
  px_est->n_mip = num_mip_level;

  assert(px_est->wd % (1 << (num_mip_level - 1)) == 0);
  assert(px_est->ht % (1 << (num_mip_level - 1)) == 0);

  int total_mip_size = 0;
  for(int i = 0; i < px_est->n_mip; i++)
  {
    px_est->mip_off[i] = total_mip_size;
    total_mip_size += (px_est->wd / (1 << i)) * (px_est->ht / (1 << i));
  }
  px_est->mip_val = malloc(sizeof(float) * total_mip_size);
  px_est->mip_n = malloc(sizeof(float) * total_mip_size);

#ifdef DETERMINISTIC
  px_est->dbor_tls = malloc(sizeof(float*) * rt.num_threads);
  for(int t = 0; t < rt.num_threads; t++)
  {
    px_est->dbor_tls[t] = calloc(NUM_DBOR_LEVEL * width * height, sizeof(float));
  }
#endif
  return px_est;
}

void pixel_est_cleanup(pixel_est_t *px_est)
{
#ifdef DETERMINISTIC
  for(int t = 0; t < rt.num_threads; t++)
  {
    free(px_est->dbor_tls[t]);
  }
  free(px_est->dbor_tls);
#endif
  free(px_est->dbor);
  free(px_est->mip_val);
  free(px_est->mip_n);
  free(px_est);
}

#if 0
// bilinear splat
static inline void filter_splat(
    float *buf,
    const float i,
    const float j,
    const float val,
    const int wd,
    const int ht)
{
  const int ii = (int)i;
  const int jj = (int)j;
  const float fi = i - ii;
  const float fj = j - jj;
  const int iip = ii+1, jjp = jj+1;
  float *p = buf + ii+wd*jj;
  if(ii  >= 0 && jj  >= 0 && ii  < wd && jj  < ht)
    common_atomic_add(p     , val * (1.0f-fi)*(1.0f-fj));
  if(iip >= 0 && jj  >= 0 && iip < wd && jj  < ht)
    common_atomic_add(p+1   , val * (     fi)*(1.0f-fj));
  if(ii  >= 0 && jjp >= 0 && ii  < wd && jjp < ht)
    common_atomic_add(p  +wd, val * (1.0f-fi)*(     fj));
  if(iip >= 0 && jjp >= 0 && iip < wd && jjp < ht)
    common_atomic_add(p+1+wd, val * (     fi)*(     fj));
}
#else
// blackman-harris splat
// (copied from blackmanharris.h because of incompatible api)
static inline float filter_bh_w(const float n)
{
  const float NN = 4.0f;
  if(n > NN-1.0f || n < 0.0f) return 0.0f;
  const float a0 = 0.35875;
  const float a1 = 0.48829;
  const float a2 = 0.14128;
  const float a3 = 0.01168;
  const float N_1 = 1.0f/(NN-1.0f);
  const float cos1 = cosf(2.0f*M_PI*n*N_1);
  const float cos2 = cosf(4.0f*M_PI*n*N_1);
  const float cos3 = cosf(6.0f*M_PI*n*N_1);
  return a0 - a1*cos1 + a2*cos2 - a3*cos3;
}

static inline void filter_splat(
    float *buf,
    const float i,
    const float j,
    const float val,
    const int wd,
    const int ht)
{
  const int x0 = (int)(i - 1.5f);
  const int y0 = (int)(j - 1.5f);
  const int u0 = x0 >= 0 ? 0 : - x0;
  const int v0 = y0 >= 0 ? 0 : - y0;
  const int u4 = x0 + 4 > wd ? wd - x0 : 4;
  const int v4 = y0 + 4 > ht ? ht - y0 : 4;

  float weight = 0.0f;
  for(int v=v0;v<v4;v++) for(int u=u0;u<u4;u++)
  {
    const float uu = (x0 + u + .5f) - i, vv = (y0 + v + .5f) - j;
    const float r = sqrtf(uu*uu + vv*vv);
    const float f = filter_bh_w(r+1.5f);
    weight += f;
  }
  if(weight <= 0) return;
  weight = 1.0f/weight;
  for(int v=v0;v<v4;v++) for(int u=u0;u<u4;u++)
  {
    const float uu = (x0 + u + .5f) - i, vv = (y0 + v + .5f) - j;
    const float r = sqrtf(uu*uu + vv*vv);
    const float f = weight * filter_bh_w(r+1.5f);
    common_atomic_add(&buf[(y0 + v) * wd + (x0 + u)], val * f);
  }
}
#endif

void pixel_est_splat(pixel_est_t *px_est, float x, float y, float throughput)
{
  if (throughput > (1 << NUM_DBOR_LEVEL)) return;
  assert(throughput > 0.f);
  const float logval = MAX(0, log2f(throughput));
  const int l = CLAMP((int)logval, 0, NUM_DBOR_LEVEL-1);
  const int u = CLAMP(l+1, 0, NUM_DBOR_LEVEL-1);
  const float lv = (l == NUM_DBOR_LEVEL-1) || throughput < 1.f ? 1.f :
    (((1<<l)/throughput)-0.5f)/0.5f;
  const float uv = 1.f - lv;
  int wd = px_est->wd;
  int ht = px_est->ht;
#ifdef DETERMINISTIC
  int t = common_get_threadid();
  filter_splat(px_est->dbor_tls[t] + l * wd * ht, x, y, lv * throughput, wd, ht);
  filter_splat(px_est->dbor_tls[t] + u * wd * ht, x, y, uv * throughput, wd, ht);
#else
  filter_splat(px_est->dbor + l * wd * ht, x, y, lv * throughput, wd, ht);
  filter_splat(px_est->dbor + u * wd * ht, x, y, uv * throughput, wd, ht);
#endif
}

static void fill_base_mip(pixel_est_t *px_est, int num_samples);
static void compute_mip_pyramid(pixel_est_t *px_est);
static void propagate_coarse_to_fine(pixel_est_t *px_est, int n_min);

void pixel_est_estimate(pixel_est_t *px_est, int num_samples, float n_min, float *out_img)
{
  fill_base_mip(px_est, num_samples);
#if WRITE_DEBUG_INFO
  int write_screenshot = (((unsigned int) num_samples) & ((unsigned int) num_samples - 1)) == 0;
  if(write_screenshot)
  {
    char filename[1024];
    snprintf(filename, sizeof(filename), "dbor_out_%05d", num_samples);
    screenshot_write(filename, px_est->mip_val, 1, 0, 1, view_width(), view_height(), 1.f);
  }
#endif
  compute_mip_pyramid(px_est);
  propagate_coarse_to_fine(px_est, n_min);
  // copy base level to output:
  memcpy(out_img, px_est->mip_val, sizeof(float) * px_est->wd * px_est->ht); 
}

static void fill_base_mip(pixel_est_t *px_est, int num_samples)
{
#ifdef DETERMINISTIC
  {
    int end = NUM_DBOR_LEVEL * px_est->wd * px_est->ht;
    for(int i = 0; i < end; i++)
    {
      px_est->dbor[i] = 0;
      for(int t = 0; t < rt.num_threads; t++) px_est->dbor[i] += px_est->dbor_tls[t][i];
    }
  }
#endif

  float kappa_min = 1.f;
  float kappa = 4.f;
  int wd = px_est->wd;
  int ht = px_est->ht;
#pragma omp parallel for
  for(int y = 0; y < ht; y++)
  {
    for(int x = 0; x < wd; x++)
    { // filter outliers with dbor
      float value = px_est->dbor[y * wd + x];
      float n_splats = value;
      for(int level = 1; level < NUM_DBOR_LEVEL; level++)
      {
        int level_below = MAX(level - 1, 0);
        int level_above = MIN(level + 1, NUM_DBOR_LEVEL - 1);
        float n_level = 0.;
        float n_average = 0.;
        int y_min = MAX(y - 1, 0);
        int x_min = MAX(x - 1, 0);
        int y_max = MIN(y + 1, ht - 1);
        int x_max = MIN(x + 1, wd - 1);
        for(int k = level_below; k <= level_above; k++)
        {
          float *level = &px_est->dbor[k * wd * ht];
          float level_norm = 1.f/((float) (1 << k));
          for(int y_ = y_min; y_ <= y_max; y_++) for(int x_ = x_min; x_ <= x_max; x_++)
          {
            float n_ = level[y_ * wd + x_] * level_norm;
            if(x_ == x && y_ == y) n_level += n_;
            n_average += n_;
          }
        }
        float w = (x_max - x_min + 1) * (y_max - y_min + 1);
        n_average /= w;
        if(n_average <= kappa_min || n_level <= kappa_min) continue;

        float weight = n_level < (kappa + kappa_min) ? (n_level - kappa_min)/kappa : 1.f;
        value += weight * px_est->dbor[level * wd * ht + y * wd + x];
        n_splats += weight * px_est->dbor[level * wd * ht + y * wd + x] / (1 << level);
      }
      px_est->mip_val[y * wd + x] = value / num_samples;
      px_est->mip_n[y * wd + x] = n_splats;
    }
  }
}

static void compute_mip_pyramid(pixel_est_t *px_est)
{
  const float wg[5] = {1.0, 4.0, 6.0, 4.0, 1.0}; // filter kernel
  int wd = px_est->wd;
  int ht = px_est->ht;
  for(int level = 1; level < px_est->n_mip; level++)
  {
    int new_wd = wd / (1 << level);
    int new_ht = ht / (1 << level);
    int offset_prev = px_est->mip_off[level - 1];
    int offset_next = px_est->mip_off[level];
#pragma omp parallel for
    for(int y = 0; y < new_ht; y++) for(int x = 0; x < new_wd; x++)
    { // downscaled image
      float sum_n = 0;
      float sum_px_val = 0;
      int wd_prev = wd / (1 << (level - 1));
      int ht_prev = ht / (1 << (level - 1));
      for(int by = 0; by < 2; by++) for(int bx = 0; bx < 2; bx++)
      { // 2x2 block in higher res image
        float weight = 0.f;
        float n = 0.f;
        float px_val = 0.f;
        for(int fy = -2; fy <= 2; fy++) for(int fx = -2; fx <= 2; fx++)
        { // 5x5 gauss filter
          int x_ = 2 * x + bx + fx;
          int y_ = 2 * y + by + fy;
          if(x_ < 0 || x_ >= wd_prev) continue;
          if(y_ < 0 || y_ >= ht_prev) continue;
          float w = wg[2 + fy] * wg[2 + fx];
          weight += w;
          int index = offset_prev + y_ * wd_prev + x_;
          n += w * px_est->mip_n[index];
          px_val += w * px_est->mip_val[index];
        }
        sum_n += n / weight;
        sum_px_val += px_val / weight;
      }
      int pos = offset_next + y * new_wd + x;
      px_est->mip_n[pos] = sum_n;
      px_est->mip_val[pos] = sum_px_val * .25f;
    }
  }
}

static void propagate_coarse_to_fine(pixel_est_t *px_est, int n_min)
{
  int wd = px_est->wd;
  int ht = px_est->ht;
  for(int fine = px_est->n_mip - 2; fine >= 0; fine--)
  {
    int coarse = fine + 1;
    int wd_fine = wd / (1 << fine);
    int ht_fine = ht / (1 << fine);
    int wd_coarse = wd / (1 << coarse);
    int ht_coarse = ht / (1 << coarse);
    int off_fine = px_est->mip_off[fine];
    int off_coarse = px_est->mip_off[coarse];
#pragma omp parallel for
    for(int y = 0; y < ht_fine; y++) for(int x = 0; x < wd_fine; x++)
    {
      int index_fine = off_fine + y * wd_fine + x;
      float n_fine = px_est->mip_n[index_fine];
      if(n_fine >= n_min) continue;
      float val_fine = px_est->mip_val[index_fine];

      float w = 0.;
      float val_coarse = 0.;
      for(int by = -1; by <= 1; by++) for(int bx = -1; bx <= 1; bx++)
      {
          int x_ = (x + bx) / 2;
          int y_ = (y + by) / 2;
          if(x_ < 0 || x_ >= wd_coarse) continue;
          if(y_ < 0 || y_ >= ht_coarse) continue;
          int index = off_coarse + y_ * wd_coarse + x_;
          val_coarse += px_est->mip_val[index];
          w++;
      }
      val_coarse /= w;

      float t = (n_min - n_fine) / n_min;
      float val = (1.f - t) * val_fine + t * val_coarse;
      px_est->mip_val[off_fine + y * wd_fine + x] = val;
    }
  }
}
