#include "reservoir.h"

#include <stdbool.h>
#include <stdlib.h>

void reservoir_init(reservoir_t *r, size_t capacity)
{
  // we reserve one more slot for the new sample coming in to avoid special cases
  *r= (reservoir_t) {
    .heap = heap_init(capacity + 1),
    .samples = malloc(sizeof(reservoir_sample_t) * (capacity + 1)),
    .samples_begin = capacity + 1,
    .equal_weight = 0,
    .capacity = capacity,
    .samples_seen = 0,
  };
}

void reservoir_cleanup(reservoir_t *r)
{
  heap_cleanup(r->heap);
  free(r->samples);
}

void reservoir_clear(reservoir_t *reservoir)
{
  reservoir->samples_begin = reservoir->capacity + 1;
  reservoir->equal_weight = 0;
  heap_clear(reservoir->heap);
  reservoir->samples_seen = 0;
}

void reservoir_decay(reservoir_t *reservoir, float decay)
{
  reservoir->equal_weight *= decay;
  for(int i = 0; i < reservoir->heap->end; i++)
    reservoir->heap->vals[i] *= decay;
  reservoir->samples_seen = 0;
}

void reservoir_push(reservoir_t *r, const reservoir_sample_t *sample, float weight, float xi[2])
{
  if(!(weight > 0 && weight < FLT_MAX)) return;
  r->samples_seen++;

  uint32_t len_T = (r->capacity + 1) - r->samples_begin;
  if(len_T == 0 && r->heap->end < r->capacity)
  {
    r->samples[r->heap->end] = *sample;
    heap_insert(r->heap, r->heap->end, -weight);
    return;
  }

  assert(len_T + r->heap->end == r->capacity);
  double W = r->equal_weight * len_T;
  uint32_t len_X = 0; // store X simply at the end of the heap
  if(weight <= r->equal_weight)
  {
    assert(r->samples_begin > 0);
    len_X++;
    assert(r->samples_begin - len_X >= r->heap->end);
    r->samples[r->samples_begin - len_X] = *sample;
    assert(r->heap->end + len_X <= r->heap->size);
    r->heap->vals[r->heap->size - len_X] = weight;
    W += weight;
  }
  else
  {
    assert(r->heap->end < r->capacity + 1);
    assert(r->heap->end < r->samples_begin);
    r->samples[r->heap->end] = *sample;
    heap_insert(r->heap, r->heap->end, -weight);
  }

  while(!heap_empty(r->heap) && W >= (-heap_max_val(r->heap)) * ((int) (len_X + len_T) - 1))
  {
    uint64_t key;
    float val;
    heap_remove(r->heap, &key, &val);
    val = -val;

    reservoir_sample_t this_sample = r->samples[key];
    // swap end of heap samples to fill hole
    uint32_t end_pos = r->heap->end;
    if(end_pos > 0 && end_pos != key)
    {
      // TODO this needs to go through all elements to correct the index of the other element
      // use an adressable heap to make this faster
      r->samples[key] = r->samples[end_pos];
      bool found = false;
      for(uint32_t i = 0; i < r->heap->end; i++)
      {
        if(r->heap->keys[i] == end_pos)
        {
          r->heap->keys[i] = key;
          found = true;
          break;
        }
      }
      assert(found);
    }

    // append to X
    len_X++;
    assert(r->samples_begin - len_X >= r->heap->end);
    r->samples[r->samples_begin - len_X] = this_sample;
    assert(r->heap->end + len_X <= r->heap->size);
    r->heap->vals[r->heap->size - len_X] = val;
    W += val;
  }
  assert(((int) (len_X + len_T) - 1) > 0);

  r->equal_weight = W / (len_X + len_T - 1);
  double Wx = 0;
  bool deleted_from_X = false;
  float *weights_X = &r->heap->vals[r->heap->size - len_X];
  for(uint32_t d = 0; d < len_X; d++)
  {
    Wx += weights_X[d];
    if((d + 1) - Wx / r->equal_weight >= xi[0])
    { // delete element d from X
      r->samples[r->samples_begin - len_X + d] = r->samples[r->samples_begin - len_X];
      len_X--;
      deleted_from_X = true;
      break;
    }
  }
  if(!deleted_from_X && len_T > 0)
  {
    assert(len_T > 0);
    // delete random from T
    uint32_t d = (uint32_t) (xi[1] * len_T);
    assert(d >= 0 && d < len_T);
    r->samples[r->samples_begin + d] = r->samples[r->samples_begin - len_X];
    r->samples_begin++;
  }
  r->samples_begin -= len_X;
  assert(r->samples_begin > 0);
  assert(r->samples_begin <= r->capacity + 1);
}

uint32_t reservoir_stored_samples(const reservoir_t *r)
{
  return r->heap->end + (r->capacity + 1 - r->samples_begin);
}
