#include "config.h"
#include "stats.h"

#include "corona_common.h"
#include <assert.h>
#include <string.h>

#define NUM_PROJECT_ERRORS 16

typedef struct guided_stats_t
{
  uint64_t err_extend[NUM_EXTEND_ERRORS];
  uint64_t err_project[NUM_PROJECT_ERRORS];
  uint64_t num_paths_success;
  uint64_t num_paths_sampled;
  uint64_t num_paths_recorded;

  uint64_t lobes_found_sum[PATHSPACE_MAX_VERTS];
  uint64_t lobes_found_2_sum[PATHSPACE_MAX_VERTS];
  uint64_t n_lobes_found[PATHSPACE_MAX_VERTS];

  uint64_t lobes_filtered_sum[PATHSPACE_MAX_VERTS];
  uint64_t lobes_filtered_2_sum[PATHSPACE_MAX_VERTS];
  uint64_t n_lobes_filtered[PATHSPACE_MAX_VERTS];

  uint64_t lobes_used_sum[PATHSPACE_MAX_VERTS];
  uint64_t lobes_used_2_sum[PATHSPACE_MAX_VERTS];
  uint64_t n_lobes_used[PATHSPACE_MAX_VERTS];

  double em_time;
  double em_init_time;
  double em_init_fraction;
  double em_init_fraction2;
  uint64_t em_iter;
  uint64_t em_iter2;
  uint64_t n_em_records;
}
guided_stats_t;

guided_stats_t *stats_init(void)
{
  return calloc(rt.num_threads, sizeof(guided_stats_t));
}

void stats_cleanup(guided_stats_t *stats)
{
  free(stats);
}

static guided_stats_t* get_tls_stats(guided_stats_t *stats)
{
  return &stats[common_get_threadid()];
}

void stats_collect(guided_stats_t *stats_)
{
  guided_stats_t *stats_acc = stats_;
  for(int t = 1; t < rt.num_threads; t++)
  {
    guided_stats_t *stats = &stats_[t];
    for(int i = 0; i < NUM_EXTEND_ERRORS; i++)
    {
      stats_acc->err_extend[i] += stats->err_extend[i];
    }
    for(int i = 0; i < NUM_PROJECT_ERRORS; i++)
    {
      stats_acc->err_project[i] += stats->err_project[i];
    }
    stats_acc->num_paths_sampled += stats->num_paths_sampled;
    stats_acc->num_paths_success += stats->num_paths_success;
    stats_acc->num_paths_recorded += stats->num_paths_recorded;

    for(int i = 0; i < PATHSPACE_MAX_VERTS; i++)
    {
      stats_acc->lobes_found_sum[i] += stats->lobes_found_sum[i];
      stats_acc->lobes_found_2_sum[i] += stats->lobes_found_2_sum[i];
      stats_acc->n_lobes_found[i] += stats->n_lobes_found[i];

      stats_acc->lobes_used_sum[i] += stats->lobes_used_sum[i];
      stats_acc->lobes_used_2_sum[i] += stats->lobes_used_2_sum[i];
      stats_acc->n_lobes_used[i] += stats->n_lobes_used[i];

      stats_acc->lobes_filtered_sum[i] += stats->lobes_filtered_sum[i];
      stats_acc->lobes_filtered_2_sum[i] += stats->lobes_filtered_2_sum[i];
      stats_acc->n_lobes_filtered[i] += stats->n_lobes_filtered[i];
    }

    stats_acc->em_time += stats->em_time;
    stats_acc->em_init_time += stats->em_init_time;
    stats_acc->em_init_fraction += stats->em_init_fraction;
    stats_acc->em_init_fraction2 += stats->em_init_fraction2;
    stats_acc->em_iter += stats->em_iter;
    stats_acc->em_iter2 += stats->em_iter2;
    stats_acc->n_em_records += stats->n_em_records;
  }
}

void stats_print(guided_stats_t *stats, guided_cache_stats_t cache_stats, FILE *f)
{
  static const char *err_extend_str[NUM_EXTEND_ERRORS] = {
    "no error",
    "maximum path length reached",
    "uniform extend failed (first vertex)",
    "camera sampling failed",
    "uniform extend failed",
    "bsdf sampling failed",
    "distance sampling failed",
    "first vertex sampled outside lobe aabb",
    "wrong mode at sampled vertex",
    "wrong flags at sampled vertex",
    "bsdf is zero",
    "absorbing environment hit",
    "wrong mode at middle vertex",
    "sampled vertex has zero pdf (2d)",
    "sampled vertex has zero pdf (3d)",
    "no material to continue path"
  };

  uint64_t total_extend = 0;
  for(int i = 0; i < NUM_EXTEND_ERRORS; i++)
  {
    total_extend += stats->err_extend[i];
  }
  for(int i = 0; i < NUM_PROJECT_ERRORS; i++)
  {
    total_extend += stats->err_project[i];
  }
  uint64_t total_fail = total_extend - stats->err_extend[0];
  fprintf(f, "lobes in cache: %d camera, %d scene\n", cache_stats.num_lobes_v1, cache_stats.num_lobes_total);
  fprintf(f, "sampled %ld paths (%.2f%% miss)\n", stats->num_paths_success, (1. - (double) stats->num_paths_success / stats->num_paths_sampled) * 100.);
  fprintf(f, "guided extend fail stats (%.2f%% fail):\n", (double) total_fail / total_extend * 100.);
  for(int i = 1; i < NUM_EXTEND_ERRORS; i++)
  {
    if(stats->err_extend[i] == 0) continue;
    fprintf(f, "  %.2f%% %s\n", (double) stats->err_extend[i] / total_fail * 100., err_extend_str[i]);
  }
  for(int i = 1; i < NUM_PROJECT_ERRORS; i++)
  {
    if(stats->err_project[i] == 0) continue;
    fprintf(f, "  %.2f%% project error %d\n", (double) stats->err_project[i] / total_fail * 100., i);
  }
  fprintf(f, "\n");

  // em stats
  {
    double avg_em_time = stats->em_time / stats->n_em_records;
    double avg_init_time = stats->em_init_time / stats->n_em_records;
    double avg_iter = (double) stats->em_iter / stats->n_em_records; 
    double avg_iter2 = (double) stats->em_iter2 / stats->n_em_records;
    double stddev_iter = sqrt(MAX(avg_iter2 - avg_iter * avg_iter, 0.));
    double avg_fraction = stats->em_init_fraction / stats->n_em_records;
    double avg_fraction2 = stats->em_init_fraction2 / stats->n_em_records;
    double stddev_fraction = sqrt(MAX(avg_fraction2 - avg_fraction * avg_fraction, 0.));
    fprintf(f, "ran em %ld times in total\n", stats->n_em_records);
    fprintf(f, "avg time init %f em %f\n", avg_init_time, avg_em_time);
    fprintf(f, "avg init time %f%% (stddev %f%%) of total em time\n", 100. * avg_fraction, 100. * stddev_fraction);
    fprintf(f, "avg number of em iterations %f stddev %f\n", avg_iter, stddev_iter);
  }

  // mean and var lobes found
  fprintf(f, "\n");
  for(int i = 0; i < PATHSPACE_MAX_VERTS; i++)
  {
    uint64_t n = stats->n_lobes_found[i];
    uint64_t sum = stats->lobes_found_sum[i];
    uint64_t sum2 = stats->lobes_found_2_sum[i];
    double avg = n == 0 ? 0 : (double) sum / n;
    double stddev = n == 0 ? 0 : sqrt(MAX((double) sum2 / n - avg * avg, 0.));
    fprintf(f, "n_lobes[%d]: mean %f stddev %f (n = %ld)\n", i, avg, stddev, n);
  }

  fprintf(f, "\n");
}

void stats_print_json(guided_stats_t *stats, guided_cache_stats_t cache_stats, FILE *f)
{
  fprintf(f, "{\n");
  fprintf(f, "\t\"num_scene_lobes\": %d,\n", cache_stats.num_lobes_total);
  fprintf(f, "\t\"num_cam_lobes\": %d,\n", cache_stats.num_lobes_v1);

  double mean[PATHSPACE_MAX_VERTS], stddev[PATHSPACE_MAX_VERTS];
  for(int i = 0; i < PATHSPACE_MAX_VERTS; i++)
  {
    uint64_t n = stats->n_lobes_found[i];
    mean[i] = n == 0 ? 0 : (double) stats->lobes_found_sum[i] / n;
    stddev[i] = n == 0 ? 0 : sqrt(MAX((double) stats->lobes_found_2_sum[i] / n - mean[i] * mean[i], 0));
  }
  fprintf(f, "\t\"num_lobes_bvh_mean\": [%g", mean[0]);
  for(int i = 1; i < PATHSPACE_MAX_VERTS; i++) fprintf(f, ", %g", mean[i]);
  fprintf(f, "],\n");
  fprintf(f, "\t\"num_lobes_bvh_stddev\": [%g", stddev[0]);
  for(int i = 1; i < PATHSPACE_MAX_VERTS; i++) fprintf(f, ", %g", stddev[i]);
  fprintf(f, "],\n");

  for(int i = 0; i < PATHSPACE_MAX_VERTS; i++)
  {
    uint64_t n = stats->n_lobes_used[i];
    mean[i] = n == 0 ? 0 : (double) stats->lobes_used_sum[i] / n;
    stddev[i] = n == 0 ? 0 : sqrt(MAX((double) stats->lobes_used_2_sum[i] / n - mean[i] * mean[i], 0));
  }
  fprintf(f, "\t\"num_lobes_mean\": [%g", mean[0]);
  for(int i = 1; i < PATHSPACE_MAX_VERTS; i++) fprintf(f, ", %g", mean[i]);
  fprintf(f, "],\n");
  fprintf(f, "\t\"num_lobes_stddev\": [%g", stddev[0]);
  for(int i = 1; i < PATHSPACE_MAX_VERTS; i++) fprintf(f, ", %g", stddev[i]);
  fprintf(f, "],\n");

  for(int i = 0; i < PATHSPACE_MAX_VERTS; i++)
  {
    uint64_t n = stats->n_lobes_filtered[i];
    mean[i] = n == 0 ? 0 : (double) stats->lobes_filtered_sum[i] / n;
    stddev[i] = n == 0 ? 0 : sqrt(MAX((double) stats->lobes_filtered_2_sum[i] / n - mean[i] * mean[i], 0));
  }
  fprintf(f, "\t\"num_lobes_filtered_mean\": [%g", mean[0]);
  for(int i = 1; i < PATHSPACE_MAX_VERTS; i++) fprintf(f, ", %g", mean[i]);
  fprintf(f, "],\n");
  fprintf(f, "\t\"num_lobes_filtered_stddev\": [%g", stddev[0]);
  for(int i = 1; i < PATHSPACE_MAX_VERTS; i++) fprintf(f, ", %g", stddev[i]);
  fprintf(f, "],\n");

  fprintf(f, "\t\"num_paths_sampled\": %ld,\n", stats->num_paths_sampled);
  fprintf(f, "\t\"num_paths_success\": %ld,\n", stats->num_paths_success);
  fprintf(f, "}\n");
}

void stats_reset(guided_stats_t *stats)
{
  for(int t = 0; t < rt.num_threads; t++)
  {
    memset(&stats[t], 0, sizeof(guided_stats_t));
  }
}

void stats_record_lobes_found(guided_stats_t *stats, uint64_t n_lobes, int v)
{
  guided_stats_t *tls_stats = get_tls_stats(stats);
  tls_stats->lobes_found_sum[v] += n_lobes;
  tls_stats->lobes_found_2_sum[v] += n_lobes * n_lobes;
  tls_stats->n_lobes_found[v]++;
}

void stats_record_lobes_filtered(guided_stats_t *stats, uint64_t n_lobes, int v)
{
  guided_stats_t *tls_stats = get_tls_stats(stats);
  tls_stats->lobes_filtered_sum[v] += n_lobes;
  tls_stats->lobes_filtered_2_sum[v] += n_lobes * n_lobes;
  tls_stats->n_lobes_filtered[v]++;
}

void stats_record_lobes_used(guided_stats_t *stats, uint64_t n_lobes, int v)
{
  guided_stats_t *tls_stats = get_tls_stats(stats);
  tls_stats->lobes_used_sum[v] += n_lobes;
  tls_stats->lobes_used_2_sum[v] += n_lobes * n_lobes;
  tls_stats->n_lobes_used[v]++;
}

void stats_record_err_extend(guided_stats_t *stats, extend_error_t err)
{
  get_tls_stats(stats)->err_extend[err]++;
}

void stats_record_err_project(guided_stats_t *stats, int err)
{
  assert(err < NUM_PROJECT_ERRORS);
  get_tls_stats(stats)->err_project[err]++;
}

void stats_record_successful_path(guided_stats_t *stats)
{
  get_tls_stats(stats)->num_paths_success++;
}

void stats_record_sampled_path(guided_stats_t *stats)
{
  get_tls_stats(stats)->num_paths_sampled++;
}

void stats_record_em(guided_stats_t *stats, double start, double after_init, double after_em, int iterations)
{
  guided_stats_t *tls_stats = get_tls_stats(stats);
  tls_stats->em_time += after_em - after_init;
  tls_stats->em_init_time += after_init - start;
  double fraction = (after_init - start) / (after_em - start);
  tls_stats->em_init_fraction += fraction;
  tls_stats->em_init_fraction2 += fraction * fraction;
  tls_stats->em_iter += iterations;
  tls_stats->em_iter2 += iterations * iterations;
  tls_stats->n_em_records++;
}
