// include the c file since we want to test static functions too
#include "octree.c"
#include "unittest.h"

static void test_child_aabb(void)
{
  float aabb[] = {
    0.5, -5., 6.,
    8.5, -3., 7.,
  };
  octree_compute_child_aabb(aabb, 0);
  float expected_aabb[] = {
    0.5, -5., 6.,
    4.5, -4., 6.5,
  };
  EXPECT_ALL_CLOSE(expected_aabb, aabb, 6);
  PASS();
}

static void test_octree_lookup(void)
{
  octree_t octree;
  float aabb[] = {
    0., 1., -2.,
    1., 2., -1.
  };
  octree_init(&octree, aabb, 0);
  {
    float pos[] = {0.45, 1.6, -1.7};
    uint32_t index = octree_lookup(&octree, pos);
    uint32_t index2 = octree_get_child_offset(octree.aabb, pos);
    EXPECT(index == index2);
  }
  {
    float pos[] = {0.99, 1.0, -1.3};
    uint32_t index = octree_lookup(&octree, pos);
    uint32_t index2 = octree_get_child_offset(octree.aabb, pos);
    EXPECT(index == index2);
  }

  octree_cleanup(&octree);
  PASS();
}

static void test_octree_adapt_empty(void)
{
  octree_t octree;
  float aabb[] = {
    0., 0., 0.,
    1., 1., 1.,
  };
  octree_init(&octree, aabb, 0);
  for(int i = 0; i < 8; i++) octree_adapt_subdivide_cell(&octree, i);
  octree_adapt_compact(&octree);
  octree_cleanup(&octree);
  PASS();
}

static void test_octree_adapt_two_levels(void)
{
  octree_t octree;
  float aabb[] = {
    0., 0., 0.,
    1., 1., 1.,
  };
  octree_init(&octree, aabb, 0);
  for(int i = 0; i < 8 + 8 * 8; i++) octree_adapt_subdivide_cell(&octree, i);
  octree_adapt_compact(&octree);
  EXPECT(octree.nodes.n == 8 + 8 * 8 + 8 * 8 * 8);
  octree_cleanup(&octree);
  PASS();
}

static void test_octree_insert_prims(void)
{
  octree_t octree;
  float aabb[] = {
    0., 0., 0.,
    1., 1., 1.,
  };
  octree_init(&octree, aabb, 0);
  for(int i = 0; i < 8; i++) octree_adapt_subdivide_cell(&octree, i);
  srand(4);
  float pos[4 * 4 * 4 * 3];
  for(int i = 0; i < 4; i++) for(int j = 0; j < 4; j++) for(int k = 0; k < 4; k++)
  {
    uint32_t idx = (i * 4 + j) * 4 + k;
    float *p = &pos[idx * 3];
    p[0] = i * 0.25f;
    p[1] = j * 0.25f;
    p[2] = k * 0.25f;
    for(int k = 0; k < 3; k++) p[k] += (rand() / (double) RAND_MAX) * .125f;
    octree_set_prim(&octree, p, idx);
  }
  octree_adapt_compact(&octree);
  EXPECT(octree.nodes.n == 8 + 8 * 8);
  for(uint32_t i = 0; i < 4 * 4 * 4; i++)
  {
    uint32_t prim = octree_lookup_prim(&octree, &pos[i * 3]);
    EXPECT(prim == i);
  }
  PASS();
}

static void test_octree_adapt_twice(void)
{
  octree_t octree;
  float aabb[] = {
    0., 0., 0.,
    1., 1., 1.,
  };
  octree_init(&octree, aabb, 0);

  {
    for(int i = 0; i < 8; i++) octree_adapt_subdivide_cell(&octree, i);
    octree_adapt_compact(&octree);
    EXPECT(octree.nodes.n == 8 + 8 * 8);
  }
  {
    uint32_t leaf_indices[8 * 8];
    uint32_t leaf_indices_len = 0;
    for(uint32_t i = 0; i < octree.nodes.n; i++)
      if(octree.nodes.nodes[i].leaf)
        leaf_indices[leaf_indices_len++] = i;
    EXPECT(leaf_indices_len == 64);
    for(uint32_t i = 0; i < leaf_indices_len; i++)
      octree_adapt_subdivide_cell(&octree, leaf_indices[i]);
    octree_adapt_compact(&octree);
    EXPECT(octree.nodes.n == 8 + 8 * 8 + 8 * 8 * 8);
  }
  octree_cleanup(&octree);
  PASS();
}

int main(int argc, char *argv[])
{
  (void) argc, (void) argv;

  test_child_aabb();
  test_octree_lookup();
  test_octree_adapt_empty();
  test_octree_adapt_two_levels();
  test_octree_insert_prims();
  test_octree_adapt_twice();
  // TODO add tests for collapse, node data
  return 0;
}
