// adapted from the roughplastic bsdf in Mitsuba

/*
    Copyright (c) 2007-2014 by Wenzel Jakob and others.

    Mitsuba is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License Version 3
    as published by the Free Software Foundation.

    Mitsuba is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "corona_common.h"
#include "shader.h"
#include "spectrum.h"
#include "ggx.h"
#include "fresnel.h"

#include <stdio.h>
#include <math.h>
#include <assert.h>

// - does not support specular, clamps roughness to GLOSSY_THR
// - doesn't use path's eta, always uses 1 as outer ior

// roughness below which specular reflection would be used
#define GLOSSY_THR 1e-4f

typedef struct plastic_t
{
  float n_d;
  int nonlinear;
}
plastic_t;

// precomputed data for rough transmittance loaded from file
// (needs data/microfacet/ggx.dat from mitsuba in data/gxx.dat)
typedef struct rtrans_t
{
  float alpha_min;
  float alpha_max;
  float eta_min;
  float eta_max;
  size_t eta_samples;
  size_t alpha_samples;
  size_t theta_samples;

  float *data;
  float *diff_data;
} rtrans_t;
static rtrans_t rtrans;
static int rtrans_users = 0; // refcount

static int load_rtrans_data(const char *filename);
static float fresnel(float cosThetaI, float eta);
float rtrans_eval(float cosTheta, float alpha, float eta);
static float rtrans_eval_diff(float alpha, float eta);
static mf_t prob_specular(
    float cos_in,
    float alpha,
    float eta,
    mf_t glossy_reflectance,
    mf_t diffuse_reflectance);

int init(FILE *f, void **data)
{

  plastic_t *plastic = malloc(sizeof(plastic_t));
  plastic->nonlinear = 0;
  int i = fscanf(f, "%f %d", &plastic->n_d, &plastic->nonlinear);
  if(i < 1)
  {
    fprintf(stderr, "[plastic] could not parse all arguments! expecting: n_d [nonlinear]\n");
  }
  *data = plastic;
  int dreggn = 0;
  dreggn = fscanf(f, "%*[^\n]\n");
  if(dreggn == -1) return 1;

  if(rtrans_users == 0)
  {
    const char *filename = "data/ggx.dat";
    int err = load_rtrans_data(filename);
    if(err)
    {
      fprintf(stderr, "[plastic] could not load %s\n", filename);
      return 1;
    }
  }
  rtrans_users++;
  return 0;
}

void cleanup(void *data)
{
  rtrans_users--;
  if(rtrans_users == 0)
  {
    free(rtrans.data);
    free(rtrans.diff_data);
  }
  free(data);
}

float prepare(path_t *p, int v, void *data)
{
  p->v[v].material_modes = s_reflect | s_glossy;
  p->v[v].shading.roughness = MAX(p->v[v].shading.roughness, GLOSSY_THR);
  return 1.0f;
}

mf_t pdf(path_t *p, int e1, int v, int e2, void *data)
{
  if(!(p->v[v].mode & s_reflect)) return mf_set1(0.0f);
  float wi[3];
  float wo[3];
  const float *n = p->v[v].hit.n;
  if(e1 < e2)
  {
    for(int k=0;k<3;k++)
    {
      wi[k] = p->e[e1].omega[k];
      wo[k] = p->e[e2].omega[k];
    }
  }
  else
  {
    for(int k=0;k<3;k++)
    {
      wi[k] = -p->e[e1].omega[k];
      wo[k] = -p->e[e2].omega[k];
    }
  }
  // these should always be positive:
  const float cos_in  = -dotproduct(n, wi);
  const float cos_out =  dotproduct(n, wo);
  if(cos_in < 0.0f) return mf_set1(0.0f);
  if(cos_out < 0.0f) return mf_set1(0.0f);

  float h[3];
  for(int k=0;k<3;k++) h[k] = -wi[k] + wo[k];
  normalise(h);

  const float r = p->v[v].shading.roughness;
  float eta = ((plastic_t*) data)->n_d;
  mf_t probSpecular = prob_specular(cos_in, r, eta, p->v[v].shading.rg, p->v[v].shading.rd);
  float pdf_h = ggx_pdf_h(wi, h, n, r);
  const float cosr = -dotproduct(wi, h);
  // pdfs are in projected solid angle measure
  float specPdf = pdf_h / (4.0f * cosr * fabsf(cos_out));
  float diffPdf = 1.f/M_PI;
  mf_t spec = mf_mul(probSpecular, mf_set1(specPdf));
  mf_t diff = mf_mul(mf_sub(mf_set1(1.f), probSpecular), mf_set1(diffPdf));
  mf_t pdf = mf_add(spec, diff);

  // mask out for degenerate divisions etc:
  mf_t mask = mf_not(mf_gt(pdf, mf_set1(0.0f)));
  return mf_select(mf_set1(0.0f), pdf, mask);
}

mf_t brdf(path_t *p, int v, void *data)
{
  const float cos_in  = -dotproduct(p->v[v].hit.n, p->e[v].omega);
  const float cos_out =  dotproduct(p->v[v].hit.n, p->e[v+1].omega);
  if(cos_out <= 0.0f || cos_in <= 0.0f) return mf_set1(0.0f);

  const float r = p->v[v].shading.roughness;
  float eta = ((plastic_t*) data)->n_d;
  int nonlinear = ((plastic_t*) data)->nonlinear;

  mf_t glossy;
  { // glossy part
    float h[3];
    for(int k = 0; k < 3; k++) h[k] = -p->e[v].omega[k] + p->e[v+1].omega[k];
    normalise(h);

    const float DG1 = ggx_pdf_h(p->e[v].omega, h, p->v[v].hit.n, r);
    const float cosr = -dotproduct(h, p->e[v].omega);
    if(DG1 == 0.f || cosr < 0.f) glossy = mf_set1(0);
    else
    {
      const float G1 = ggx_shadowing_smith_G1(p->e[v+1].omega, p->v[v].hit.n, h, r);
      const mf_t R = mf_set1(fresnel(cosr, eta));
      glossy = mf_mul(mf_mul(p->v[v].shading.rg, R), mf_set1(DG1 * G1 / (4.0f*fabsf(cosr*cos_out))));
    }
  }
  mf_t diff;
  { // diffuse part
    float T12 = rtrans_eval(cos_in, r, eta);
    float T21 = rtrans_eval(cos_out, r, eta);
    float Fdr = 1. - rtrans_eval_diff(r, 1./eta);
    mf_t albedo = p->v[v].shading.rd;
    if(nonlinear)
      diff = mf_div(albedo, mf_sub(mf_set1(1.), mf_mul(albedo, mf_set1(Fdr))));
    else
      diff = mf_div(albedo, mf_set1(1. - Fdr));
    diff = mf_mul(diff, mf_set1(T12 * T21 / (M_PI * eta * eta)));
  }
  p->v[v].mode = s_reflect | s_glossy;
  return mf_add(glossy, diff);
}

mf_t sample(path_t *p, void* data)
{
  const int v = p->length-1; // current vertex.

  const float cos_in  = -dotproduct(p->v[v].hit.n, p->e[v].omega);
  if(cos_in <= 0.0f) return mf_set1(0.f);

  const float r = p->v[v].shading.roughness;
  float eta = ((plastic_t*) data)->n_d;
  float probSpecular = mf(prob_specular(cos_in, r, eta, p->v[v].shading.rg, p->v[v].shading.rd), 0);
  const float x1 = pointsampler(p, s_dim_omega_x);
  const float x2 = pointsampler(p, s_dim_omega_y);
  if(pointsampler(p, s_dim_scatter_mode) <= probSpecular)
  { // glossy
    const float *n = p->v[v].hit.n;
    float h[3] = {n[0], n[1], n[2]}; // init for smooth dielectric
    // flip incoming direction to point away from intersection point
    const float wit[3] = {
      - dotproduct(p->v[v].hit.a, p->e[v].omega),
      - dotproduct(p->v[v].hit.b, p->e[v].omega),
      - dotproduct(p->v[v].hit.n, p->e[v].omega)};
    float ht[3]; // h in tangent space
    ggx_sample_h(wit, r, r, pointsampler(p, s_dim_omega_x), pointsampler(p, s_dim_omega_y), ht);
    // world space:
    for(int k=0;k<3;k++) h[k] = ht[0]*p->v[v].hit.a[k] + ht[1]*p->v[v].hit.b[k] + ht[2]*n[k];
    const float cosr = -dotproduct(p->e[v].omega, h);
    if(!(cosr > 0.0f)) return mf_set1(0.0f); // sampled a microfacet from the wrong side :(

    for(int k=0;k<3;k++) p->e[v+1].omega[k] = p->e[v].omega[k] + 2.0f * cosr * h[k];
    if(dotproduct(p->e[v+1].omega, n) <= 0.0f) return mf_set1(0.0f);

    if(dotproduct(p->e[v+1].omega, n)*dotproduct(p->e[v+1].omega, h) < 0.0f) return mf_set1(0.0f);
  }
  else
  { // diffuse
    float *n = p->v[v].hit.n;
    float s = sqrtf(x1);
    for(int k=0;k<3;k++)
      p->e[v+1].omega[k] =
        sqrtf(1.0 - x1)   * n[k] +
        s*cosf(2*M_PI*x2) * p->v[v].hit.a[k] +
        s*sinf(2*M_PI*x2) * p->v[v].hit.b[k];
  }
  mf_t brdf_ = brdf(p, v, data);
  mf_t pdf_ = pdf(p, v, v, v+1, data);
  p->v[v+1].pdf = pdf_;
  return mf_div(brdf_, pdf_);
}

// fresnelDielectricExt from mitsuba
static float fresnel(float cosThetaI, float eta) {
  if(eta == 1) return 0;

  /* Using Snell's law, calculate the squared sine of the
     angle between the normal and the transmitted ray */
  float scale = (cosThetaI > 0) ? 1/eta : eta;
  float cosThetaTSqr = 1 - (1-cosThetaI*cosThetaI) * (scale*scale);

  /* Check for total internal reflection */
  if (cosThetaTSqr <= 0.0f) return 1.f;

  /* Find the absolute cosines of the incident/transmitted rays */
  cosThetaI = fabsf(cosThetaI);
  float cosThetaT = sqrtf(cosThetaTSqr);

  float Rs = (cosThetaI - eta * cosThetaT)
    / (cosThetaI + eta * cosThetaT);
  float Rp = (eta * cosThetaI - cosThetaT)
    / (eta * cosThetaI + cosThetaT);

  /* No polarization -- return the unpolarized reflectance */
  return 0.5f * (Rs * Rs + Rp * Rp);
}

static float warped_alpha(float alpha)
{
  float alpha_min = rtrans.alpha_min;
  float alpha_max = rtrans.alpha_max;
  return powf((alpha - alpha_min) / (alpha_max - alpha_min), 0.25f);
}

static float warped_eta(float eta)
{
  float eta_min = rtrans.eta_min;
  float eta_max = rtrans.eta_max;
  eta = MAX(eta, eta_min);
  return powf((eta - eta_min) / (eta_max - eta_min), 0.25f);
}

static float evalCubicInterp3D(const float p[3], const float *values, const size_t size[3])
{
  float knotWeights[3][4];
  size_t knot[3];

  /* Compute interpolation weights separately for each dimension */
  for (int dim=0; dim<3; ++dim) {
    float *weights = knotWeights[dim];
    /* Give up when given an out-of-range or NaN argument */
    if (!(p[dim] >= 0.f && p[dim] <= 1.f)) return 0.0f;

    /* Transform 'p' so that knots lie at integer positions */
    float t = (p[dim] * (size[dim] - 1));

    /* Find the index of the left knot in the queried subinterval, be
       robust to cases where 't' lies exactly on the right endpoint */
    knot[dim] = MIN((size_t) t, size[dim] - 2);

    /* Compute the relative position within the interval */
    t = t - (float) knot[dim];

    /* Compute node weights */
    float t2 = t*t, t3 = t2*t;
    weights[0] = 0.0f;
    weights[1] = 2*t3 - 3*t2 + 1;
    weights[2] = -2*t3 + 3*t2;
    weights[3] = 0.0f;

    /* Derivative weights */
    float d0 = t3 - 2*t2 + t,
          d1 = t3 - t2;

    /* Turn derivative weights into node weights using
       an appropriate chosen finite differences stencil */
    if (knot[dim] > 0) {
      weights[2] +=  0.5f * d0;
      weights[0] -=  0.5f * d0;
    } else {
      weights[2] += d0;
      weights[1] -= d0;
    }

    if (knot[dim] + 2 < size[dim]) {
      weights[3] += 0.5f * d1;
      weights[1] -= 0.5f * d1;
    } else {
      weights[2] += d1;
      weights[1] -= d1;
    }
  }

  float result = 0.0f;
  for (int z=-1; z<=2; ++z) {
    float wz = knotWeights[2][z+1];
    for (int y=-1; y<=2; ++y) {
      float wyz = knotWeights[1][y+1] * wz;
      for (int x=-1; x<=2; ++x) {
        float wxyz = knotWeights[0][x+1] * wyz;

        if (wxyz == 0)
          continue;

        size_t pos = ((knot[2] + z) * size[1] + (knot[1] + y))
          * size[0] + knot[0] + x;

        result += values[pos] * wxyz;
      }
    }
  }
  return result;
}

// RoughTransmittance::eval from mitsuba
float rtrans_eval(float cosTheta, float alpha, float eta)
{
  if (cosTheta < 0) {
    assert(0);
    cosTheta = -cosTheta;
    eta = 1.0f / eta;
  }

  const float *data = rtrans.data;
  if (eta < 1) {
    /* Entering a less dense medium -- skip ahead to the
       second data block */
    data += rtrans.eta_samples * rtrans.alpha_samples * rtrans.theta_samples;
    eta = 1.0f / eta;
  }

  /* Transform the roughness and IOR values into the warped parameter space */
  float point[] = {
    powf(fabsf(cosTheta), 0.25f),
    warped_alpha(alpha),
    warped_eta(eta),
  };
  size_t size[] = { rtrans.theta_samples, rtrans.alpha_samples, rtrans.eta_samples };
  float result = evalCubicInterp3D(point, data, size);
  return CLAMP(result, 0.f, 1.f);
}

static float evalCubicInterp2D(const float p[2], const float *values, const size_t size[2])
{
  float knotWeights[2][4];
  size_t knot[2];

  /* Compute interpolation weights separately for each dimension */
  for (int dim=0; dim<2; ++dim) {
    float *weights = knotWeights[dim];
    /* Give up when given an out-of-range or NaN argument */
    if (!(p[dim] >= 0.f && p[dim] <= 1.f)) return 0.0f;

    /* Transform 'p' so that knots lie at integer positions */
    float t = p[dim] * (size[dim] - 1);

    /* Find the index of the left knot in the queried subinterval, be
       robust to cases where 't' lies exactly on the right endpoint */
    knot[dim] = MIN((size_t) t, size[dim] - 2);

    /* Compute the relative position within the interval */
    t = t - (float) knot[dim];

    /* Compute node weights */
    float t2 = t*t, t3 = t2*t;
    weights[0] = 0.0f;
    weights[1] = 2*t3 - 3*t2 + 1;
    weights[2] = -2*t3 + 3*t2;
    weights[3] = 0.0f;

    /* Derivative weights */
    float d0 = t3 - 2*t2 + t,
          d1 = t3 - t2;

    /* Turn derivative weights into node weights using
       an appropriate chosen finite differences stencil */
    if (knot[dim] > 0) {
      weights[2] +=  0.5f * d0;
      weights[0] -=  0.5f * d0;
    } else {
      weights[2] += d0;
      weights[1] -= d0;
    }

    if (knot[dim] + 2 < size[dim]) {
      weights[3] += 0.5f * d1;
      weights[1] -= 0.5f * d1;
    } else {
      weights[2] += d1;
      weights[1] -= d1;
    }
  }

  float result = 0.0f;
  for (int y=-1; y<=2; ++y) {
    float wy = knotWeights[1][y+1];
    for (int x=-1; x<=2; ++x) {
      float wxy = knotWeights[0][x+1] * wy;

      if (wxy == 0)
        continue;

      size_t pos = (knot[1] + y) * size[0] + knot[0] + x;

      result += values[pos] * wxy;
    }
  }
  return result;
}

// RoughTransmittance::evalDiffuse from mitsuba
static float rtrans_eval_diff(float alpha, float eta)
{
  const float *data = rtrans.diff_data;
  if (eta < 1) {
    /* Entering a less dense medium -- skip ahead to the
       second data block */
    data += rtrans.eta_samples * rtrans.alpha_samples;
    eta = 1.0f / eta;
  }

  /* Transform the roughness and IOR values into the warped parameter space */
  float point[] = {
    warped_alpha(alpha),
    warped_eta(eta),
  };
  size_t size[] = { rtrans.alpha_samples, rtrans.eta_samples };
  float result = evalCubicInterp2D(point, data, size);
  return CLAMP(result, 0.f, 1.f);
}

static mf_t prob_specular(
    float cos_in,
    float alpha,
    float eta,
    mf_t glossy_reflectance,
    mf_t diffuse_reflectance)
{
  float T12 = rtrans_eval(cos_in, alpha, eta);
  float probSpecular = 1 - T12;
  mf_t glossy = mf_mul(glossy_reflectance, mf_set1(probSpecular));
  mf_t diffuse = mf_mul(diffuse_reflectance, mf_set1(1.f - probSpecular));
  return mf_div(glossy, mf_add(glossy, diffuse));
}

static int load_rtrans_data(const char *filename)
{
  FILE *f = fopen(filename, "rb");
  if(f == NULL) return 1;

  const char magic[] = "MTS_TRANSMITTANCE";
  const size_t magic_len = sizeof(magic) - 1; // strlen
  size_t head_size = magic_len + sizeof(uint64_t) * 3 + sizeof(float) * 4;
  uint8_t head_buf[128];
  assert(sizeof(head_buf) >= head_size);
  if(!fread(head_buf, head_size, 1, f)) goto fail;
  if(memcmp(magic, head_buf, magic_len) != 0) goto fail;

  size_t pos = magic_len;
  memcpy(&rtrans.eta_samples, head_buf + pos, sizeof(uint64_t));
  pos += sizeof(uint64_t);
  memcpy(&rtrans.alpha_samples, head_buf + pos, sizeof(uint64_t));
  pos += sizeof(uint64_t);
  memcpy(&rtrans.theta_samples, head_buf + pos, sizeof(uint64_t));
  pos += sizeof(uint64_t);

  memcpy(&rtrans.eta_min, head_buf + pos, sizeof(float));
  pos += sizeof(float);
  memcpy(&rtrans.eta_max, head_buf + pos, sizeof(float));
  pos += sizeof(float);
  memcpy(&rtrans.alpha_min, head_buf + pos, sizeof(float));
  pos += sizeof(float);
  memcpy(&rtrans.alpha_max, head_buf + pos, sizeof(float));
  pos += sizeof(float);

  const size_t trans_size = 2 * rtrans.eta_samples * rtrans.alpha_samples
    * rtrans.theta_samples;
  const size_t diff_trans_size = 2 * rtrans.eta_samples * rtrans.alpha_samples;
  rtrans.data = malloc(sizeof(float) * trans_size);
  rtrans.diff_data = malloc(sizeof(float) * diff_trans_size);
  float *temp = malloc(sizeof(float) * (trans_size + diff_trans_size));
  if(!fread(temp, sizeof(float) * (trans_size + diff_trans_size), 1, f)) goto fail;

  float *ptr = temp;
  size_t fdrEntry = 0, dataEntry = 0;
  for(size_t i = 0; i < 2 * rtrans.eta_samples; ++i)
  {
    for(size_t j = 0; j < rtrans.alpha_samples; ++j)
    {
      for(size_t k = 0; k < rtrans.theta_samples; ++k) rtrans.data[dataEntry++] = *ptr++;
      rtrans.diff_data[fdrEntry++] = *ptr++;
    }
  }
  free(temp);

  fclose(f);
  return 0;

fail:
  fclose(f);
  return 1;
}
