import xml.etree.ElementTree as ET
import os
import bpy
import numpy as np
import sys
import mathutils
import struct
import zlib
import io

scenefile = '/home/vincent/scenes/guiding3v/bathroom/mitsuba/scene.xml'
scenedir = os.path.dirname(scenefile)

fstop = 32.
aspect = 0. # set in import_sensor TODO

def main():
    tree = ET.parse(scenefile)
    root = tree.getroot()

    scene_transform = import_sensor(root)
    materials = import_materials(root)
    import_shapes(root, materials, scene_transform)
    import_envmap(root)

def import_sensor(root):
    sensor = root.find('sensor')
    assert(sensor.get('type') == 'thinlens')
    mitsuba_sensor = np.array([
        [-1., 0., 0., 0.],
        [0., 1., 0., 0.],
        [0.,0.,-1.,0.],
        [0.,0.,0.,1.]
    ]).reshape((4,4))
    bpy.ops.object.camera_add()
    camera = bpy.context.active_object
    matrix_sensor = read_transform(sensor)
    camera_matrix = matrix_sensor @ mitsuba_sensor
    # match default mitsuba camera to default corona camera
    # makes navigation in blender easier and environments match
    scene_transform = np.eye(4)
    scene_transform[0,0] = -1
    scene_transform[1,1] = 0
    scene_transform[2,2] = 0
    scene_transform[1,2] = 1
    scene_transform[2,1] = 1
    # flip x if necessary (works for clocks, not necessarily in general)
    # TODO envmaps don't flip
    handedness = np.sign(np.linalg.det(camera_matrix[:3,:3]))
    scene_transform[0] *= handedness
    if handedness == -1:
        print("warning: camera flipped")
    camera_matrix = scene_transform @ camera_matrix
    camera.matrix_world = camera_matrix.T

    scene = bpy.context.scene
    film = sensor.find('film')
    film_width = film.find("integer[@name='width']")
    film_height = film.find("integer[@name='height']")
    film_width = 768 if film_width is None else int(film_width.get('value'))
    film_height = 576 if film_height is None else int(film_height.get('value'))
    scene.render.resolution_x = film_width
    scene.render.resolution_y = film_height

    fov = [child for child in sensor.findall('float') if child.get('name') == 'fov']
    if not fov:
        print("no fov found for sensor")
        assert(0)
    camera.data.sensor_width = 35.0
    camera.data.angle = np.radians(float(fov[0].get('value')))
    camera.data.dof.focus_distance = float(sensor.find("float[@name='focusDistance']").get('value'))
    scene.camera = camera
    cam_text = 'camera'
    global aspect # TODO
    aspect = float(film_width) / float(film_height)
    if cam_text in bpy.data.texts:
        bpy.data.texts.remove(bpy.data.texts[cam_text])
    text = bpy.data.texts.new(cam_text)
    text.write(("{{\n"
        '"aperture_value": 12\n'
    "}}\n").format(aspect * 0.35 / aspect))
    return scene_transform

def import_materials(root):
    scene_materials = set()
    for material in root.findall('bsdf'):
        mat_id = material.get('id')
        add_material(root, scene_materials, mat_id, material)
    return scene_materials

def add_material(root, scene_materials, mat_id, material):
    scene_materials.add(mat_id)
    if mat_id in bpy.data.materials:
        bpy.data.materials.remove(bpy.data.materials[mat_id])
    mat_obj = bpy.data.materials.new(mat_id)
    if mat_id in bpy.data.texts:
        bpy.data.texts.remove(bpy.data.texts[mat_id])
    text = bpy.data.texts.new(mat_id)
    corona_material = read_material(root, material)
    text.write(corona_material)

def import_shapes(root, materials, scene_transform):
    for shape in root.findall('shape'):
        for obj in bpy.context.selected_objects:
            obj.select_set(False)
        face_normals = shape.find("boolean[@name='faceNormals']")
        face_normals = False if face_normals is None else str_to_bool(face_normals.get('value'))
        needs_edge_split = face_normals
        shape_type = shape.get('type')
        if shape_type == 'obj':
            objname = shape.find("string[@name='filename']").get('value')
            import_obj(os.path.join(scenedir, objname))
            needs_edge_split = True
        elif shape_type == 'rectangle':
            bpy.ops.mesh.primitive_plane_add()
        elif shape_type == 'sphere':
            import_sphere(shape)
        elif shape_type == 'serialized':
            import_serialized(shape)
        else:
            print("don't know how to import shape type {}".format(shape_type))
            continue

        matrix = read_transform(shape)
        imported = bpy.context.selected_objects[0]

        flip_normals = shape.find("boolean[@name='flipNormals']")
        flip_normals = False if flip_normals is None else str_to_bool(flip_normals.get('value'))
        if flip_normals:
            do_flip_normals(imported)
        if needs_edge_split:
            add_edge_split(imported, face_normals)
        matrix = scene_transform @ matrix
        imported.matrix_world = matrix.T
        imported.active_material = bpy.data.materials[resolve_material(root, shape, materials)]

anon_counter = 0
def resolve_material(root, shape, materials):
    global anon_counter
    maybe_emitter = shape.find("emitter[@type='area']")
    if maybe_emitter is not None:
        anon_id = 'emitter{:05d}'.format(anon_counter)
        anon_counter += 1
        add_material(root, materials, anon_id, maybe_emitter)
        return anon_id
    maybe_bsdf = shape.find("bsdf")
    if maybe_bsdf is not None:
        anon_id = 'bsdf{:05d}'.format(anon_counter)
        anon_counter += 1
        add_material(root, materials, anon_id, maybe_bsdf)
        return anon_id
    for ref in shape.findall('ref'):
        ref_id = ref.get('id')
        if ref_id in materials:
            return ref_id
        else:
            print("ref {} not understood in shape {}".format(ref, imported.name))
    assert(0) # no material found

def add_edge_split(obj, face_normals):
    if face_normals:
        # mark all edges as sharp and set split normals to face normals (quite slow)
        bpy.context.view_layer.objects.active = obj
        bpy.ops.object.editmode_toggle()
        bpy.ops.mesh.select_all(action='SELECT')
        bpy.ops.mesh.split_normals()
        bpy.ops.object.editmode_toggle()
        bpy.context.view_layer.objects.active = None
    mod = obj.modifiers.new(name='EdgeSplit', type='EDGE_SPLIT')
    mod.use_edge_sharp = True
    mod.use_edge_angle = False

def do_flip_normals(obj):
    bpy.context.view_layer.objects.active = obj
    bpy.ops.object.editmode_toggle()
    bpy.ops.mesh.select_all(action='SELECT')
    bpy.ops.mesh.flip_normals()
    bpy.ops.object.editmode_toggle()

def import_sphere(shape):
    radius_attr = shape.find("float[@name='radius']").get('value')
    radius = 1 if radius_attr is None else float(radius_attr)
    point_node = shape.find("point[@name='center']")
    x = 0 if point_node is None else float(point_node.get('x'))
    y = 0 if point_node is None else float(point_node.get('y'))
    z = 0 if point_node is None else float(point_node.get('z'))
    bpy.ops.mesh.primitive_ico_sphere_add(subdivisions=6, radius=radius, location=(x,y,z))
    bpy.ops.object.transform_apply()

def import_serialized(shape):
    filename = shape.find("string[@name='filename']").get('value')
    shape_index = int(shape.find("integer[@name='shapeIndex']").get('value'))
    mesh = read_serialized_mesh(os.path.join(scenedir, filename), shape_index)
    (name, pos, normal, uv, _, indices) = mesh
    mesh_obj = bpy.data.meshes.new(name)
    mesh_obj.vertices.add(pos.shape[0])
    mesh_obj.vertices.foreach_set("co", pos.flatten())
    mesh_obj.vertices.foreach_set("normal", normal.flatten())
    indices_flat = indices.flatten()
    mesh_obj.loops.add(indices_flat.shape[0])
    mesh_obj.loops.foreach_set("vertex_index", indices_flat)
    num_triangles = indices.shape[0]
    mesh_obj.polygons.add(num_triangles)
    mesh_obj.polygons.foreach_set("loop_start", np.arange(0, num_triangles * 3, 3))
    mesh_obj.polygons.foreach_set("loop_total", np.repeat(3, num_triangles))
    if uv is not None:
        uv_layer = mesh_obj.uv_layers.new()
        uv_layer.data.foreach_set("uv", uv[indices_flat].flatten())
    mesh_obj.update()
    mesh_obj.validate()
    blender_obj = bpy.data.objects.new(name, mesh_obj)
    bpy.context.collection.objects.link(blender_obj)
    blender_obj.select_set(True)

def read_serialized_mesh(path, shape_index):
    with open(path, 'rb') as f:
        check_header(f)
        offsets = read_offsets(f)
        offset = offsets[shape_index]
        end = offsets[shape_index + 1]
        f.seek(header_format.size + offset)
        mesh_buffer = read_compressed_mesh(f, end - offset)
    return parse_mesh(mesh_buffer)

# ==============================================================================
# serialized mesh format
header_format = struct.Struct('<HH')

def check_header(infile):
    header_raw = infile.read(header_format.size)
    ident, version = header_format.unpack(header_raw)
    if ident != 0x041c:
        print("not a serialized file")
        assert(0)
    if version != 4:
        print("warning: version {} not supported", version)

def read_offsets(infile):
    num_meshes_format = struct.Struct('<I')
    infile.seek(-num_meshes_format.size, io.SEEK_END)
    num_meshes_raw = infile.read(num_meshes_format.size)
    num_meshes, = num_meshes_format.unpack(num_meshes_raw)

    # seek back to beginning of offsets
    offset_format = struct.Struct('<Q')
    seek_offset = num_meshes_format.size + num_meshes * offset_format.size
    infile.seek(-seek_offset, io.SEEK_END)
    last = infile.tell()
    offsets_raw = infile.read(num_meshes * offset_format.size)
    offsets = list(x[0] for x in offset_format.iter_unpack(offsets_raw))
    offsets.append(last)
    return offsets

def read_compressed_mesh(infile, length):
    compressed = infile.read(length)
    return zlib.decompress(compressed)

flags_format = struct.Struct('<I')
flags_normals = 0x0001
flags_tex = 0x0002
flags_colors = 0x0008
flags_use_face = 0x0010
flags_single = 0x1000
flags_double = 0x2000
def read_flags(buf, offset):
    flags, = flags_format.unpack_from(buf, offset)
    return (flags, offset + flags_format.size)

def read_string(buf, offset):
    end = buf.index(0, offset)
    return (buf[offset:end].decode(), end + 1)

def read_array(buf, offset, n, components, form, dtype=float):
    end = offset + n * components * form.size
    it = (x[0] for x in form.iter_unpack(buf[offset:end]))
    array = np.fromiter(it, dtype=dtype)
    #for _ in range(n):
    #    vec = []
    #    for _ in range(components):
    #        val, = form.unpack_from(buf, offset)
    #        vec.append(val)
    #        offset += form.size
    #    array.append(vec)
    return (array.reshape(n, components), end)

num_vert_tri_format = struct.Struct('<QQ')
def parse_mesh(mesh_buffer):
    offset = 0
    (flags, offset) = read_flags(mesh_buffer, offset)
    (name, offset) = read_string(mesh_buffer, offset)
    num_vertices, num_triangles = num_vert_tri_format.unpack_from(mesh_buffer, offset)
    offset += num_vert_tri_format.size
    is_single = flags & flags_single
    is_double = flags & flags_double
    if is_single and not is_double:
        float_format = struct.Struct('f')
    elif not is_single and is_double:
        float_format = struct.Struct('d')
    else:
        print("warning: mesh '{}' has undefined precision, skipping".format(name))
        return
    (vertex_positions, offset) = read_array(mesh_buffer, offset, num_vertices, 3, float_format)
    vertex_normals = None
    if flags & flags_normals:
        (vertex_normals, offset) = read_array(mesh_buffer, offset, num_vertices, 3, float_format)
    vertex_tex = None
    if flags & flags_tex:
        (vertex_tex, offset) = read_array(mesh_buffer, offset, num_vertices, 2, float_format)
    vertex_colors = None
    if flags & flags_colors:
        (vertex_colors, offset) = read_array(mesh_buffer, offset, num_vertices, 2, float_format)
    index_format = struct.Struct('<I' if num_vertices <= 0xFFFFFFFF else '<Q')
    (indices, offset) = read_array(mesh_buffer, offset, num_triangles, 3, index_format, int)
    if flags & flags_use_face:
        print("mesh '{}' should use face normals".format(name))
    return (
        name,
        vertex_positions,
        vertex_normals,
        vertex_tex,
        vertex_colors,
        indices
    )
# ==============================================================================

def str_to_bool(string):
    if string != 'false' and string != 'true':
        print('warning: got boolean value "{}"'.format(string))
    return string == 'true'

def read_transform(node):
    transform_matrix = np.eye(4)
    transform = node.find('transform')
    if transform is None:
        return transform_matrix
    assert(transform.get('name') == 'toWorld')
    for basic_transform in transform:
        tag = basic_transform.tag
        if tag == 'translate':
            basic_matrix = read_translation(basic_transform)
        elif tag == 'rotate':
            basic_matrix = read_rotation(basic_transform)
        elif tag == 'scale':
            basic_matrix = read_scaling(basic_transform)
        elif tag == 'matrix':
            basic_matrix = read_matrix(basic_transform)
        elif tag == 'lookat':
            basic_matrix = read_lookat(basic_transform)
        else:
            print('transform not recognized')
            assert(0)
        transform_matrix = basic_matrix @ transform_matrix
    return transform_matrix

def read_translation(translation):
    x = translation.get('x')
    y = translation.get('y')
    z = translation.get('z')
    matrix = np.eye(4)
    matrix[0, 3] = 0 if x is None else float(x)
    matrix[1, 3] = 0 if y is None else float(y)
    matrix[2, 3] = 0 if z is None else float(z)
    return matrix

def read_rotation(rotation):
    x = rotation.get('x')
    y = rotation.get('y')
    z = rotation.get('z')
    angle = rotation.get('angle')
    axis = np.array([
        0. if x is None else float(x),
        0. if y is None else float(y),
        0. if z is None else float(z)
    ])
    axis /= np.linalg.norm(axis)
    angle = 0. if angle is None else float(angle)
    cos_theta = np.cos(np.radians(angle))
    sin_theta = np.sin(np.radians(angle))
    matrix = np.eye(4)
    matrix[0, 0] = axis[0] * axis[0] + (1. - axis[0] * axis[0]) * cos_theta
    matrix[0, 1] = axis[0] * axis[1] * (1. - cos_theta) - axis[2] * sin_theta
    matrix[0, 2] = axis[0] * axis[2] * (1. - cos_theta) + axis[1] * sin_theta

    matrix[1, 0] = axis[0] * axis[1] * (1. - cos_theta) + axis[2] * sin_theta
    matrix[1, 1] = axis[1] * axis[1] + (1. - axis[1] * axis[1]) * cos_theta
    matrix[1, 2] = axis[1] * axis[2] * (1. - cos_theta) - axis[0] * sin_theta

    matrix[2, 0] = axis[0] * axis[2] * (1. - cos_theta) - axis[1] * sin_theta
    matrix[2, 1] = axis[1] * axis[2] * (1. - cos_theta) + axis[0] * sin_theta
    matrix[2, 2] = axis[2] * axis[2] + (1. - axis[2] * axis[2]) * cos_theta
    return matrix

def read_scaling(scaling):
    scale = scaling.get('value')
    x = scaling.get('x')
    y = scaling.get('y')
    z = scaling.get('z')
    if scale is not None:
        assert(x is None)
        assert(y is None)
        assert(z is None)
        x = float(scale)
        y = float(scale)
        z = float(scale)
    matrix = np.eye(4)
    matrix[0, 0] = 1 if x is None else float(x)
    matrix[1, 1] = 1 if y is None else float(y)
    matrix[2, 2] = 1 if z is None else float(z)
    return matrix

def read_matrix(matrix):
    return parse_vec(matrix.get('value')).reshape((4,4))

def read_lookat(lookat):
    origin = parse_vec(lookat.get('origin'))
    target = parse_vec(lookat.get('target'))
    up = parse_vec(lookat.get('up'))
    d = target - origin
    d /= np.linalg.norm(d)
    left = np.cross(up, d)
    left /= np.linalg.norm(left)
    new_up = np.cross(d, left)
    matrix = np.eye(4)
    matrix[:3, 0] = left
    matrix[:3, 1] = new_up
    matrix[:3, 2] = d
    matrix[:3, 3] = origin
    return matrix

def parse_vec(value):
    value = value.strip()
    separator = ',' if ',' in value else ' '
    return np.array([float(x) for x in value.split(separator)])

def import_obj(filename):
    bpy.ops.import_scene.obj(filepath=filename,
        use_split_objects=False,
        use_image_search=False)

def read_material(root, material):
    mat_type = material.get('type')
    if material.tag == 'emitter':
        assert(mat_type == 'area')
        return read_area_emitter(root, material)
    assert(material.tag == 'bsdf')
    if mat_type == 'diffuse':
        return read_diffuse_material(root, material)
    elif mat_type == 'conductor':
        return read_conductor_material(root, material)
    elif mat_type == 'roughconductor':
        return read_conductor_material(root, material, rough=True)
    elif mat_type == 'dielectric':
        return read_dielectric_material(root, material)
    elif mat_type == 'roughdielectric':
        return read_dielectric_material(root, material, rough=True)
    elif mat_type == 'plastic':
        print('please convert plastic to roughplastic with (near zero) roughness')
        assert(0)
        return 'plastic not supported - convert to roughplastic'
    elif mat_type == 'roughplastic':
        return read_plastic_material(root, material, rough=True)
    else:
        print('no support for material type: {}'.format(mat_type))
        assert(0)
        return 'no support for material type: {}'.format(mat_type)

def read_area_emitter(root, node):
    mts_color = read_color(node.find("*[@name='radiance']"))
    alpha_node = node.find("float[@name='alpha']")
    alpha = 1. if alpha_node is None else float(alpha_node.get('value'))
    rec709_to_ergb = np.array([0.830127, 1.054504, 1.099916])
    scale = (106.86535 # camera sensor response
        * (100. / 125.) # exposure time
        * (0.35 * (0.35 / aspect)) # sensor width * sensor height
        / (4. * fstop * fstop)
        * 106.8 / (830. - 360.) # illum e norm / wavelength range
    )
    color = (rec709_to_ergb * mts_color) / scale
    shader = (
        'diffuse #\n'
        'color e {:g} {:g} {:g} {:g} #\n'
    ).format(*color, alpha)
    shader = shader_append_mult(shader)
    return shader

def read_diffuse_material(root, node):
    refl_node = node.find("*[@name='reflectance']")
    if refl_node is None:
        refl = np.array([0.5, 0.5, 0.5])
    else:
        refl = read_texture_or_color(root, refl_node)
    shader = 'diffuse #\n'
    shader += get_shader_set_slot('d', refl)
    shader = shader_append_mult(shader)
    return shader

def read_dielectric_material(root, node, rough=False):
    ref_id = node.get('id') # for errors, might not be set
    alpha = 0. if not rough else read_rough(node)
    refl_node = node.find("*[@name='specularReflectance']")
    if refl_node is None:
        refl = np.array([1., 1., 1.])
    else:
        refl = read_texture_or_color(root, refl_node)
    trans_node = node.find("*[@name='specularTransmittance']")
    if trans_node is None:
        trans = np.array([1., 1., 1.])
    else:
        trans = read_texture_or_color(root, trans_node)
    if not np.all(refl == trans):
        print('reflectance/transmittance needs to match for material {}'.format(ref_id))
        assert(0)
    ext_ior_node = node.find("float[@name='extIOR']")
    if ext_ior_node is None or float(ext_ior_node.get('value')) != 1.:
        print('extIOR needs to be set to 1 for material {}'.format(ref_id))
        assert(0)
    int_ior_node = node.find("*[@name='intIOR']")
    if int_ior_node.tag != 'float':
        print('intIOR needs to be specified as float for material {}'.format(ref_id))
        assert(0)
    ior = float(int_ior_node.get('value'))
    shader = 'dielectric {:g} 0 #\n'.format(ior)
    shader += get_shader_set_slot('g', refl, alpha)
    shader = shader_append_mult(shader)
    return shader

def read_conductor_material(root, node, rough=False):
    refl_node = node.find("*[@name='specularReflectance']")
    if refl_node is None:
        refl = np.array([1., 1., 1.])
    else:
        refl = read_texture_or_color(root, refl_node)
    alpha = 0. if not rough else read_rough(node)
    metal_node = node.find("string[@name='material']")
    metal = 'Cu' if metal_node is None else metal_node.get('value')
    if metal not in {'Ag', 'Au', 'Cu'}:
        print("unsupported metal: {}".format(metal))
        assert(0)
    if node.find("*[@name='eta']") is not None or node.find("*[@name='k']") is not None:
        print('eta/k not supported for rough conductor')
        assert(0)
    ext_eta_node = node.find("float[@name='extEta']")
    if ext_eta_node is None or float(ext_eta_node.get('value')) != 1.:
        print('extEta needs to be set to 1 for material {}'.format(node.get('id')))
        assert(0)
    shader = 'metal {} #\n'.format(metal)
    shader += get_shader_set_slot('g', refl, alpha)
    shader = shader_append_mult(shader)
    return shader

def read_plastic_material(root, node, rough=False):
    ref_id = node.get('id') # for errors, might not be set
    alpha = 0. if not rough else read_rough(node)
    ext_ior_node = node.find("float[@name='extIOR']")
    if ext_ior_node is None or float(ext_ior_node.get('value')) != 1.:
        print('extIOR needs to be set to 1 for material {}'.format(ref_id))
        assert(0)
    int_ior_node = node.find("*[@name='intIOR']")
    if int_ior_node.tag != 'float':
        print('intIOR needs to be specified as float for material {}'.format(ref_id))
        assert(0)
    ior = float(int_ior_node.get('value'))

    g_refl_node = node.find("*[@name='specularReflectance']")
    if g_refl_node is None:
        g_refl = np.array([1.0, 1.0, 1.0])
    else:
        g_refl = read_texture_or_color(root, g_refl_node)

    d_refl_node = node.find("*[@name='diffuseReflectance']")
    if d_refl_node is None:
        d_refl = np.array([0.5, 0.5, 0.5])
    else:
        d_refl = read_texture_or_color(root, d_refl_node)

    nonlinear_node = node.find("boolean[@name='nonlinear']")
    nonlinear = False if nonlinear_node is None else str_to_bool(nonlinear_node.get('value'))

    shader = 'plastic {} {} #\n'.format(ior, 1 if nonlinear else 0)
    shader += get_shader_set_slot('g', g_refl, alpha)
    shader += get_shader_set_slot('d', d_refl) # alpha is already set by glossy color
    shader = shader_append_mult(shader)
    return shader

def read_rough(node):
    ref_id = node.get('id') # for errors, might not be set
    distr = node.find("string[@name='distribution']").get('value')
    if distr != 'ggx':
        print('distribution not set to ggx for material {}'.format(ref_id))
        assert(0)
    if node.find("*[@name='alphaU']") is not None or node.find("*[@name='alphaV']") is not None:
        print('anisotropy not supported (in material {})'.format(ref_id))
        assert(0)
    alpha = node.find("*[@name='alpha']")
    if alpha is not None and alpha.tag != 'float':
        print('alpha texture not supported (in material {})'.format(ref_id))
        assert(0)
    return float(alpha.get('value')) if alpha is not None else 0.1

def read_color(node):
    if node.tag != 'spectrum' and node.tag != 'rgb':
        print("did not understand color of type {}".format(node.tag))
        assert(0)
    color = parse_vec(node.get('value'))
    if color.shape[0] == 1:
        color = np.repeat(color[0], 3)
    if color.shape[0] != 3:
        print("invalid color of length {}".format(color.shape[0]))
        assert(0)
    return color

def read_texture_or_color(root, node):
    if node.tag == 'ref':
        ref_id = node.get('id')
        node = root.find("texture[@id='{}']".format(ref_id))
    if node.tag == 'texture':
        assert(node.get('type') == 'bitmap')
        filename = node.find("string[@name='filename']").get('value')
        filter_type = node.find("string[@name='filterType']")
        if filter_type is None or filter_type.get('value') != 'nearest':
            print("filter not set to nearest for texture {}".format(filename))
            assert(0)
        gamma = node.find("float[@name='gamma']")
        if gamma is not None and float(gamma.get('value')) != -1:
            print("gamma not set to sRGB for texture {}".format(filename))
            assert(0)
        uscale = node.find("float[@name='uscale']")
        vscale = node.find("float[@name='vscale']")
        uoffset = node.find("float[@name='uoffset']")
        voffset = node.find("float[@name='voffset']")
        if uscale is not None or vscale is not None or uoffset is not None or voffset is not None:
            print("modified uvs for texture {}".format(filename))
            assert(0)
        name = os.path.splitext(filename)[0]
        return name + '.fb'
    else:
        return read_color(node)

def get_shader_set_slot(slot, texture_or_color, alpha=1.):
    if type(texture_or_color) == str:
        if alpha < 1.:
            return 'texture {slot} {name} #\ncolor {slot} 1 1 1 {alpha} #\n'.format(slot=slot, name=texture_or_color, alpha=alpha)
        else:
            return 'texture {slot} {name} #\n'.format(slot=slot, name=texture_or_color)
    else:
        return 'color {} {:g} {:g} {:g} {:g} #\n'.format(slot, *texture_or_color, alpha)

def shader_append_mult(shader):
    n_before = shader.count('\n')
    assert(n_before > 0)
    if n_before == 1:
        return # don't need a mult
    sequence = ' '.join(str(-x) for x in range(1, n_before + 1))
    shader += 'mult {} {} #\n'.format(n_before - 1, sequence)
    return shader

def import_envmap(root):
    env_shader = 'black'
    env_node = root.find("emitter[@type='envmap']")
    if env_node:
        filename = env_node.find("string[@name='filename']").get('value')
        assert(filename.endswith('.exr'))
        filename = os.path.splitext(filename)[0] + '.fb'
        to_world = read_transform(env_node)
        mts2c = np.eye(4)
        mts2c[0,0] = -1
        mts2c[1,1] = 0
        mts2c[2,2] = 0
        mts2c[1,2] = 1
        mts2c[2,1] = 1
        rotation = np.array(mathutils.Matrix(mts2c @ to_world @ mts2c.T).to_euler('ZYX')) 
        angles = np.degrees(rotation)
        scale = (106.86535 # camera sensor response
            * (100. / 125.) # exposure time
            * (0.35 * (0.35 / aspect)) # sensor width * sensor height
            / (4. * fstop * fstop)
            * 106.8 / (830. - 360.) # illum e norm / wavelength range
            * np.pi
        )
        env_shader = 'envmap {} {:g} {:.5f} {:.5f} {:.5f} #'.format(filename, 1./scale, *angles)
    env_text = 'Environment'
    if env_text in bpy.data.texts:
        bpy.data.texts.remove(bpy.data.texts[env_text])
    text = bpy.data.texts.new(env_text)
    text.write(env_shader)

if __name__ == '__main__':
    main()
