# run this script in blender (with guided.so present)
# keybindings:
# u - update sampling
# alt 1, 2 or 3 - switch between 1d, 2d or 3d sampling
# alt 5, 6, 7, 8 - switch between independent, specular, smooth ggx and rough ggx

import numpy as np

import bmesh
import bpy

from mathutils import *

# add working directory to the import path
import sys, os
sys.path.append(os.getcwd())
try:
    import guided
except ImportError:
    print('expected guided.so in the directory where blender was started')
    raise

# interpret bounds as 2 stddevs
stddevs = 2.

bl_info = {"name": "Gaussian Sampling Test", "category": "Object", "blender": (2, 80, 0)}

def normalize(v):
    return v / np.linalg.norm(v)

def reflect(i, n):
    return i - 2. * np.dot(n, i) * n

# Sampling the GGX Distribution of Visible Normals [Heitz 2018]
# Input Ve: view direction
# Input alpha_x, alpha_y: roughness parameters
# Input U1, U2: uniform random numbers
# Output Ne: normal sampled with PDF D_Ve(Ne) = G1(Ve) * max(0, dot(Ve, Ne)) * D(Ne) / Ve.z
def sampleGGXVNDF(Ve, alpha_x, alpha_y, u):
    U1 = u[0]
    U2 = u[1]
    # Section 3.2: transforming the view direction to the hemisphere configuration
    Vh = normalize(np.array((alpha_x * Ve[0], alpha_y * Ve[1], Ve[2])));
    # Section 4.1: orthonormal basis
    lensq = Vh[0] * Vh[0] + Vh[1] * Vh[1]
    T1 = np.array((-Vh[1], Vh[0], 0.)) / np.sqrt(lensq) if lensq > 1e-8 else np.array((1., 0., 0.))
    T2 = np.cross(Vh, T1)
    # Section 4.2: parameterization of the projected area
    r = np.sqrt(U1)
    phi = 2.0 * np.pi * U2
    t1 = r * np.cos(phi)
    t2 = r * np.sin(phi)
    s = 0.5 * (1.0 + Vh[2])
    t2 = (1.0 - s)*np.sqrt(1.0 - t1*t1) + s*t2
    # Section 4.3: reprojection onto hemisphere
    Nh = t1*T1 + t2*T2 + np.sqrt(max(0.0, 1.0 - t1*t1 - t2*t2))*Vh
    # Section 3.4: transforming the normal back to the ellipsoid configuration
    Ne = normalize(np.array((alpha_x * Nh[0], alpha_y * Nh[1], max(0.0, Nh[2]))))
    return Ne

def get_onb(n):
    if abs(n[1]) < 0.5:
        up = np.array((0., 1., 0.))
        u = normalize(np.cross(n, up))
    else:
        rg = np.array((1., 0., 0.))
        u = normalize(np.cross(n, rg))
    v = normalize(np.cross(n, u))
    return np.vstack((u, v, n))

def sample_ggx(i, n, roughness):
    # more linear roughness
    alpha = roughness * roughness
    u = np.random.rand(2)
    onb = get_onb(n)
    micronormal = onb.T @ sampleGGXVNDF(onb @ (-i), alpha, alpha, u)
    return reflect(i, micronormal)

def get_gauss_object(name, default_location=(0., 0., 0.), default_scale=0.2):
    if name not in bpy.data.objects:
        # create a new object with a sphere mesh
        mesh = bpy.data.meshes.new(name)
        obj = bpy.data.objects.new(name, mesh)
        bpy.context.scene.collection.objects.link(obj)
        bm = bmesh.new()
        bm.from_mesh(mesh)
        bmesh.ops.create_uvsphere(bm, u_segments=64, v_segments=32, diameter=1.)
        bm.to_mesh(mesh)
        bm.free()
        obj.matrix_world = Matrix.Translation(default_location) @ Matrix.Scale(default_scale, 4)
    return bpy.data.objects[name]

def get_samples_object(name):
    if name not in bpy.data.objects:
        mesh = bpy.data.meshes.new(name)
        obj = bpy.data.objects.new(name, mesh)
        bpy.context.scene.collection.objects.link(obj)
    return bpy.data.objects[name]

def mvn_from_object(obj):
    transform = np.array(obj.matrix_world)
    cov = (transform[0:3, 0:3] / stddevs) @ (transform[0:3, 0:3] / stddevs).T
    mean = transform[0:3, 3]
    return mean, cov

def mvn_from_objects(v1, v2, v3):
    mean_v1, cov_v1 = mvn_from_object(v1)
    mean_v2, cov_v2 = mvn_from_object(v2)
    mean_v3, cov_v3 = mvn_from_object(v3)

    cov = np.zeros((9, 9))
    cov[0:3, 0:3] = cov_v1
    cov[3:6, 3:6] = cov_v2
    cov[6:9, 6:9] = cov_v3

    mean = np.zeros(9)
    # mean xpp
    mean[0:3] = mean_v1
    # mean xp
    mean[3:6] = mean_v2
    # mean xn
    mean[6:9] = mean_v3

    return mean, cov

def lobe_from_mvn(mean, cov):
    accum = guided.gmm_lobe_accum_t()
    lobe = guided.gmm_lobe_t()

    accum_array = np.zeros(55)
    accum_array[0] = 1.

    accum_array[1:10] = mean

    acc = 10
    for i in range(9):
        for j in range(i, 9):
            accum_array[acc] = cov[i][j]
            acc += 1

    guided.accum_set(accum, accum_array)
    guided.gmm_update_lobe(lobe, accum)

    return lobe

class Mode:
    sample1D = 'sample1D'
    sample2D = 'sample2D'
    sample3D = 'sample3D'

class Correlation:
    independent = 'independent'
    specular = 'specular'
    ggx_smooth = 'ggx_smooth'
    ggx_rough = 'ggx_rough'

mode = Mode.sample3D
correlation = Correlation.independent

def sample_triplets(lobe, mean, cov, n=100):
    points = np.random.multivariate_normal(mean=mean, cov=cov, size=n)

    for i in range(n):
        if mode is Mode.sample1D:
            dir_ = normalize(points[i,6:9] - points[i,3:6])
            dist = guided.gmm_sample_dist(lobe, np.random.rand(), points[i,3:6], dir_)
            points[i,6:9] = points[i,3:6] + dist * dir_
        elif mode is Mode.sample2D:
            points[i,6:9] = guided.gmm_sample_pos2d(lobe, np.random.rand(2), points[i,0:3], points[i,3:6])
        elif mode is Mode.sample3D:
            points[i,6:9] = guided.gmm_sample_pos3d(lobe, np.random.rand(3), points[i,0:3], points[i,3:6])
        else:
            points[i,6:9] = 0.

    return points

def sample_with_reflect(mean, cov, n=1000):
    # sample first two vertices and reflect for third vertex
    normal = normalize(normalize(mean[0:3] - mean[3:6]) + normalize(mean[6:9] - mean[3:6]))
    normal_plane = normalize(mean[6:9] - mean[3:6])
    triplets = np.random.multivariate_normal(mean=mean, cov=cov, size=n)
    for i in range(n):
        if correlation is Correlation.specular:
            outgoing = normalize(reflect(triplets[i, 3:6] - triplets[i, 0:3], normal))
        elif correlation is Correlation.ggx_smooth:
            outgoing = normalize(sample_ggx(triplets[i, 3:6] - triplets[i, 0:3], normal, 0.05))
        elif correlation is Correlation.ggx_rough:
            outgoing = normalize(sample_ggx(triplets[i, 3:6] - triplets[i, 0:3], normal, 0.3))
        d = np.dot(mean[6:9] - triplets[i, 3:6], normal_plane) / np.dot(normal_plane, outgoing)
        d *= 1. + .1 * (2. * np.random.rand() - 1.)
        triplets[i, 6:9] = triplets[i, 3:6] + d * outgoing
    return triplets

def main(context):
    v1 = get_gauss_object('Gauss v1', (0., -1., 1.))
    v2 = get_gauss_object('Gauss v2')
    v3 = get_gauss_object('Gauss v3', (0., 1., 1.))
    v3_out = get_gauss_object('Gauss v3 out', (0., 1., 1.))

    # apply potential object transformations
    bpy.context.view_layer.update()

    mean, cov = mvn_from_objects(v1, v2, v3)
    lobe = lobe_from_mvn(mean, cov)

    if correlation is not Correlation.independent:
        triplets = sample_with_reflect(mean, cov)
        mean = np.mean(triplets, axis=0)
        cov = np.cov(triplets.T)
        lobe = lobe_from_mvn(mean, cov)

    transform_result = np.eye(4)
    u, s, vh = np.linalg.svd(cov[6:9,6:9])
    transform_result[0:3,0:3] = u @ np.diag(np.sqrt(s) * stddevs)
    transform_result[0:3, 3] = mean[6:9]
    v3_out.matrix_world = Matrix(transform_result)

    triplets = sample_triplets(lobe, mean, cov)

    samples_obj = get_samples_object('Samples')

    mesh = samples_obj.data
    bm = bmesh.new()
    for triplet in triplets:
        bm.verts.new(triplet[0:3])
        bm.verts.new(triplet[3:6])
        bm.verts.new(triplet[6:9])
    bm.verts.ensure_lookup_table()
    for i in range(len(triplets)):
        bm.edges.new((bm.verts[i * 3], bm.verts[i * 3 + 1]))
        bm.edges.new((bm.verts[i * 3 + 1], bm.verts[i * 3 + 2]))
    bm.edges.ensure_lookup_table()
    bm.to_mesh(mesh)
    bm.free()

    bpy.context.view_layer.update()

class SampleOperator(bpy.types.Operator):
    bl_idname = "object.sample_gauss"
    bl_label = "Sample Gauss"

    @classmethod
    def poll(cls, context):
        return True

    def execute(self, context):
        main(context)
        return {'FINISHED'}

class Sample1DOperator(bpy.types.Operator):
    bl_idname = "object.sample_gauss1d"
    bl_label = "Sample Gauss 1D"

    @classmethod
    def poll(cls, context):
        return True

    def execute(self, context):
        global mode
        mode = Mode.sample1D
        main(context)
        return {'FINISHED'}

class Sample2DOperator(bpy.types.Operator):
    bl_idname = "object.sample_gauss2d"
    bl_label = "Sample Gauss 2D"

    @classmethod
    def poll(cls, context):
        return True

    def execute(self, context):
        global mode
        mode = Mode.sample2D
        main(context)
        return {'FINISHED'}

class Sample3DOperator(bpy.types.Operator):
    bl_idname = "object.sample_gauss3d"
    bl_label = "Sample Gauss 3D"

    @classmethod
    def poll(cls, context):
        return True

    def execute(self, context):
        global mode
        mode = Mode.sample3D
        main(context)
        return {'FINISHED'}

class SampleIndependentOperator(bpy.types.Operator):
    bl_idname = "object.sample_independent"
    bl_label = "Sample Gauss Independent"

    @classmethod
    def poll(cls, context):
        return True

    def execute(self, context):
        global correlation
        correlation = Correlation.independent
        main(context)
        return {'FINISHED'}

class SampleSpecularOperator(bpy.types.Operator):
    bl_idname = "object.sample_specular"
    bl_label = "Sample Gauss Specular"

    @classmethod
    def poll(cls, context):
        return True

    def execute(self, context):
        global correlation
        correlation = Correlation.specular
        main(context)
        return {'FINISHED'}

class SampleGGXSmoothOperator(bpy.types.Operator):
    bl_idname = "object.sample_ggx_smooth"
    bl_label = "Sample Gauss GGX Smooth"

    @classmethod
    def poll(cls, context):
        return True

    def execute(self, context):
        global correlation
        correlation = Correlation.ggx_smooth
        main(context)
        return {'FINISHED'}

class SampleGGXRoughOperator(bpy.types.Operator):
    bl_idname = "object.sample_ggx_rough"
    bl_label = "Sample Gauss GGX Rough"

    @classmethod
    def poll(cls, context):
        return True

    def execute(self, context):
        global correlation
        correlation = Correlation.ggx_rough
        main(context)
        return {'FINISHED'}

addon_keymaps = []

def register():
    bpy.utils.register_class(SampleOperator)
    bpy.utils.register_class(Sample1DOperator)
    bpy.utils.register_class(Sample2DOperator)
    bpy.utils.register_class(Sample3DOperator)

    bpy.utils.register_class(SampleIndependentOperator)
    bpy.utils.register_class(SampleSpecularOperator)
    bpy.utils.register_class(SampleGGXSmoothOperator)
    bpy.utils.register_class(SampleGGXRoughOperator)

    wm = bpy.context.window_manager
    km = wm.keyconfigs.addon.keymaps.new(name='Object Mode', space_type='EMPTY')
    km.keymap_items.new(SampleOperator.bl_idname, 'U', 'PRESS')
    km.keymap_items.new(Sample1DOperator.bl_idname, 'ONE', 'PRESS', alt=True)
    km.keymap_items.new(Sample2DOperator.bl_idname, 'TWO', 'PRESS', alt=True)
    km.keymap_items.new(Sample3DOperator.bl_idname, 'THREE', 'PRESS', alt=True)

    km.keymap_items.new(SampleIndependentOperator.bl_idname, 'FIVE', 'PRESS', alt=True)
    km.keymap_items.new(SampleSpecularOperator.bl_idname, 'SIX', 'PRESS', alt=True)
    km.keymap_items.new(SampleGGXSmoothOperator.bl_idname, 'SEVEN', 'PRESS', alt=True)
    km.keymap_items.new(SampleGGXRoughOperator.bl_idname, 'EIGHT', 'PRESS', alt=True)
    addon_keymaps.append(km)

def unregister():
    bpy.utils.unregister_class(SampleOperator)
    bpy.utils.unregister_class(Sample1DOperator)
    bpy.utils.unregister_class(Sample2DOperator)
    bpy.utils.unregister_class(Sample3DOperator)

    bpy.utils.unregister_class(SampleIndependentOperator)
    bpy.utils.unregister_class(SampleSpecularOperator)
    bpy.utils.unregister_class(SampleGGXSmoothOperator)
    bpy.utils.unregister_class(SampleGGXRoughOperator)

    wm = bpy.context.window_manager
    for km in addon_keymaps:
        wm.keyconfigs.addon.keymaps.remove(km)
    del addon_keymaps[:]

if __name__ == '__main__':
    register()
