import numpy as np
import struct

class Lobe:
    def __init__(self, id_, lobe_raw=None, is_guide_pairs=True, is_cam_lobe=False):
        self.id = id_
        if lobe_raw is None:
            return
        pos = 0
        self.weight = lobe_raw[0]
        pos += 1
        self.modes_flags = lobe_raw[1]
        pos += 1
        self.E_cond = np.array(lobe_raw[pos:pos+9]).reshape((3,3))
        pos += 9
        self.w_cond = np.array(lobe_raw[pos:pos+3])
        pos += 3
        if is_cam_lobe:
            self.mean = np.array(lobe_raw[pos:pos+3])
            pos += 3
        else:
            if is_guide_pairs:
                self.mean = np.array(lobe_raw[pos:pos+6])
                pos += 6
                self.cond_mean = np.array(lobe_raw[pos:pos+9]).reshape((3, 3))
                pos += 9
                self.E_v1v2_T = np.array(lobe_raw[pos:pos+9]).reshape((3,3))
                pos += 9
                self.w_v1v2_inv = np.array(lobe_raw[pos:pos+3])
                pos += 3
            else:
                self.mean = np.array(lobe_raw[pos:pos+9])
                pos += 6
                self.cond_mean = np.array(lobe_raw[pos:pos+18]).reshape((3, 6))
                pos += 18
                self.E_v1v2_T = np.array(lobe_raw[pos:pos+36]).reshape((6,6))
                pos += 36
                self.w_v1v2_inv = np.array(lobe_raw[pos:pos+6])
                pos += 6
        #self.w_v2, self.E_v2 = np.linalg.eigh(S[0:3, 0:3])
        #self.w_v2 = np.sqrt(self.w_v2)

    def is_guide_pairs(self):
        return len(self.mean) == 6

    def is_cam_lobe(self):
        return len(self.mean) == 3

    def get_cond_weight(self, pos):
        if self.is_cam_lobe():
            return self.weight
        if self.is_guide_pairs():
            logdet = -2. * np.sum(np.log(self.w_v1v2_inv))
            dim = 3
            norm = -0.5 * (dim * np.log(2. * np.pi) + logdet)
            cutoff = 4.
            contained = np.all(np.abs(self.E_v1v2_T @ (pos[0:3] - self.mean[0:3])) * self.w_v1v2_inv <= cutoff)
            if not contained:
                return 0
            d = np.diag(self.w_v1v2_inv) @ self.E_v1v2_T @ (pos[0:3] - self.mean[0:3])
            exponent = -0.5 * np.dot(d, d)
            logpdf = norm + exponent
        else:
            assert(False)
        return self.weight * np.exp(logpdf)

    def contains_point(self, point):
        pass
        """
        cutoff = 4. / 0.729
        mean_v2 = self.mean[0:3] if self.is_guide_pairs() else self.mean[3:6]
        transformed_point = (self.E_v2.T @ (point - mean_v2)) / self.w_v2
        contained = np.all(np.abs(transformed_point) <= cutoff)
        aabb = np.concatenate((mean_v2, mean_v2))
        Bi = self.E_v2.copy()
        for k in range(3):
            Bi[:,k] *= self.w_v2[k] 
        for i in range(8):
            g = np.zeros(3)
            g[0] = cutoff if (i & 1) else -cutoff
            g[1] = cutoff if (i & 2) else -cutoff
            g[2] = cutoff if (i & 4) else -cutoff
            gt = Bi @ g
            corner = mean_v2 + gt
            aabb[:3] = np.minimum(corner, aabb[:3])
            aabb[3:] = np.maximum(corner, aabb[3:])
        aabb_contained = np.all(np.logical_and(
            np.less_equal(aabb[:3], point),
            np.less_equal(point, aabb[3:])
        ))

        if contained:
            assert(aabb_contained)
        return contained, aabb_contained
        """

    def read_from_file(f):
        lobes = []
        guide_dim_file, = struct.unpack('i', f.read(4))
        guide_pairs = guide_dim_file == 6
        assert(guide_dim_file in {6, 9})
        lobe_format = struct.Struct(
            '=' # no alignment
            + 'f' # weight
            + 'I' # modes/flags
            + '9f' # E_cond
            + '3f' # w_cond
            + ('6f' if guide_pairs else '9f') # mean
            + ('9f' if guide_pairs else '18f') # cond mean
            + ('9f' if guide_pairs else '36f') # E_v1v2_T
            + ('3f' if guide_pairs else '6f') # w_v1v2_inv
        )
        n_lobes, = struct.unpack('i', f.read(4))
        buf = f.read(n_lobes * lobe_format.size)
        lobes_raw = lobe_format.iter_unpack(buf)
        lobes = [Lobe(i, lobe_raw, guide_pairs) for i, lobe_raw in enumerate(lobes_raw)]
        return lobes

    def read_from_file_camera(f):
        lobes = []
        lobe_format = struct.Struct(
            '='
            + 'f' # weight
            + 'I' # modes/flags
            + '9f' # E
            + '3f' # w
            + '3f' # mean
        )
        n_lobes, = struct.unpack('i', f.read(4))
        buf = f.read(n_lobes * lobe_format.size)
        lobes_raw = lobe_format.iter_unpack(buf)
        lobes = [Lobe(i, lobe_raw, is_cam_lobe=True, ) for i, lobe_raw in enumerate(lobes_raw)]
        return lobes
