import bpy
from bpy import data as D

from mathutils import *
import bmesh
import numpy as np

def plasma(t):
    t = np.clip(t, 0., 1.)
    c0 = np.array([0.05873234392399702, 0.02333670892565664, 0.5433401826748754])
    c1 = np.array([2.176514634195958, 0.2383834171260182, 0.7539604599784036])
    c2 = np.array([-2.689460476458034, -7.455851135738909, 3.110799939717086])
    c3 = np.array([6.130348345893603, 42.3461881477227, -28.51885465332158])
    c4 = np.array([-11.10743619062271, -82.66631109428045, 60.13984767418263])
    c5 = np.array([10.02306557647065, 71.41361770095349, -54.07218655560067])
    c6 = np.array([-3.658713842777788, -22.93153465461149, 18.19190778539828])
    return np.clip(c0+t*(c1+t*(c2+t*(c3+t*(c4+t*(c5+t*c6))))), 0., 1.)

def apply_lobe_to_collection(lobe, coll, point, cond_weight=1):
    if lobe.is_cam_lobe():
        n_marginals = 1
    else:
        assert(lobe.is_guide_pairs()) # TODO implement for full triplets
        n_marginals = 2 if lobe.is_guide_pairs() else 3

    stddevs = 4.0/0.729
    if lobe.is_cam_lobe():
        means = [lobe.mean[0:3], lobe.mean[0:3]]
    else:
        means = [
            lobe.mean[0:3],
            lobe.mean[3:6] + lobe.cond_mean @ (point - lobe.mean[0:3]),
        ]
    for lobe_obj in coll.objects:
        if lobe_obj.name[:2] in {'v1', 'v2', 'v3'}:
            k = int(lobe_obj.name[1]) - 1
            if lobe.is_guide_pairs():
                k -= 1
            if lobe.is_cam_lobe():
                k -= 1
            offset = 3 * k
            W_, E_ = None, None
            if k == 0:
                w_ = 1./lobe.w_v1v2_inv
                E_ = lobe.E_v1v2_T.T
            elif k == 1:
                w_, E_ = lobe.w_cond, lobe.E_cond
            else:
                assert(False)
            transform = np.eye(4)
            transform[0:3, 0:3] = E_ @ np.diag(w_ * stddevs)
            transform[0:3, 3] = means[k]
            lobe_obj.matrix_world = Matrix(transform)
            color = plasma(cond_weight)**(2.2)
            lobe_obj.active_material.diffuse_color = (color[0], color[1], color[2], 1.)
        else:
            assert(lobe_obj.name.startswith('conn'))
            bm = bmesh.new()
            bm.from_mesh(lobe_obj.data)
            bm.verts.ensure_lookup_table()
            bm.verts[0].co = means[0]
            bm.verts[1].co = means[1]
            bm.to_mesh(lobe_obj.data)
            bm.free()

class LobeVis:
    def __init__(self, all_lobes, all_octrees=None, all_mixtures=None, show_all=False):
        if 'lobes' in D.collections:
            collection = D.collections['lobes']
            # clear collection
            while collection.children:
                collection.children.unlink(collection.children[0])
        else:
            coll = D.collections.new('lobes')
            bpy.context.scene.collection.children.link(coll)
        self.all_lobes = all_lobes
        self.all_octrees = all_octrees
        self.all_mixtures = all_mixtures
        self.show_all = show_all

    def get_sphere_mesh():
        mesh_name = 'vis_sphere'
        if mesh_name in D.meshes:
            return D.meshes[mesh_name]
        sphere_mesh = bpy.data.meshes.new(mesh_name)
        bm = bmesh.new()
        bm.from_mesh(sphere_mesh)
        bmesh.ops.create_uvsphere(bm, u_segments=128, v_segments=64, radius=0.5)
        for f in bm.faces:
            f.smooth = True
        bm.to_mesh(sphere_mesh)
        bm.free()
        return sphere_mesh

    def new_conn_mesh():
        conn_mesh = bpy.data.meshes.new('vis_conn')
        bm = bmesh.new()
        bm.from_mesh(conn_mesh)
        bm.verts.new((0, 0, 0))
        bm.verts.new((1, 0, 0))
        bm.verts.ensure_lookup_table()
        bm.edges.new((bm.verts[0], bm.verts[1]))
        bm.edges.ensure_lookup_table()
        bm.to_mesh(conn_mesh)
        bm.free()
        return conn_mesh

    def update(self, frame, pos):
        frame = np.clip(frame, 1, len(self.all_lobes)) - 1
        lobes = self.all_lobes[frame]
        lobes_to_show = []
        if self.all_octrees is not None:
            octree = self.all_octrees[frame]
            octree_aabb = octree.lookup_aabb(pos)
        else:
            octree_aabb = None
        """
        for lobe in lobes:
            do_show = True
            if octree_aabb is not None:
                do_show = np.all(lobe.octree_pos >= octree_aabb[:3]) and np.all(lobe.octree_pos <= octree_aabb[3:])
            # to show lobes in support
            #do_show, _ = lobe.contains_point(np.array(pos))
            if do_show:
                lobes_to_show.append(lobe)
        """
        if self.all_mixtures is not None:
            matching_mixtures = 0
            for mixture in self.all_mixtures[frame]:
                if np.all(mixture.octree_pos >= octree_aabb[:3]) and np.all(mixture.octree_pos <= octree_aabb[3:]):
                    lobes_to_show.extend(lobes[mixture.offset:mixture.offset+mixture.n_lobes])
                    matching_mixtures += 1
            assert(matching_mixtures <= 1)
        else:
            lobes_to_show.extend(lobes)

        if self.show_all:
            cond_weights = np.array([lobe.weight for lobe in lobes_to_show], dtype=np.float64)
        else:
            cond_weights = np.zeros(len(lobes_to_show))
            for i, lobe in enumerate(lobes_to_show):
                cond_weights[i] = lobe.get_cond_weight(np.array(pos))
        weight_sum = np.sum(cond_weights)
        if np.isclose(weight_sum, 0):
            return
        cond_weights /= weight_sum

        lobes_to_show = [lobe for lobe, w in zip(lobes_to_show, cond_weights) if w > 0.]
        cond_weights = cond_weights[cond_weights > 0.]
        assert(len(lobes_to_show) == len(cond_weights))

        lobe_collection = D.collections['lobes']
        sphere_mesh = LobeVis.get_sphere_mesh()
        colls_needed = len(lobes_to_show)
        while len(lobe_collection.children) < colls_needed:
            coll = D.collections.new('lobe ?')
            if not lobes_to_show[0].is_cam_lobe():
                if not lobes_to_show[0].is_guide_pairs():
                    obj_v1 = bpy.data.objects.new('v1', sphere_mesh)
                    material_v1 = bpy.data.materials.new('v1 material')
                    obj_v1.active_material = material_v1
                    obj_v1.material_slots[0].link = 'OBJECT'
                    obj_v1.material_slots[0].material = material_v1
                    coll.objects.link(obj_v1)
                obj_v2 = bpy.data.objects.new('v2', sphere_mesh)
                material_v2 = bpy.data.materials.new('v2 material')
                obj_v2.active_material = material_v2
                obj_v2.material_slots[0].link = 'OBJECT'
                obj_v2.material_slots[0].material = material_v2
                coll.objects.link(obj_v2)
            obj_v3 = bpy.data.objects.new('v3', sphere_mesh)
            material_v3 = bpy.data.materials.new('v3 material')
            obj_v3.active_material = material_v3
            obj_v3.material_slots[0].link = 'OBJECT'
            obj_v3.material_slots[0].material = material_v3
            coll.objects.link(obj_v3)
            obj_conn = bpy.data.objects.new('conn', LobeVis.new_conn_mesh())
            coll.objects.link(obj_conn)
            lobe_collection.children.link(coll)
        for lobe_coll in lobe_collection.children:
            lobe_coll.name = 'unused'
        for i, lobe_coll in enumerate(lobe_collection.children):
            if i < colls_needed:
                lobe = lobes_to_show[i]
                lobe_coll.hide_viewport = False
                lobe_coll.name = 'lobe {}'.format(lobe.id)
                apply_lobe_to_collection(lobe, lobe_coll, np.array(pos), cond_weights[i])
            else:
                lobe_coll.hide_viewport = True
